///	\file			MFCManager.h
///	\date			2022
///	\copyright (c)	Metrolab Technology S.A., Geneva, Switzerland (www.metrolab.com)
///	See the included file License Terms.pdf for the licensing conditions.

#pragma once

// Standard includes
#include <iostream>
#include <atomic>
#include <exception>
#include <unordered_map>

// Personnal includes
#include "InstrumentCatalog.h"
#include "MFCController.h"
#include "MFC3045Device.h"
#include "MFC2x46Device.h"
#include "USBRemoteBoxController.h"
#include "Helpers.h"
#include "MFCManagerCatalog.h"

using namespace MTL;
using namespace MTL::Instrument;
using namespace MTL::ThreadedMailbox;

namespace MTL {
	//--------------------------------------------------------------------------//
	//								MFC MANAGER
	//--------------------------------------------------------------------------//
	struct sMFCManagerMailBoxMsg {
		CAsyncNotification *						pCatNot;		// Global catalog notification
		CAsyncNotification *						pContrNot;		// Controller notification
		sMFCManagerMailBoxMsg()
			: pCatNot(nullptr), pContrNot(nullptr) {}
		sMFCManagerMailBoxMsg(CAsyncNotification * pCat, CAsyncNotification * pContr)
			: pCatNot(pCat), pContrNot(pContr) {}
	};
	class CMFCManager : private CThreadedMailBox<sMFCManagerMailBoxMsg>
	{
	//--------------------------------------------------------------------------//
	//								Utilities
	//--------------------------------------------------------------------------//
	private:
		typedef std::string		tUniqueID;
		struct sPrettyResource {
			std::string			PrettyName;
			sParsedResource		ParsedResource;
		};
		typedef std::unordered_map<tUniqueID, sPrettyResource>	tResourceMap;
		tResourceMap l_PrettyResourceNameList(const CParsedResourceList & rRsrcList);

	private:
		//--------------------------------------------------------------------------//
		//							Class attributes
		//--------------------------------------------------------------------------//
		CInstrManCatalogInterface					m_Cat;
		CMFCManagerCatalog							m_ContCatInstance;
		CMFCManagerInterface						m_ContCat;
		CAsyncDynamicNotifier						m_ADNotifier;
		CAsyncSnaphsotNotifier						m_ASControllerNotifier;
		CVISAResourceManager						m_RsrcManager;
		tResourceMap								m_RsrcListMap;
		CMFCDevice *								m_pMFCDevice = nullptr;
		CMFCController								m_MFCController;
		CMFCManagerCatalog::eControllerState		m_MFCControllerState;
		CUSBRemoteBoxController						m_RBController;
		CInstrumentCatalog::eRemoteBoxBusyLedState	m_RBBusyLedState;

		std::string									m_DefaultPAConfScriptFilePath;		// Path of the default PA configuration script (if no specific file exists)
		std::string									m_CustomPAConfScriptDirectory;		// Directory containing PA specific configuration scripts

		// Local versions of controls
		CInstrumentCatalog::tInstrumentList			m_PrettyInstrList;
		CInstrumentCatalog::sInstrumentInformation	m_InstrInfo;
		//CInstrumentCatalog::sProbeArrayInformation	m_PAInfo; // no need to keep in Manager
		//CInstrumentCatalog::sProbeArrayAngle		m_PAAngle;
		CInstrumentCatalog::sInstrumentStatus		m_InstrStatus;
		//		Normalization is ephemeral
		//		Search result is ephemeral
		//		Measurement results are ephemeral
		//		Error is ephemeral
		// Local versions of observers
		CInstrumentCatalog::sInstrumentConnection		m_Instr;
		CInstrumentCatalog::sMFC3045SerialPortSettings	m_MFC3045SerialSett;
		CInstrumentCatalog::sInstrumentState			m_InstrState;
		//		Instrument calibration date is ephemeral
		//		Normalization returned is ephemeral
		CInstrumentCatalog::sLicenseInformation			m_LicInfo;

		//--------------------------------------------------------------------------//
		//							Local private methods
		//--------------------------------------------------------------------------//
		// Handlers for new values of catalog observers
		void l_ClearInstrumentInfo();
		void l_NewInstrumentCurrent(const CInstrumentCatalog::sInstrumentConnection & rNewInstr);
		void l_NewMFC3045SerialSettings(const CInstrumentCatalog::sMFC3045SerialPortSettings & rNewSerialSett);
		void l_NewInstrumentState(const CInstrumentCatalog::sInstrumentState & rNewInstrState);
		void l_NewNormalizationRequest(const CInstrumentCatalog::sNormalization & rNewNormalizationRequest);
		void l_GetDongleLicenseInfo();
		void l_NewRBBusyLedState(const CInstrumentCatalog::eRemoteBoxBusyLedState & rNewRBBusyLedState);

		// Private operations
		bool l_UpdateRBBusyLedState(const CInstrumentCatalog::eRemoteBoxBusyLedState & rNewRBBusyLedState);
		void l_UpdateInstrumentStatus();
		void l_SendError(const CMFCDevice::sMFCDeviceError & rMFCDeviceError);
		void l_SendError(CInstrumentCatalog::eErrorLevel Level, std::string Descr, std::string Source);
		void l_SendError(CInstrumentCatalog::sError Error);
		void l_Log(const CInstrumentCatalog::sLogMessage & rMessage);
		void l_UpdateSPR(uint8_t SearchProgress);

		// Local Thread
		void l_ThreadTask(CThreadedMailBox::eEventType EvtType, sMFCManagerMailBoxMsg * pMsg);
		// Lives in the main thread (outside of this class)
		void l_CatalogNotificationHandler(CAsyncNotification * pANot);
		void l_ControllerNotificationHandler(CAsyncNotification * pASNot);

		//--------------------------------------------------------------------------//
		//						Catalog Notification handling
		//--------------------------------------------------------------------------//
	public :
		void PostCatalogNotification(CAsyncNotification * pANot);		// Receives asynchronous notifications from global catalog
		//--------------------------------------------------------------------------//
		//					MFCManager Catalog Notification handling
		//--------------------------------------------------------------------------//
		void PostControllerNotification(CAsyncNotification * pASNot);	// Receives asyncrhonous notifications from MFCController

		//--------------------------------------------------------------------------//
		//							MFC Controller Interface
		//--------------------------------------------------------------------------//
	public:
		void SlotControllerThreadIsReady();

		void SlotControllerState(CMFCManagerCatalog::eControllerState State);
		void SlotInstrumentList(CParsedResourceList List);
		void SlotPAInfo(CInstrumentCatalog::sProbeArrayInformation PAInfo);
		void SlotInstrConfig(CInstrumentCatalog::sInstrumentConfiguration InstrConfig);
		void SlotProbeArrayAngle(CInstrumentCatalog::sProbeArrayAngle Angle);
		void SlotRemoteBoxTriggerEvent();
		void SlotSearchProgress(U8 Progress);
		void SlotSearchResult(CInstrumentCatalog::sSearchResult SearRes);
		void SlotMeasResult(CInstrumentCatalog::sMeasurementResults MeasRes);
		void SlotControllerError(CMFCDevice::sMFCDeviceError error);
		void SlotControllerLogMessage(CInstrumentCatalog::sLogMessage Message);

		//--------------------------------------------------------------------------//
		//						Remote Box Controller Interface
		//--------------------------------------------------------------------------//
	public :
		void SlotRBReset();
		void SlotRBStart();

	public:
		//--------------------------------------------------------------------------//
		//						Constructor / Destructor
		//--------------------------------------------------------------------------//
		CMFCManager(CInstrumentCatalog & rCatalog, const std::string & rDefaultPAConfScriptFilePath, const std::string & rCustomPAConfScriptDir);
		virtual ~CMFCManager();

		//--------------------------------------------------------------------------//
		//							Class control
		//--------------------------------------------------------------------------//
		void SyncControls();
		void SyncObservers();
		void Subscribe();
		void Operate();

	public:
		void Terminate();
	}; // class CMFCManager

} // namespace MTL
