///	\file			MFC3045Device.cpp
///	\date			2022
///	\copyright (c)	Metrolab Technology S.A., Geneva, Switzerland (www.metrolab.com)
///	See the included file License Terms.pdf for the licensing conditions.

// Standard includes
#include <regex>
#include <bitset>

// Personal includes
#include "MFC3045Device.h"
#include "Exception.h"

#if defined(MFC_NORMALIZATION_ENABLED)
#include "MFCProLicInterface.h"
#endif

// External tools includes
#include "date.h"

#define LOG(__msg__, __level__,__location__)		m_rMFCController.l_SendLogMessage(__msg__, __level__, ".MFC3045Device", __location__)


//------------------------------------------------------------------//
//					Constructors / Destructors
//------------------------------------------------------------------//
CMFC3045Device::CMFC3045Device (CMFCController &	rMFCController, CVISAResourceManager & rRM, tResourceName Rsrc, CInstrumentCatalog::sMFC3045SerialPortSettings & rSerialSetting)
	:
	  m_rMFCController(rMFCController),
	  m_MFC3045(rRM, Rsrc),
	  m_PaSerialNum(""),
	  m_MaxMeasTimeOutMs(0)
{
	l_ParseSerialSettings(rSerialSetting, m_SerialParms);
}

CMFC3045Device::~CMFC3045Device ()
{

}

//--------------------------------------------------------------------------//
//							Utilities
//--------------------------------------------------------------------------//
template<typename T>
static std::string toBinaryString(const T& x)
{
	std::stringstream ss;
	ss << std::bitset<sizeof(T) * 8>(x);
	return ss.str();
}

F64 l_ConvertdHzToMHz (const U32 & rFreqdHz)
{
	return static_cast <F64>(rFreqdHz)*1.0e-7;
}

U32 l_ConvertMHzTodHz (const F64 & rFreqMHz)
{
	return static_cast <U32>(rFreqMHz*1.0e7);
}

F32 l_ConvertStdDevdHzToPpm (const U32 & rStdDevdHz, const U32 & rFreqRefdHz)
{
	F64 l_StdDevppm(DPqNaN);
	if (rFreqRefdHz > 0 && rStdDevdHz > 0)
		l_StdDevppm= (static_cast <F64>(rStdDevdHz) * 1.0e6) / static_cast <F64>(rFreqRefdHz);
	return static_cast <F32>(l_StdDevppm);
}

bool CMFC3045Device::l_UtcToStrDateISO8601(const std::time_t & rUtcDate, std::string & rStrDate)
{
	try {
		auto l_TimePointUTC = std::chrono::system_clock::from_time_t(rUtcDate);
		rStrDate = date::format("%FT%TZ", date::floor<std::chrono::seconds>(l_TimePointUTC)); // convert date time_t UTC in YYYY-MM-DDTHH:MM:SSZ ISO8601 UTC string format
		if(rStrDate.empty())
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Failed to format time_t UTC date to string", MTL__LOCATION__);
	}
	catch (sMFCDeviceError & rE)
	{
		MTL_Unused(rE)
		rStrDate.clear();
		return false;
	}
	return true;
}

bool CMFC3045Device::l_ParseSerialSettings(const CInstrumentCatalog::sMFC3045SerialPortSettings & rSerialSetting, sSerialPortSettings & rVisaSerialSetting)
{
	bool l_Ret(true);
	// Convert CInstrumentCatalog::sMFC3045SerialPortSettings to sSerialPortSettings (VISAInstrumentTypes.h)

	switch (rSerialSetting.Baudrate)
	{
	case CInstrumentCatalog::eSerialBaudrate::k2400:
		rVisaSerialSetting.Baudrate = eSerialBaudrate::k2400;
		break;
	case CInstrumentCatalog::eSerialBaudrate::k4800:
		rVisaSerialSetting.Baudrate = eSerialBaudrate::k4800;
		break;
	case CInstrumentCatalog::eSerialBaudrate::k9600:
		rVisaSerialSetting.Baudrate = eSerialBaudrate::k9600;
		break;
	case CInstrumentCatalog::eSerialBaudrate::k19200:
		rVisaSerialSetting.Baudrate = eSerialBaudrate::k19200;
		break;
	case CInstrumentCatalog::eSerialBaudrate::k28800:
		rVisaSerialSetting.Baudrate = eSerialBaudrate::k28800;
		break;
	case CInstrumentCatalog::eSerialBaudrate::k38400:
		rVisaSerialSetting.Baudrate = eSerialBaudrate::k38400;
		break;
	case CInstrumentCatalog::eSerialBaudrate::k57600:
		rVisaSerialSetting.Baudrate = eSerialBaudrate::k57600;
		break;
	case CInstrumentCatalog::eSerialBaudrate::k115200:
		rVisaSerialSetting.Baudrate = eSerialBaudrate::k115200;
		break;
	default:
		LOG("Unhandled Serial Baudrate type", CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);
		l_Ret = false;
		break;
	}

	switch (rSerialSetting.Parity)
	{
	case CInstrumentCatalog::eSerialParity::kNone:
		rVisaSerialSetting.Parity = eSerialParity::kNone;
		break;
	case CInstrumentCatalog::eSerialParity::kOdd:
		rVisaSerialSetting.Parity = eSerialParity::kOdd;
		break;
	case CInstrumentCatalog::eSerialParity::kEven:
		rVisaSerialSetting.Parity = eSerialParity::kEven;
		break;
	default:
		LOG("Unhandled Serial Parity type", CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);
		l_Ret = false;
		break;
	}

	switch (rSerialSetting.DataBits)
	{
	case CInstrumentCatalog::eSerialDataBits::k7:
		rVisaSerialSetting.DataBits = eSerialDataBits::k7;
		break;
	case CInstrumentCatalog::eSerialDataBits::k8:
		rVisaSerialSetting.DataBits = eSerialDataBits::k8;
		break;
	default:
		LOG("Unhandled Serial Databits type", CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);
		l_Ret = false;
		break;
	}

	switch (rSerialSetting.StopBits)
	{
	case CInstrumentCatalog::eSerialStopBits::k1:
		rVisaSerialSetting.StopBits = eSerialStopBits::k1;
		break;
	case CInstrumentCatalog::eSerialStopBits::k2:
		rVisaSerialSetting.StopBits = eSerialStopBits::k2;
		break;
	default:
		LOG("Unhandled Serial Stopbits type", CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);
		l_Ret = false;
		break;
	}

	switch (rSerialSetting.Handshake)
	{
	case CInstrumentCatalog::eSerialHandshake::kNone:
		rVisaSerialSetting.Handshake = eSerialHandshake::kNone;
		break;
	case CInstrumentCatalog::eSerialHandshake::kXonXoff:
		rVisaSerialSetting.Handshake = eSerialHandshake::kXonXoff; // software handshake
		break;
	case CInstrumentCatalog::eSerialHandshake::kHardware:
		rVisaSerialSetting.Handshake = eSerialHandshake::kHardware; // CTS/RTS hardware handshake
		break;
	default:
		LOG("Unhandled Serial Handshake type", CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);
		l_Ret = false;
		break;
	}
	return l_Ret;
}

bool CMFC3045Device::l_SerialCommunicationSuccess()
{
	bool l_Ret(true);

	if(!m_MFC3045.StatusGet(eStatusRegister::kStatus5, m_AllRegStatus.StReg5.RawSTB5))
		l_Ret = false;
	else
	{
		if(m_AllRegStatus.StReg5.StatusByte5.EFR || m_AllRegStatus.StReg5.StatusByte5.EOR || m_AllRegStatus.StReg5.StatusByte5.EPY)
			l_Ret = false;
	}
	return l_Ret;
}

bool CMFC3045Device::l_CheckPAConnected(bool & rPAPresent) //Check PA connection with a shorter communication time out
{
	U32 l_TimeoutMs(0);
	m_MFC3045.GetCurrentTimeoutRdDataMs(l_TimeoutMs);
	try
	{
		m_MFC3045.SetTimeoutRdNewDataMs(2000); //Set temporary to 2 seconds timeout to abort quickly if no more Instrument connected (communication error)
		if(!m_MFC3045.StatusGet(eStatusRegister::kStatus2, m_AllRegStatus.StReg2.RawSTB2))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Unable to read PA connection in Status register", MTL__LOCATION__);
		m_MFC3045.SetTimeoutRdNewDataMs(l_TimeoutMs); //Restore previous timeout
		rPAPresent = (0 == m_AllRegStatus.StReg2.StatusByte2.PANC);
		return true;
	}
	catch (sMFCDeviceError & rE)
	{
		MTL_Unused(rE)
		m_MFC3045.SetTimeoutRdNewDataMs(l_TimeoutMs); //Restore previous timeout
		rPAPresent = false;
		return false;
	}
	return false;
}

bool CMFC3045Device::l_ConfigurationChanged (CInstrumentCatalog::sInstrumentState lastcfg, CInstrumentCatalog::sInstrumentState newcfg)
{
	return (lastcfg.OperatingMode != newcfg.OperatingMode ||
			lastcfg.ProbeSelection.SelectionMode != newcfg.ProbeSelection.SelectionMode ||
			*(reinterpret_cast<U64*>(&lastcfg.FreqMin)) != *(reinterpret_cast<U64*>(&newcfg.FreqMin)) ||		// Compare binary representation of the double values to detect value change
			*(reinterpret_cast<U64*>(&lastcfg.FreqMax)) != *(reinterpret_cast<U64*>(&newcfg.FreqMax)) ||		// Compare binary representation of the double values to detect value change
			lastcfg.NbAveragedMeasurements != newcfg.NbAveragedMeasurements);
}

bool CMFC3045Device::l_ConfigureForOperation (const CInstrumentCatalog::sInstrumentState & rInstrState, sMFCDeviceError & rError)
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);

	//-----------------------------------------------------
	// According to the current sInstrumentState, set main unit (MFC3045) parameters:
	//  FreqMin
	//  FreqMax
	//  NbAveragedMeasurements
	//  Data block mode to decimal

	//Configure Main unit only if configuration has changed
	if (l_ConfigurationChanged(m_LastConfigured, rInstrState)) //Check if new configuration is different than current one
	{
		m_MaxMeasTimeOutMs =0;
		uStatusRegBitSel l_SelStatus;
		LOG("Reconfiguration needed", CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);
		try
		{
			U32 l_FreqdHz(0);
			U16 l_NbMeasCylce(0);

			switch (rInstrState.OperatingMode)
			{
			case CInstrumentCatalog::eOperatingMode::kSearch:
				if (!m_MFC3045.BreakMeasSet())
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not stop any instrument acquisition", MTL__LOCATION__);
				if (!m_MFC3045.DisableSMA())
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not disable SMA on instrument", MTL__LOCATION__);
				if (!m_MFC3045.DataBlockModeDecimalSet())
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not set block mode to decimal", MTL__LOCATION__);
				// Search is always done on all Probe array range. Don't need to configure measurement frequency limits
				break;

			case CInstrumentCatalog::eOperatingMode::kMeasure:
				if (!m_MFC3045.BreakMeasSet())
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not stop any instrument acquisition", MTL__LOCATION__);
				if (!m_MFC3045.DisableSMA())
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not disable SMA on instrument", MTL__LOCATION__);
				// Set measurement limits
				if(!m_MFC3045.AdvancedCommandModeSet(eAdvancedMode::k1))
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Failed to change MU Advanced mode", MTL__LOCATION__);
				//Set MRE to reference = MLF (Fmin) for future measurement
				if(!m_MFC3045.ModulationReferenceSet(eModulationRef::kMLF))
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Failed to set MRE register", MTL__LOCATION__);
				if(!m_MFC3045.AdvancedCommandModeSet(eAdvancedMode::kNone))
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Failed to change MU Advanced mode", MTL__LOCATION__);
				// set MLF & MHF (MDA et MCF will be computed from MLF and MHF in the main unit
				l_FreqdHz = l_ConvertMHzTodHz(rInstrState.FreqMin);
				if (!m_MFC3045.ModulationLowestFreqSet(l_FreqdHz))
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not set min Freq measurement", MTL__LOCATION__);
				l_FreqdHz = l_ConvertMHzTodHz(rInstrState.FreqMax);
				if (!m_MFC3045.ModulationHighestFreqSet(l_FreqdHz))
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not set max Freq measurement", MTL__LOCATION__);

				// Set measurement averaging (number of measurement cycle)
				if(rInstrState.NbAveragedMeasurements <= NCY_MIN)
					l_NbMeasCylce = static_cast<U16>(NCY_MIN);
				else if(rInstrState.NbAveragedMeasurements >= NCY_MAX)
					l_NbMeasCylce = static_cast<U16>(NCY_MAX);
				else
					l_NbMeasCylce = static_cast<U16>(rInstrState.NbAveragedMeasurements);
				if (!m_MFC3045.NumberMeasCycleSet(l_NbMeasCylce))
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not set number of measurement cycle", MTL__LOCATION__);

				if (!m_MFC3045.DataBlockModeDecimalSet())
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not set block mode to decimal", MTL__LOCATION__);
				break;

			default:
				break;
			}
			if(!l_ComputeMaxMeasTimeout(m_MaxMeasTimeOutMs)) //Get maximum time to wait for a measuremenr with the modulation parameters set
			{
				m_MaxMeasTimeOutMs = 0;
			}
			rError = sMFCDeviceError(sMFCDeviceError::eLevel::kNoError, "Configuration done.", MTL__LOCATION__);
		}
		catch (sMFCDeviceError & rE)
		{
			rError = rE;
			return false;
		}
		l_SelStatus.SelSTReg = BitSelST1;
		if(!m_MFC3045.StatusGet(l_SelStatus, m_AllRegStatus)) //read status register 1 to see if errors occured
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Status registers read failed", MTL__LOCATION__);
		if (m_AllRegStatus.StReg1.StatusByte1.ECMD || m_AllRegStatus.StReg1.StatusByte1.ECOM) //check serial command and communication error
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Error on Status 1: 0b" + toBinaryString(m_AllRegStatus.StReg1.RawSTB1), MTL__LOCATION__);
		if (m_AllRegStatus.StReg1.StatusByte1.EMOD) //check modulation error
		{
			l_SelStatus.SelSTReg = BitSelST2; // Read details of modulation error in Status 2 register.
			if(!m_MFC3045.StatusGet(l_SelStatus, m_AllRegStatus))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Status registers read failed", MTL__LOCATION__);
			if (m_AllRegStatus.StReg2.StatusByte2.MED || m_AllRegStatus.StReg2.StatusByte2.PANC) //check Modulation setting error
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Error on Status 2: 0b" + toBinaryString(m_AllRegStatus.StReg2.RawSTB2), MTL__LOCATION__);
		}
	}
	m_LastConfigured = rInstrState;
	return true;
}

bool CMFC3045Device::l_ComputeMaxMeasTimeout(U64 & rMaxTimeOutMs)
{
	rMaxTimeOutMs = 0;
	try
	{
		U16 l_NbMeasCylce(0);
		U16 l_NbPreMeasCylce(100);
		U16 l_NbModPeriod(0);

		if (!m_MFC3045.NumberMeasCycleGet(l_NbMeasCylce))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not get number of measurement cycle", MTL__LOCATION__);

		if (!m_MFC3045.ModulationPeriodGet(l_NbModPeriod))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not get modulation period", MTL__LOCATION__);

		// register access with Advance mode 1:
		if(!m_MFC3045.AdvancedCommandModeSet(eAdvancedMode::k1))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Failed to change MU Advanced mode", MTL__LOCATION__);
		if (!m_MFC3045.NumberPreCycleGet(l_NbPreMeasCylce))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not get number of pre cycle", MTL__LOCATION__);
		if(!m_MFC3045.AdvancedCommandModeSet(eAdvancedMode::kNone))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Failed to change MU Advanced mode", MTL__LOCATION__);
		rMaxTimeOutMs = 4*(l_NbModPeriod * (l_NbMeasCylce + l_NbPreMeasCylce));
	}
	catch (sMFCDeviceError & rE)
	{
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Compute the average of the measurements made during the search operation.
void CMFC3045Device::l_ComputeAveraging (const std::vector<U32> & rFluxListdHz, F64 & rReturnedAveragingMHz, F64 & rReturnedStdDevppm)
{
	// Compute the mean of the measurements.
	U8 l_ValidCount = 0;
	rReturnedAveragingMHz	= 0.0l;
	rReturnedStdDevppm		= 0.0l;
	try
	{
		// Check content
		if (rFluxListdHz.empty())
			throw false;

		// Compute averaging
		for (std::vector<U32>::const_iterator l_it = rFluxListdHz.begin(); l_it != rFluxListdHz.end(); l_it++)
		{
			// Make sure that the measurement taken to compute the mean are valid.
			// An invalid measurement is marked with a mean value being equal to 0.
			if (*l_it >0)
			{
				// Sum up all the NMR frequencies measured.
				rReturnedAveragingMHz += l_ConvertdHzToMHz(*l_it);
				// Update the number of valid heads.
				l_ValidCount++;
			}
		}
		// If we had some valid heads, compute the average.
		if (l_ValidCount > 0)
			rReturnedAveragingMHz = rReturnedAveragingMHz / l_ValidCount;
		else
			rReturnedAveragingMHz = DPqNaN;

		// We already know how many valid heads are available.
		if (l_ValidCount > 1)
		{
			for (std::vector<U32>::const_iterator l_it = rFluxListdHz.begin(); l_it != rFluxListdHz.end(); l_it++)
			{
				// Make sure that the measurement taken to compute the mean are valid.
				// An invalid measurement is marked with a mean value being equal to 0.
				if (*l_it >0)
				{
					F64 l_Value = l_ConvertdHzToMHz(*l_it) - rReturnedAveragingMHz;
					rReturnedStdDevppm += l_Value * l_Value;
				}
			}
			// Compute the standard deviation in ppm from the computed mean.
			rReturnedStdDevppm = sqrt(rReturnedStdDevppm / (l_ValidCount - 1)) / rReturnedAveragingMHz * 1.0e6;
		}
		else
			rReturnedStdDevppm = DPqNaN;
	}
	catch (bool & rE)
	{
		MTL_Unused(rE)
		rReturnedAveragingMHz	= DPqNaN;
		rReturnedStdDevppm		= DPqNaN;
	}
}

bool CMFC3045Device::AbortInstrument ()
{
	return m_MFC3045.Abort();
}

bool CMFC3045Device::IsInstrumentConnected()
{
	U32 l_TimeoutMs(0);
	bool l_Connected(false);
	m_MFC3045.GetCurrentTimeoutRdDataMs(l_TimeoutMs);
	m_MFC3045.SetTimeoutRdNewDataMs(2000);
	l_Connected = m_MFC3045.CheckInstrumentConnection(2);
	m_MFC3045.SetTimeoutRdNewDataMs(l_TimeoutMs);
	return l_Connected;
}

//--------------------------------------------------------------------------//
//							Available Operations
//--------------------------------------------------------------------------//
// Connect to MFC3045 and get instrument information
CMFC3045Device::sMFCDeviceError CMFC3045Device::Connect(CInstrumentCatalog::sInstrumentInformation & rInstrInfo)
{
	try
	{
		uStatusRegBitSel l_SelStatus;

		if (!m_MFC3045.ConnectInstrument(m_SerialParms,8000)) //connect instrument and configure 8 seconds of timeout on communication
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Cannot connect to requested resource", MTL__LOCATION__);
		l_SelStatus.SelSTReg = BitSelSTALL; // read all status registers
		if(!m_MFC3045.StatusGet(l_SelStatus, m_AllRegStatus))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Status registers read failed", MTL__LOCATION__);
		if (m_AllRegStatus.StReg1.StatusByte1.ECMD || m_AllRegStatus.StReg1.StatusByte1.ECOM) //check serial command and communication error
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Error on Status 1: 0b" + toBinaryString(m_AllRegStatus.StReg1.RawSTB1), MTL__LOCATION__);
		if (m_AllRegStatus.StReg4.StatusByte4.HWP || m_AllRegStatus.StReg4.StatusByte4.EMM) //check hardware and MUEEPROM error
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Error on Status 4: 0b" + toBinaryString(m_AllRegStatus.StReg4.RawSTB4), MTL__LOCATION__);
		// register access with Advance mode 1:
		if(!m_MFC3045.AdvancedCommandModeSet(eAdvancedMode::k1))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Failed to change MU Advanced mode", MTL__LOCATION__);
		// Check host and MFC3045 serial config to be the same.
		U16 l_ReadRSPRegValue(0);
		if(!m_MFC3045.SerialRSPGet(l_ReadRSPRegValue))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Failed to read RSP register", MTL__LOCATION__);
		U16 l_SerialSettings;
		if(!m_MFC3045.ParseParmForRSP(m_SerialParms,l_SerialSettings))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Failed to parse serial settings to RSP register type", MTL__LOCATION__);
		if(l_ReadRSPRegValue != l_SerialSettings)
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Incompatible host and instrument serial settings. MFC3045 Settings: 0b" + toBinaryString(l_ReadRSPRegValue), MTL__LOCATION__);

		std::string l_RdString;
		F32 l_FWVersionValue(0.0);
		//rInstrInfo.InstrumentConnection is filled in MFCManager
		rInstrInfo.Type = CInstrumentCatalog::eInstrumentType::kMFC3045;
		if(!m_MFC3045.VersionGet(l_RdString, l_FWVersionValue, eWhoVersion::kMUFirmVer))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Failed to get MU Firmware version", MTL__LOCATION__);
		rInstrInfo.Description = l_RdString;
		if(!m_MFC3045.VersionGet(l_RdString, l_FWVersionValue, eWhoVersion::kMUEepromVer))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Failed to get MU-EEPROM Firmware version", MTL__LOCATION__);
		rInstrInfo.Description += ", MU-EEPROM Version " + l_RdString;
		if(!m_MFC3045.SerialNumberGet(l_RdString, eWhoSerialNum::kMUSerialNum))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Failed to get MU serial number", MTL__LOCATION__);
		rInstrInfo.Description += ", S/N " + l_RdString;
		rInstrInfo.SerialNumber = l_RdString;

		// Get main unit calibration date
		std::string l_CalDateUTCStr;
		std::time_t l_CalDateUTC;
		if(!m_MFC3045.CalibrationDateMUGet(l_CalDateUTCStr, l_CalDateUTC)) // get UTC date in string and time_t
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Failed to get MU calibration date", MTL__LOCATION__);
		if(!l_UtcToStrDateISO8601(l_CalDateUTC, rInstrInfo.CalibrationDate))	// convert time_t UTC to YYYY-MM-DDTHH:MM:SSZ ISO8601 UTC format
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Cannot parse MU date", MTL__LOCATION__);
		if(!m_MFC3045.AdvancedCommandModeSet(eAdvancedMode::kNone))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Failed to change MU Advanced mode", MTL__LOCATION__);

		l_SelStatus.SelSTReg = BitSelSTALL; // read all status registers
		if(!m_MFC3045.StatusGet(l_SelStatus, m_AllRegStatus))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Status registers read failed", MTL__LOCATION__);
		if (m_AllRegStatus.StReg1.StatusByte1.ECMD || m_AllRegStatus.StReg1.StatusByte1.ECOM) //check serial command and communication error
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Error on Status 1: 0b" + toBinaryString(m_AllRegStatus.StReg1.RawSTB1), MTL__LOCATION__);
	}
	catch (sMFCDeviceError & rE)
	{
		return rE;
	}
	return sMFCDeviceError(sMFCDeviceError::eLevel::kNoError, "Instrument Connection done.", MTL__LOCATION__);
}

// Scan Probe Array connected to main unit and get information
CMFC3045Device::sMFCDeviceError CMFC3045Device::ScanPA(CInstrumentCatalog::sProbeArrayInformation &	rPAInfo,
														CInstrumentCatalog::sInstrumentConfiguration & rInstrConfig, bool &rPaInfoChanged)
{
	try
	{
		uStatusRegBitSel l_SelStatus;
		std::string l_RdString;
		rPaInfoChanged = false;

		bool l_PAConnected(false);
		if(!l_CheckPAConnected(l_PAConnected))
		{//error on read PA connection state. Check if Instrument is still connected
			if(!IsInstrumentConnected()) //takes 80 to 100ms
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Instrument connection lost", MTL__LOCATION__);
			l_SelStatus.SelSTReg = BitSelST1;
			if(!m_MFC3045.StatusGet(l_SelStatus, m_AllRegStatus)) //read status 1
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Status registers read failed", MTL__LOCATION__);
			if (m_AllRegStatus.StReg1.StatusByte1.ECMD || m_AllRegStatus.StReg1.StatusByte1.ECOM) //check serial command and communication error
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Error on Status 1: 0b" + toBinaryString(m_AllRegStatus.StReg1.RawSTB1), MTL__LOCATION__);
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Status registers read failed", MTL__LOCATION__); //throw error anyway
		}
		if(!l_PAConnected)
			throw false; // PA is not connected. It's not an error but don't go further.

		// Get probe array information
		if(!m_MFC3045.SerialNumberGet(l_RdString, eWhoSerialNum::kPASerialNum))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Failed to get PA serial number", MTL__LOCATION__);
		if(l_RdString != m_PaSerialNum) //Probe array is different than before
		{
			m_LastConfigured = CInstrumentCatalog::sInstrumentState();	// Clear last known configuration
			// read status register ST4
			l_SelStatus.SelSTReg = BitSelST4;
			if(!m_MFC3045.StatusGet(l_SelStatus, m_AllRegStatus))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Status registers read failed", MTL__LOCATION__);
			if (m_AllRegStatus.StReg4.StatusByte4.HWP || m_AllRegStatus.StReg4.StatusByte4.EPM) //check hardware and PA EEPROM error
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Error on Status 4: 0b" + toBinaryString(m_AllRegStatus.StReg4.RawSTB4), MTL__LOCATION__);

			m_PaSerialNum = l_RdString;
			m_NbProbesPA = 0;
			rPAInfo.SerialNumber = l_RdString;
			rPAInfo.Type = CInstrumentCatalog::eProbeArrayType::kMFC3048;

			// Read modulation (acquisition) period (PA need to be connected)
			U16 l_MDP_Ms(0);
			if(!m_MFC3045.ModulationPeriodGet(l_MDP_Ms))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Failed to get acquisition period", MTL__LOCATION__);
			rInstrConfig.AcquisitionPeriod = static_cast<F64>(l_MDP_Ms) / 1000.0; // convert ms to second

			U32 l_RdFreq(0);
			// Get PA Central frequency dHz
			if(!m_MFC3045.PaCentralFreqGet(l_RdFreq))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Failed to get PA central Freq", MTL__LOCATION__);
			rPAInfo.FreqCentral = l_ConvertdHzToMHz(l_RdFreq);
			// Get PA lowset frequency dHz
			if(!m_MFC3045.PaLowestFreqGet(l_RdFreq))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Failed to get PA min Freq", MTL__LOCATION__);
			rPAInfo.FreqMin = l_ConvertdHzToMHz(l_RdFreq);
			// Get PA highest frequency dHz
			if(!m_MFC3045.PaHighestFreqGet(l_RdFreq))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Failed to get PA max Freq", MTL__LOCATION__);
			rPAInfo.FreqMax = l_ConvertdHzToMHz(l_RdFreq);
			// Get PA number of probes
			U8 l_NbProbe(0);
			if(!m_MFC3045.NumberProbesGet(l_NbProbe))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Failed to get PA number of probes", MTL__LOCATION__);
			if(0==l_NbProbe)
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Failed to get PA number of probes", MTL__LOCATION__);
			m_NbProbesPA = l_NbProbe;
			rPAInfo.ProbeList.clear();
			for (U8 l_it =1; l_it <= l_NbProbe; l_it++)
			{
				CInstrumentCatalog::sProbeInformation l_PI;
				l_PI.ProbeNumber	= l_it;
				l_PI.FreqMin		= rPAInfo.FreqMin;
				l_PI.FreqMax		= rPAInfo.FreqMax;
				rPAInfo.ProbeList.push_back(l_PI);
			}

			rPAInfo.Description = "MFC3048 : Continuous Wave Camera";
			rPAInfo.GyroRatio = static_cast <F64>(GYRORATIO_MFC3048_HZT)/1.0e6; // From Hz/T to MHz/T

			// Get PA normalization date
			std::string l_NormDateUTCStr;
			sMFCDeviceError l_RetError = GetNormalizationDate(l_NormDateUTCStr);
			if(l_RetError.HasError())
			{
				throw l_RetError;
			}
			rPAInfo.NormalizationDate = l_NormDateUTCStr;

			// Ok to send new PA information
			rPaInfoChanged = true;
			l_SelStatus.SelSTReg = BitSelST1 | BitSelST2 | BitSelST4 | BitSelST5; // read status registers to clear bits
			if(!m_MFC3045.StatusGet(l_SelStatus, m_AllRegStatus))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Status registers read failed", MTL__LOCATION__);
			if (m_AllRegStatus.StReg1.StatusByte1.ECMD || m_AllRegStatus.StReg1.StatusByte1.ECOM) //check serial command and communication error
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Error on Status 1: 0b" + toBinaryString(m_AllRegStatus.StReg1.RawSTB1), MTL__LOCATION__);
		}
	}
	// There is a warning or error
	catch (sMFCDeviceError & rE)
	{
		m_PaSerialNum = "";
		// If we previously had a PA
		if (!rPAInfo.ProbeList.empty())
		{
			// Update PA Info (empty)
			rPAInfo = CInstrumentCatalog::sProbeArrayInformation();
			rInstrConfig = CInstrumentCatalog::sInstrumentConfiguration();
			m_LastConfigured = CInstrumentCatalog::sInstrumentState();	// Clear last known configuration
			rPaInfoChanged = true;
		}
		return rE; // Return received error
	}
	// There is no error but no camera is available
	catch (bool & rUsableCamera)
	{
		MTL_Unused(rUsableCamera)
		m_PaSerialNum = "";
		// If we previously had a PA
		if (!rPAInfo.ProbeList.empty())
		{
			// Update PA Info (empty)
			rPAInfo = CInstrumentCatalog::sProbeArrayInformation();
			rInstrConfig = CInstrumentCatalog::sInstrumentConfiguration();
			m_LastConfigured = CInstrumentCatalog::sInstrumentState();	// Clear last known configuration
			rPaInfoChanged = true;
		}
	}
	return sMFCDeviceError(sMFCDeviceError::eLevel::kNoError, "Scan PA done.", MTL__LOCATION__);
}

CMFC3045Device::sMFCDeviceError CMFC3045Device::Search(const CInstrumentCatalog::sInstrumentState & rInstrState)
{
	uStatusRegBitSel l_SelStatus;
	sMFCDeviceError l_ReturnError;

	try
	{
		l_SelStatus.SelSTReg = BitSelST1;
		if(!m_MFC3045.StatusGet(l_SelStatus, m_AllRegStatus)) //read Status register 1 to check any previous error before to start measurement operation
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Status registers read failed", MTL__LOCATION__);
		if (m_AllRegStatus.StReg1.StatusByte1.ECMD || m_AllRegStatus.StReg1.StatusByte1.ECOM) //check serial command and communication error
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Error on Status 1: 0b" + toBinaryString(m_AllRegStatus.StReg1.RawSTB1), MTL__LOCATION__);
		if (m_AllRegStatus.StReg1.StatusByte1.EMOD) //check if modulation error occurs
		{
			l_SelStatus.SelSTReg = BitSelST2; // Read details of modulation error in Status 2 register.
			if(!m_MFC3045.StatusGet(l_SelStatus, m_AllRegStatus))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Status registers read failed", MTL__LOCATION__);
			if (m_AllRegStatus.StReg2.StatusByte2.PANC) //check if PA connected
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "No PA available", MTL__LOCATION__);
		}

		sMFCDeviceError l_Error;
		if(!l_ConfigureForOperation(rInstrState,l_Error)) // configure Main unit modulation parameter for measurement if required
			throw l_Error;

		// Start Search NMR signal over full PA range
		bool l_IsContinuous = (CInstrumentCatalog::eRepeatMode::kContinuous == rInstrState.RepeatMode);
		do
		{
			U8 l_SearchProg =0;
			// ENABLE SMA
			if (!m_MFC3045.SearchStartWithSMA()) // Start search NMR with SMA enable on PA not connected and Data ready.
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not start search", MTL__LOCATION__);
			// WARNING: From now, no Read commands should be sent to main unit until SMA is disable.
			// We use this search method with SMA waiting on data ready message instead of a status register polling, because the Main unit firmware seems
			// to have occasionnaly a problem on parsing a command during measurement process (Command error detects "TST" instead of STx)
			MTL_SleepMs(200); // wait measurement to start in MFC3045

			// Note: no read status allowed while waiting for SMA (send automatic message from main unit) message
			// Wait continuously while
			//  - we are not requested to stop
			//  - we don't have our measurement results
			// If system indicates Searching, report search progress
			uSMAByte l_SmaMsgRd;
			l_SmaMsgRd.RawSMAB =0;
			auto l_EndTime = std::chrono::steady_clock::now() + std::chrono::milliseconds(m_MaxMeasTimeOutMs);

			for (;
				 !m_rMFCController.l_IsAbortRequested() && !l_SmaMsgRd.SMAByte.DRST1 && !l_SmaMsgRd.SMAByte.PANST2;
				 )
			{
				if (m_MaxMeasTimeOutMs > 0)
				{
					if(std::chrono::steady_clock::now() > l_EndTime) //An error occured during measurement. MFC3045 seems to be blocked or serial communication is not working anymore
						throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Search timeout", MTL__LOCATION__);
				}
				// MFC3045 doesn't have search progress. Make one here.
				l_SearchProg = (l_SearchProg + 1) % 99;
				m_rMFCController.l_SendSearchProgress(l_SearchProg); //notify controller for search progress

				MTL_SleepMs(50);
				// read for received SMA (Data Ready or PA disconnected) message if any (with a timeout)
				m_MFC3045.WaitSmaMeasureDataReady(SEARCH_POLLING_PERIOD_MS, l_SmaMsgRd);
			}
			// DISABLE SMA
			if(!m_MFC3045.DisableSMA())
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "SMA disable failed", MTL__LOCATION__);
			// Now we can send any Read commands to main unit!
			// Read status register to confirm Data ready
			l_SelStatus.SelSTReg = BitSelST1 | BitSelST3;
			if(!m_MFC3045.StatusGet(l_SelStatus, m_AllRegStatus))	//SetST1 | SetST3
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Status registers read failed", MTL__LOCATION__);
			if (m_AllRegStatus.StReg1.StatusByte1.ECMD || m_AllRegStatus.StReg1.StatusByte1.ECOM) //check serial command and communication error
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Error on Status 1: 0b" + toBinaryString(m_AllRegStatus.StReg1.RawSTB1), MTL__LOCATION__);
			if (m_AllRegStatus.StReg1.StatusByte1.EMOD) //check modulation error
			{
				l_SelStatus.SelSTReg = BitSelST2; // Read details of modulation error in Status 2 register.
				if(!m_MFC3045.StatusGet(l_SelStatus, m_AllRegStatus))
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Status registers read failed", MTL__LOCATION__);
				if (m_AllRegStatus.StReg2.StatusByte2.PANC) //check if PA is still connected
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "No PA available", MTL__LOCATION__);
			}
			if (!m_rMFCController.l_IsAbortRequested())
			{
				// If we have a result to read, read search result and return it
				if (m_AllRegStatus.StReg3.StatusByte3.DAA) // ||(m_AllRegStatus.StReg1.StatusByte1.DRDY))
				{
					std::vector<U32> l_ReadMesDatadHz;
					if (!m_MFC3045.DataMagneticFieldGet(l_ReadMesDatadHz))
						throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not fetch search result", MTL__LOCATION__);
					if(l_ReadMesDatadHz.size() < m_NbProbesPA)
						throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Missing measurment result", MTL__LOCATION__);

					MTL_Assert(!l_ReadMesDatadHz.empty());
					if (l_ReadMesDatadHz.empty())
						throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Invalid search result", MTL__LOCATION__);

					l_SearchProg = 99;
					m_rMFCController.l_SendSearchProgress(l_SearchProg);//notify controller for search progress complete

					// Compute search averaging result
					F64 l_AverageMHz, l_StdDevPpm;
					l_ComputeAveraging(l_ReadMesDatadHz, l_AverageMHz, l_StdDevPpm); //do average on all valid search probes result

					// Send result
					CInstrumentCatalog::sSearchResult l_SearchResult;
					l_SearchResult.Freq = l_AverageMHz; //in MHz
					l_SearchResult.StdDev = static_cast<F32>(l_StdDevPpm); // in ppm
					m_rMFCController.l_SendSearchResult(l_SearchResult);
				}
				else
				{
					//No error but no measurement result available from instrument. Instrument wasn't able to detect any NMR signal. Return NaN
					l_SearchProg = 99;
					m_rMFCController.l_SendSearchProgress(l_SearchProg);//notify controller for search progress complete
					// Send NaN results
					CInstrumentCatalog::sSearchResult l_SearchResult;
					l_SearchResult.Freq = DPqNaN;
					l_SearchResult.StdDev = static_cast<F32>(DPqNaN);
					m_rMFCController.l_SendSearchResult(l_SearchResult);
					LOG("No search result returned by instrument", CInstrumentCatalog::eLogLevel::kWarning, MTL__LOCATION__);
				}
			}
			else
				m_MFC3045.BreakMeasSet(); // stop any acquisition in main unit
		} while (!m_rMFCController.l_IsAbortRequested() && l_IsContinuous);

		l_SelStatus.SelSTReg = BitSelST1 | BitSelST4; // read only status register 1 and 4 (ST4 to clear remote  button bit)
		if(!m_MFC3045.StatusGet(l_SelStatus, m_AllRegStatus))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Status registers read failed", MTL__LOCATION__);
		if (m_AllRegStatus.StReg1.StatusByte1.ECMD || m_AllRegStatus.StReg1.StatusByte1.ECOM) //check serial command and communication error
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Error on Status 1: 0b" + toBinaryString(m_AllRegStatus.StReg1.RawSTB1), MTL__LOCATION__);
		if (m_AllRegStatus.StReg1.StatusByte1.EMOD) //check modulation error
		{
			l_SelStatus.SelSTReg = BitSelST2; // Read details of modulation error in Status 2 register (and clear bits from ST2).
			if(!m_MFC3045.StatusGet(l_SelStatus, m_AllRegStatus))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Status registers read failed", MTL__LOCATION__);
			if (m_AllRegStatus.StReg2.StatusByte2.PANC) //check if PA is still connected
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "No PA available", MTL__LOCATION__);
		}
		l_ReturnError = sMFCDeviceError(sMFCDeviceError::eLevel::kNoError, "Search operation done.", MTL__LOCATION__);
	}
	catch (sMFCDeviceError & rE)
	{
		U32 l_ReadTimeoutMs(0);
		// Signal error
		l_ReturnError = rE;
		m_MFC3045.GetCurrentTimeoutRdDataMs(l_ReadTimeoutMs);
		m_MFC3045.SetTimeoutRdNewDataMs(1000);// set shorter timeout on serial read communication
		m_MFC3045.BreakMeasSet(); // stop any acquisition in main unit
		if(m_MFC3045.DisableSMA()) // DisableSMA() contains a readback of SMA register. If DisableSMA failed don't try to read Status Register
		{
			l_SelStatus.SelSTReg = BitSelSTALL; // Read all status registers to clear them
			m_MFC3045.StatusGet(l_SelStatus, m_AllRegStatus);
		}
		m_MFC3045.SetTimeoutRdNewDataMs(l_ReadTimeoutMs); // set back previous serial read timeout
	}
	return l_ReturnError;
}

CMFC3045Device::sMFCDeviceError CMFC3045Device::Measure(const CInstrumentCatalog::sInstrumentState & rInstrState)
{
	bool l_MeasAllPA = true;
	uSMAByte l_SmaMsgRd;
	uStatusRegBitSel l_SelStatus;
	sMFCDeviceError l_ReturnError;

	try
	{
		// read some status registers
		l_SelStatus.SelSTReg = BitSelST1;
		if(!m_MFC3045.StatusGet(l_SelStatus, m_AllRegStatus)) //read Status register 1 to check any previous error before to start measurement operation
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Status registers read failed", MTL__LOCATION__);
		if (m_AllRegStatus.StReg1.StatusByte1.ECMD || m_AllRegStatus.StReg1.StatusByte1.ECOM) //check serial command and communication error
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Error on Status 1: 0b" + toBinaryString(m_AllRegStatus.StReg1.RawSTB1), MTL__LOCATION__);
		if (m_AllRegStatus.StReg1.StatusByte1.EMOD) //check if modulation error occurs
		{
			l_SelStatus.SelSTReg = BitSelST2; // Read details of modulation error in Status 2 register.
			if(!m_MFC3045.StatusGet(l_SelStatus, m_AllRegStatus))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Status registers read failed", MTL__LOCATION__);
			if (m_AllRegStatus.StReg2.StatusByte2.PANC) //check if PA connected
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "No PA available", MTL__LOCATION__);
		}
		// Main unit reconfiguration check
		sMFCDeviceError l_Error;
		if(!l_ConfigureForOperation(rInstrState,l_Error)) //configure Main unit modulation parameter for measurement if required
			throw l_Error;

		// Start to measure NMR
		bool l_IsContinuous = (CInstrumentCatalog::eRepeatMode::kContinuous == rInstrState.RepeatMode);
		do
		{
			U8 l_SearchProg =0;
			switch (rInstrState.ProbeSelection.SelectionMode)
			{
			case CInstrumentCatalog::eMeasureSelectionMode::kProbeArray:
			{
				// ENABLE SMA
				if (!m_MFC3045.MeasStartWithSMA()) // Start measurement with SMA enabled with only "PA not connected" and "Data ready".
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not start measurment", MTL__LOCATION__);
			}
				break;

			case CInstrumentCatalog::eMeasureSelectionMode::kSingleProbe:
			{
				l_MeasAllPA = false;
				// ENABLE SMA and start measurement
				if (!m_MFC3045.MeasStartWithSMA(rInstrState.ProbeSelection.ProbeNumber)) // Start measurement with SMA enabled with only "PA not connected" and "Data ready".
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not start measurment", MTL__LOCATION__);
			}
				break;
			default:
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Probe selection mode not supported", MTL__LOCATION__);
				break;
			}
			// WARNING: From now, no read command should be sent until SMA is disabled.
			// We use this measurement method with SMA waiting on data ready message instead of a status register polling, because the Main unit firmware seems
			// to have occasionnaly a problem on parsing a command during measurement process (eg. Command error detects "TST" instead of STx)

			MTL_SleepMs(200); // wait measurement to start in MFC3045
			// Note: no read status allowed while waiting for SMA (send automatic message from main unit)
			// Wait continuously while
			//  - we are not requested to stop
			//  - we don't have our measurement results
			// If system indicates Searching, report search progress
			l_SmaMsgRd.RawSMAB =0;
			auto l_EndTime = std::chrono::steady_clock::now() + std::chrono::milliseconds(m_MaxMeasTimeOutMs);
			for (;
				 !m_rMFCController.l_IsAbortRequested() && !l_SmaMsgRd.SMAByte.DRST1 && !l_SmaMsgRd.SMAByte.PANST2;
				 )
			{
				if (m_MaxMeasTimeOutMs > 0)
				{
					if(std::chrono::steady_clock::now() > l_EndTime)
					{//An error occured during measurement. MFC3045 seems to be stopped or serial communication is not working anymore
						throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Measurement timeout", MTL__LOCATION__);
					}
				}
				// MFC3045 doesn't have search progress. Make one here.
				l_SearchProg = (l_SearchProg + 2) % 98; // never go to 100
				m_rMFCController.l_SendSearchProgress(l_SearchProg); // notify controller for search progress
				MTL_SleepMs(50);
				// wait and read for received SMA (Data Ready or PA disconnected) message if any (with a timeout)
				m_MFC3045.WaitSmaMeasureDataReady(MEAS_POLLING_PERIOD_MS, l_SmaMsgRd);
			}

			// DISABLE SMA
			if(!m_MFC3045.DisableSMA())
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "SMA disable failed", MTL__LOCATION__);
			// Now, we can send any read command to main unit!
			// read status register to confirm Data ready
			l_SelStatus.SelSTReg = BitSelST1 | BitSelST3;
			if(!m_MFC3045.StatusGet(l_SelStatus, m_AllRegStatus))	//SetST1 | SetST3
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Status registers read failed", MTL__LOCATION__);
			if (m_AllRegStatus.StReg1.StatusByte1.ECMD || m_AllRegStatus.StReg1.StatusByte1.ECOM) //check serial command and communication error
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Error on Status 1: 0b" + toBinaryString(m_AllRegStatus.StReg1.RawSTB1), MTL__LOCATION__);
			if (m_AllRegStatus.StReg1.StatusByte1.EMOD) //check modulation error
			{
				l_SelStatus.SelSTReg = BitSelST2; // Read details of modulation error in Status 2 register.
				if(!m_MFC3045.StatusGet(l_SelStatus, m_AllRegStatus))
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Status registers read failed", MTL__LOCATION__);
				if (m_AllRegStatus.StReg2.StatusByte2.PANC) //check if PA is still connected
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "No PA available", MTL__LOCATION__);
			}
			if(!m_rMFCController.l_IsAbortRequested())
			{
					l_SearchProg = 99;
					m_rMFCController.l_SendSearchProgress(l_SearchProg);//notify controller for search progress
					CInstrumentCatalog::sMeasurementResults l_MeasResult;
					l_MeasResult.Timestamp = static_cast<U64>(std::chrono::duration_cast<std::chrono::milliseconds>(std::chrono::system_clock::now().time_since_epoch()).count()); //ms since epoch
					if(l_MeasAllPA) // all probes measurement
					{
						std::vector<U32> l_ReadMesDatadHz;
						std::vector<U32> l_ReadStdDevDatadHz;
						std::vector<U32> l_ReadValidCycleNb;
						// If we have a result to read, read measurement results from instrument and return it
						if (m_AllRegStatus.StReg3.StatusByte3.DAA)  // ||(m_AllRegStatus.StReg1.StatusByte1.DRDY))
						{
							if (!m_MFC3045.DataMagneticFieldGet(l_ReadMesDatadHz))
								throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not fetch measurment result", MTL__LOCATION__);
							if(l_ReadMesDatadHz.size() < m_NbProbesPA)
								throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Missing measurment result", MTL__LOCATION__);

							if (!m_MFC3045.DataStdDeviationGet(l_ReadStdDevDatadHz))
								throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not fetch std dev result", MTL__LOCATION__);
							if(l_ReadStdDevDatadHz.size() < m_NbProbesPA)
								throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Missing std dev result", MTL__LOCATION__);

							if (!m_MFC3045.DataNumberValidCycleGet(l_ReadValidCycleNb))
								throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not fetch nb valid cycle result", MTL__LOCATION__);
							if(l_ReadValidCycleNb.size() < m_NbProbesPA)
								throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Missing Valid cycle result", MTL__LOCATION__);

							// Prepare PA measurement result
							for (U8 l_it =0; l_it < m_NbProbesPA; l_it++)
							{
								CInstrumentCatalog::sMeasurement l_Meas;
								l_Meas.ProbeNumber = l_it+1;
								l_Meas.NbValidMeasurements = l_ReadValidCycleNb.at(l_it);
								if(l_ReadValidCycleNb.at(l_it) > 0)
								{
									l_Meas.Freq = l_ConvertdHzToMHz(l_ReadMesDatadHz.at(l_it));
									l_Meas.StdDev = l_ConvertStdDevdHzToPpm(l_ReadStdDevDatadHz.at(l_it), l_ReadMesDatadHz.at(l_it));
								}
								else
								{// nb valid cycle =0 on this probe measurement, fill with NaN
									l_Meas.Freq = DPqNaN;
									l_Meas.StdDev = static_cast<F32>(DPqNaN);
								}
								l_MeasResult.MeasurementList.push_back(l_Meas);
							}
						}
						else
						{ //Intrument doesn't have measurement to return. It wasn't able to detect any NMR signal. Fill it with NaN
							for (U8 l_it =0; l_it < m_NbProbesPA; l_it++)
							{
								CInstrumentCatalog::sMeasurement l_Meas;
								l_Meas.ProbeNumber = l_it+1;
								l_Meas.NbValidMeasurements = 0;
								l_Meas.Freq = DPqNaN;
								l_Meas.StdDev = static_cast<F32>(DPqNaN);
								l_MeasResult.MeasurementList.push_back(l_Meas);
							}
							LOG("No measurement result returned by instrument", CInstrumentCatalog::eLogLevel::kWarning, MTL__LOCATION__);
						}
					}
					else // Single probe measurement
					{
						U32 l_ReadMesDatadHz;
						U32 l_ReadStdDevDatadHz;
						U32 l_ReadValidCycleNb;
						// If we have a result to read, read measurement result from instrument
						if (m_AllRegStatus.StReg3.StatusByte3.DAA)  // ||(m_AllRegStatus.StReg1.StatusByte1.DRDY))
						{
							if (!m_MFC3045.DataMagneticFieldGet(l_ReadMesDatadHz, rInstrState.ProbeSelection.ProbeNumber))
								throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not fetch measurement result", MTL__LOCATION__);
							if (!m_MFC3045.DataStdDeviationGet(l_ReadStdDevDatadHz, rInstrState.ProbeSelection.ProbeNumber))
								throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not fetch std dev result", MTL__LOCATION__);
							if (!m_MFC3045.DataNumberValidCycleGet(l_ReadValidCycleNb, rInstrState.ProbeSelection.ProbeNumber))
								throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not fetch valid cycle result", MTL__LOCATION__);
							// Prepare single probe result
							CInstrumentCatalog::sMeasurement l_Meas;
							l_Meas.ProbeNumber = rInstrState.ProbeSelection.ProbeNumber;
							l_Meas.NbValidMeasurements = l_ReadValidCycleNb;
							if(l_ReadValidCycleNb > 0)
							{
								l_Meas.Freq = l_ConvertdHzToMHz(l_ReadMesDatadHz);
								l_Meas.StdDev = l_ConvertStdDevdHzToPpm(l_ReadStdDevDatadHz, l_ReadMesDatadHz);
							}
							else
							{ // nb valid cycle =0 on this probe measurement, fill with NaN
								l_Meas.Freq = DPqNaN;
								l_Meas.StdDev = static_cast<F32>(DPqNaN);
							}
							l_MeasResult.MeasurementList.push_back(l_Meas);
						}
						else
						{
							//Intrument doesn't have measurement to return. It wasn't able to detect any NMR signal. Fill it with NaN
							CInstrumentCatalog::sMeasurement l_Meas;
							l_Meas.ProbeNumber = rInstrState.ProbeSelection.ProbeNumber;
							l_Meas.NbValidMeasurements = 0;
							l_Meas.Freq = DPqNaN;
							l_Meas.StdDev = static_cast<F32>(DPqNaN);
							l_MeasResult.MeasurementList.push_back(l_Meas);
							LOG("No measurement result returned by instrument", CInstrumentCatalog::eLogLevel::kWarning, MTL__LOCATION__);
						}
					}
					//Send results
					m_rMFCController.l_SendMeasResult(l_MeasResult);
			}
			else
				m_MFC3045.BreakMeasSet(); // send Stop any acquisition to main unit
		} while (!m_rMFCController.l_IsAbortRequested() && l_IsContinuous);

		l_SelStatus.SelSTReg = BitSelST1 | BitSelST4; // read only status register 1 and 4 (ST4 to clear remote button bit)
		if(!m_MFC3045.StatusGet(l_SelStatus, m_AllRegStatus))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Status registers read failed", MTL__LOCATION__);
		if (m_AllRegStatus.StReg1.StatusByte1.ECMD || m_AllRegStatus.StReg1.StatusByte1.ECOM) //check serial command and communication error
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Error on Status 1: 0b" + toBinaryString(m_AllRegStatus.StReg1.RawSTB1), MTL__LOCATION__);
		if (m_AllRegStatus.StReg1.StatusByte1.EMOD) //check modulation error
		{
			l_SelStatus.SelSTReg = BitSelST2; // Read details of modulation error in Status 2 register (and clear bits from ST2).
			if(!m_MFC3045.StatusGet(l_SelStatus, m_AllRegStatus))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Status registers read failed", MTL__LOCATION__);
			if (m_AllRegStatus.StReg2.StatusByte2.PANC) //check if PA is still connected
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "No PA available", MTL__LOCATION__);
		}
		l_ReturnError = sMFCDeviceError(sMFCDeviceError::eLevel::kNoError, "Measurement operation done.", MTL__LOCATION__);
	}
	catch (sMFCDeviceError & rE)
	{
		U32 l_ReadTimeoutMs(0);
		// Signal error
		l_ReturnError = rE;

		m_MFC3045.GetCurrentTimeoutRdDataMs(l_ReadTimeoutMs);
		m_MFC3045.SetTimeoutRdNewDataMs(1000);// set shorter timeout on serial read
		m_MFC3045.BreakMeasSet(); // stop any acquisition in main unit
		if(m_MFC3045.DisableSMA()) // DisableSMA() contains a readback of SMA register. If DisableSMA failed don't try to read Status Register
		{
			l_SelStatus.SelSTReg = BitSelSTALL;
			m_MFC3045.StatusGet(l_SelStatus, m_AllRegStatus);// Read all status registers to clear them
		}
		m_MFC3045.SetTimeoutRdNewDataMs(l_ReadTimeoutMs); // set back previous serial read timeout
	}
	return l_ReturnError;
}

CMFC3045Device::sMFCDeviceError CMFC3045Device::NewNormalizationRequest(	const CInstrumentCatalog::sNormalization & rNewNormalizationRequest,
																			CInstrumentCatalog::sNormalization & rReturnedNormalization)
{
#if defined(MFC_NORMALIZATION_ENABLED)
	try
	{
		CMFC3045Device::sMFCDeviceError l_Error;
		l_Error = CMFCProLicInterface::MFC3045NewNormalizationRequest(m_NbProbesPA, m_MFC3045, rNewNormalizationRequest, rReturnedNormalization);
		if(l_Error.HasError())
			throw l_Error;
	}
	catch (sMFCDeviceError & rE)
	{
		m_MFC3045.AdvancedCommandModeSet(eAdvancedMode::kNone);
		return rE;
	}
	return sMFCDeviceError(sMFCDeviceError::eLevel::kNoError, "New Normalization request done.", MTL__LOCATION__);
	#else
		MTL_Unused(rNewNormalizationRequest);
		MTL_Unused(rReturnedNormalization);
		return sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Normalization request not supported.", MTL__LOCATION__);
	#endif //defined(MFC_NORMALIZATION_ENABLED)
}

CMFC3045Device::sMFCDeviceError CMFC3045Device::GetNormalizationDate(CInstrumentCatalog::tISO8601DateTime & rNormDate)
{
	try
	{
		uStatusRegBitSel l_SelStatus;
		std::time_t l_NormDateUTC;
		std::string l_NormDateUTCStr;

		// Read status register ST1 & ST2
		l_SelStatus.SelSTReg = BitSelST1 | BitSelST2;
		if(!m_MFC3045.StatusGet(l_SelStatus, m_AllRegStatus))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Status registers read failed", MTL__LOCATION__);
		if (m_AllRegStatus.StReg1.StatusByte1.ECOM) //check serial communication error
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Error on Status 1: 0b" + toBinaryString(m_AllRegStatus.StReg1.RawSTB1), MTL__LOCATION__);
		if (m_AllRegStatus.StReg2.StatusByte2.PANC) //check if PA connected
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Error on Status 2: 0b" + toBinaryString(m_AllRegStatus.StReg2.RawSTB2), MTL__LOCATION__);
		if(!m_MFC3045.AdvancedCommandModeSet(eAdvancedMode::k1))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Failed to change MU Advanced mode", MTL__LOCATION__);

		// Get PA normalization date
		if(!m_MFC3045.CalibrationDatePAGet(l_NormDateUTCStr, l_NormDateUTC)) // Note: this methode request high advance command level mode
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Failed to get normalization date", MTL__LOCATION__);
		if(!l_UtcToStrDateISO8601(l_NormDateUTC, rNormDate)) // convert time_t UTC to YYYY-MM-DDTHH:MM:SSZ ISO8601 format
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Failed to parse normalization date", MTL__LOCATION__);

		if(!m_MFC3045.AdvancedCommandModeSet(eAdvancedMode::kNone))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Failed to change MU Advanced mode", MTL__LOCATION__);

		l_SelStatus.SelSTReg = BitSelST1;
		if(!m_MFC3045.StatusGet(l_SelStatus, m_AllRegStatus))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Status registers read failed", MTL__LOCATION__);
		if (m_AllRegStatus.StReg1.StatusByte1.ECMD || m_AllRegStatus.StReg1.StatusByte1.ECOM) //check serial command and communication error
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Error on Status 1: 0b" + toBinaryString(m_AllRegStatus.StReg1.RawSTB1), MTL__LOCATION__);
	}
	catch (sMFCDeviceError & rE)
	{
		return rE;
	}
	return sMFCDeviceError(sMFCDeviceError::eLevel::kNoError, "Get normalization date done.", MTL__LOCATION__);
}

CMFC3045Device::sMFCDeviceError CMFC3045Device::SetRBBusyLedState(const CInstrumentCatalog::eRemoteBoxBusyLedState & rRBBusyLedState)
{
	try
	{
		eRemoteBusyLed l_MFC3045RemoteLedState;
		switch (rRBBusyLedState)
		{
		case CInstrumentCatalog::eRemoteBoxBusyLedState::kRemoteBusyLedOff:
			l_MFC3045RemoteLedState = eRemoteBusyLed::kRemoteLedOff;
			break;
		case CInstrumentCatalog::eRemoteBoxBusyLedState::kRemoteBusyLedMeasuring:
			l_MFC3045RemoteLedState = eRemoteBusyLed::kRemoteLedMeasuring;
			break;
		case CInstrumentCatalog::eRemoteBoxBusyLedState::kRemoteBusyLedOffEndMeas:
			l_MFC3045RemoteLedState = eRemoteBusyLed::kRemoteLedOffEndMeas;
			break;
		case CInstrumentCatalog::eRemoteBoxBusyLedState::kRemoteBusyLedOn:
			l_MFC3045RemoteLedState = eRemoteBusyLed::kRemoteLedOn;
			break;
		case CInstrumentCatalog::eRemoteBoxBusyLedState::kRemoteBusyLedBlinkSlow:
			l_MFC3045RemoteLedState = eRemoteBusyLed::kRemoteLedBlinkSlow;
			break;
		case CInstrumentCatalog::eRemoteBoxBusyLedState::kRemoteBusyLedBlinkFast:
			l_MFC3045RemoteLedState = eRemoteBusyLed::kRemoteLedBlinkFast;
			break;
		default:
			MTL_Assert(false);
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kFatal, "Unhandled busy led mode", MTL__LOCATION__);
			break;
		}
		uStatusRegBitSel l_SelStatus;
		l_SelStatus.SelSTReg = BitSelST1;
		if(!m_MFC3045.StatusGet(l_SelStatus, m_AllRegStatus)) //read to clear any previous bit error in ST1
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Status registers read failed", MTL__LOCATION__);
		if (!m_MFC3045.RemoteBusyLedSet(l_MFC3045RemoteLedState))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Could not set remote box busy led state", MTL__LOCATION__);
		l_SelStatus.SelSTReg = BitSelST1;
		if(!m_MFC3045.StatusGet(l_SelStatus, m_AllRegStatus)) //read ST1 to check errors
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Status registers read failed", MTL__LOCATION__);
		if (m_AllRegStatus.StReg1.StatusByte1.ECMD || m_AllRegStatus.StReg1.StatusByte1.ECOM) //check serial command and communication error
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Error on Status 1: 0b" + toBinaryString(m_AllRegStatus.StReg1.RawSTB1), MTL__LOCATION__);
	}
	catch (sMFCDeviceError & rE)
	{
		return rE;
	}
	return sMFCDeviceError(sMFCDeviceError::eLevel::kNoError, "Set remote box busy led state done.", MTL__LOCATION__);
}

CMFC3045Device::sMFCDeviceError CMFC3045Device::RemoteBoxButtonStatus(bool & rRemoteStartPressed)
{
	try
	{
		uStatusRegBitSel l_SelStatus;
		l_SelStatus.SelSTReg = BitSelST4;
		if(!m_MFC3045.StatusGet(l_SelStatus, m_AllRegStatus)) // read status 4
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Status registers read failed", MTL__LOCATION__);
		rRemoteStartPressed = false;
		if(m_AllRegStatus.StReg4.StatusByte4.RBC) // beware to not clear the bit in l_ScanPA by reading the status register 4 before here
		{
			if (!m_AllRegStatus.StReg4.StatusByte4.RBS) //check remote button State
				rRemoteStartPressed = true;
		}
	}
	catch (sMFCDeviceError & rE)
	{
		rRemoteStartPressed = false;
		return rE;
	}
	return sMFCDeviceError(sMFCDeviceError::eLevel::kNoError, "Check RemoteBox Status button done.", MTL__LOCATION__);
}

CMFC3045Device::sMFCDeviceError CMFC3045Device::ChangeSerialParmsOnConnectedInstr(const CInstrumentCatalog::sMFC3045SerialPortSettings & rNewSerialSetting)
{
	bool l_SystemReset(false);
	try
	{
		sSerialPortSettings l_NewVisaSerialSettings;
		U16 l_NewRSPSettings, l_ReadRSPRegValue;
		uStatusRegBitSel l_SelStatus;
		l_SelStatus.SelSTReg = BitSelST1 | BitSelST4 | BitSelST5;

		if (!IsInstrumentConnected())
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "No more instrument connected", MTL__LOCATION__);

		if(!m_MFC3045.StatusGet(l_SelStatus, m_AllRegStatus)) // read status to clear bits
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Status registers read failed", MTL__LOCATION__);
		if(!l_ParseSerialSettings(rNewSerialSetting, l_NewVisaSerialSettings))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Failed to parse serial settings", MTL__LOCATION__);

		if(!m_MFC3045.AdvancedCommandModeSet(eAdvancedMode::k2))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Failed to change MU Advanced mode", MTL__LOCATION__);
		if(!m_MFC3045.ParseParmForRSP(l_NewVisaSerialSettings,l_NewRSPSettings))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Failed to parse serial settings to RSP register type", MTL__LOCATION__);
		if(!m_MFC3045.SerialRSPSet(l_NewRSPSettings))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Failed to set RSP register", MTL__LOCATION__);
		// read back register RSP and compare
		if(!m_MFC3045.SerialRSPGet(l_ReadRSPRegValue))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Failed to read RSP register", MTL__LOCATION__);
		if(l_ReadRSPRegValue!= l_NewRSPSettings)
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Write and read back RSP register different", MTL__LOCATION__);

		if(!m_MFC3045.StatusGet(l_SelStatus, m_AllRegStatus))	//SetST1 | SetST4 | SetST5
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Status registers read failed", MTL__LOCATION__);
		if (m_AllRegStatus.StReg1.StatusByte1.ECMD) //check serial command error
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Error on Status 1: 0b" + toBinaryString(m_AllRegStatus.StReg1.RawSTB1), MTL__LOCATION__);
		if (m_AllRegStatus.StReg5.StatusByte5.EFR | m_AllRegStatus.StReg5.StatusByte5.EOR | m_AllRegStatus.StReg5.StatusByte5.EPY) //check serial communication error
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Error on Status 5: 0b" + toBinaryString(m_AllRegStatus.StReg5.RawSTB5), MTL__LOCATION__);

		//Save new RS parms in MU-EEPROM
		if(!m_MFC3045.MainUnitEepromWrite())
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Failed to save new Serial parms to Main unit EEPROM", MTL__LOCATION__);
		MTL_SleepMs(1000);

		// Read status register 1
		l_SelStatus.SelSTReg = BitSelST1 | BitSelST4;
		if(!m_MFC3045.StatusGet(l_SelStatus, m_AllRegStatus))	//SetST1 | SetST4
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Status registers read failed", MTL__LOCATION__);
		if (m_AllRegStatus.StReg1.StatusByte1.ECMD || m_AllRegStatus.StReg1.StatusByte1.ECOM) //check serial command and communication error
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Error on Status 1: 0b" + toBinaryString(m_AllRegStatus.StReg1.RawSTB1), MTL__LOCATION__);
		if (!m_AllRegStatus.StReg4.StatusByte4.MSW)
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Failed to save RS settings in main unit EEPROM", MTL__LOCATION__);

		//Reset Main Unit
		l_SystemReset = true;
		if(!m_MFC3045.SystemResetSet()) // Beware after a Main unit reset, registers take their default value (For example advanced mode and datablock mode)
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Failed to reset main unit instrument", MTL__LOCATION__);

		m_SerialParms = l_NewVisaSerialSettings;
		if(!m_MFC3045.DisconnectInstrument()) // Need to disconnect host Serial port before to reconnect it
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Failed to close VISA connection", MTL__LOCATION__);
		MTL_SleepMs(2000); // wait a bit to let the main unit to restart
	}
	catch (sMFCDeviceError & rE)
	{
		if(!l_SystemReset)
			m_MFC3045.AdvancedCommandModeSet(eAdvancedMode::kNone);
		return rE;
	}
	return sMFCDeviceError(sMFCDeviceError::eLevel::kNoError, "Change instrument serial settings done.", MTL__LOCATION__);
}

