//////////////////////////////////////////////////////////////////////////
/// \file		PT2026.h
/// \brief 		Interface definition for C++ API for Metrolab PT2026.
/// \date       2022
/// \copyright 	Metrolab Technology SA
/// \author 	Simon Tellier


#pragma once

// Standard includes
#include <ctime>
#include <queue>

// Personal includes
#include "VISAInstrument.h"
#include "PT2026Types.h"

// Version
#define PT2026_CPP_DRIVER_BCD_VERSION		0x0105

using namespace MTL::Instrument::PT2026Types;

namespace MTL {
	namespace Instrument {

	/// \brief PT2026 instrument class.
	class CPT2026Instrument : public CVISAInstrument
	{
	//----------------------------------------------------------------------//
	//							Private Attributes							//
	//----------------------------------------------------------------------//
	private:
		// General
		ViUInt32				m_Timeout_ms;
		std::queue<sError>		m_ErrorList;
		CVISABuffer				m_ReadBuffer;
	//----------------------------------------------------------------------//
	//								Utilities								//
	//----------------------------------------------------------------------//
	private:
		bool l_GetErrorList(const std::string & rContext);
		bool l_WriteAndRead(const std::string & rWriteStr, CVISABuffer & rReadBuffer, bool WaitForMAV = false);

	//----------------------------------------------------------------------//
	//							PT2026 Interface							//
	//----------------------------------------------------------------------//
	public:
		///
		/// \brief Constructors / destructors
		///
		
		//----------------------------------------------------------------------//
		// Constructors / destructors
		CPT2026Instrument(CVISAResourceManager & rRM, tResourceName Rsrc);
		virtual ~CPT2026Instrument();

		///
		/// @name ErrorList
		///
		/// List containing detailed information about the last operation performed.
		///@{
		
		//----------------------------------------------------------------------//
		// General Information
		const std::vector<sError> ErrorList();

		///@}
		
		/// @name Open / Close
		///
		/// Create a link to the instrument. The actual handle to the instrument is private to the class.
		///
		///@{
		
		//----------------------------------------------------------------------//
		// Open / Close
		bool Connect(U32 InitialTimeout_ms);
		void Disconnect();

		///@}
		
		/// @name Interface lock & unlock
		/// 
		/// Several communication ports are available on the instrument, including Ethernet that allows for remote 
		/// control of the instrument. Locking a port will enforce the interface right to use get / set accessors 
		/// to modify the behavior of the instrument whereas ports that don't have the lock are only allowed to read 
		/// the instrument parameters.
		/// 
		///@{
		
		//----------------------------------------------------------------------//
		// Interface lock
		bool LockInterface();
		bool UnlockInterface();

		///@}
		
		/// @name Channels handling
		///
		/// The PT2026 allow several probes to be connected to it using either a probe multiplexer or a field camera adapter. 
		///
		///@{
		
		//----------------------------------------------------------------------//
		// Channels
		bool ChannelsGetList(MTL::SCPI::tChannelList & rChanList);
		bool ChannelsGetList(CExtendedChannelList & rChanList);
		bool ChannelsGetState(MTL::SCPI::tChannelList & rSelChanList, MTL::SCPI::tChannelList & rActiveChan);
		bool ChannelsSetLimitedListState(bool State);
		bool ChannelsGetLimitedListState(bool & rState);
		bool ChannelsSelect(const MTL::SCPI::tChannelList & rChanList);
		bool ChannelsGetInfo(const MTL::SCPI::tChannelList & rChanList, std::vector<sChannelIformation> & rChanInfo);
		bool ChannelsGetLimits(const MTL::SCPI::tChannelList & rChanList, std::vector<F64> & rLowLimit_UNITS, std::vector<F64> & rHighLimit_UNITS);
		
		bool ChannelGetRawHall(sProbeRawHall & rRawHall); ///<Returns the raw 16-bit ADC measurements performed by the hall sensor located at the tip of the NMR probe. Notice that all NMR probes aren't necessarily equipped with such sensor. Magnetic field cameras aren't equipped with such sensor.
		bool ChannelGetHall(sProbeHall & rHall); ///<Returns in the current units the calibrated Hall measurements.

		bool ChannelGetRemoteLEDMode(eRemoteBusyLEDmode & rMode);
		bool ChannelSetRemoteLEDMode(eRemoteBusyLEDmode Mode);
		
		///@}

		//----------------------------------------------------------------------//
		// Parameters

		/// @name Search parameters
		///
		/// An NMR measurement always starts with a search operation consisting in finding the NMR resonance. 
		/// This search operation may take a very long time, especially with probe measuring high fields. 
		/// To reduce the time taken by the search operation, one may lessen the search range. Failing to set 
		/// the search range with respect to the actual magnetic field density that one tries to measure will 
		/// result in an infinite search operation. The search range must be set to fit the probe dynamic range. 
		///
		/// If present, the Hall probe can be enabled allowing the search range to be set automatically. 
		/// The search span set by the Hall subsystem will depend on the quality of the calibration. 
		///
		///@{
			
			// Search
		bool ParmSearchSet(const sSearch<uParm> & rSearch); ///<Configure the search operation, including frequency steps and the search bounds specified in the selected unit.
		bool ParmSearchGet(sSearch<uParm> & rSearch);
		bool ParmSearchGet(sSearch<sBoundedParm> & rSearch);
		
		bool ParmSearchSetManualValue(F64 ManualValue_UNITS); ///<Ask the instrument to perform a measurement at fixed frequency. Automatic search mode is left.
		bool ParmSearchSetLimits(F64 LowLimit_UNITS, F64 HighLimit_UNITS); ///<Set the search bounds in the selected unit.
		bool ParmSearchSetHallEnable(bool Enabled); ///<Allow the Hall probe to take control over the search range.
		bool ParmSearchGetHallEnable(bool & rEnabled);
		
		///@}
		/// @name Measure parameters
		///
		///  Accessors to configure the way the measurement is processed to compute the result. These accessors are very specific, it is recommended not to use them.   
		///@{
			
			// Measure
		bool ParmMeasureSet(const sMeasure<uParm> & rMeasure);
		bool ParmMeasureGet(sMeasure<uParm> & rMeasure);
		bool ParmMeasureGet(sMeasure<sBoundedParm> & rMeasure);
		
		///@}
		/// @name Averaging
		///
		/// An NMR measurement provided by the instrument can be the result of several measurements that are averaged. 
		/// Two levels of averaging are available, one consists in averaging several NMR signals together using a 
		/// repeat or exponential scheme and the other level consists in averaging NMR measurements using repeat, 
		/// moving or exponential averaging. 
		/// Be very careful using signal averaging as it may provide poor measurement results or even oscillation if 
		/// used with a measured magnetic field density that isn't perfectly stable. As such, signal averaging isn't 
		/// recommended and should be used only when confronted to stable but inhomogeneous conditions.
		///
		///@{
			
			// Averaging
		bool ParmAveragingSignalSet(const sSignalAveraging<uParm> & rSigAvg);
		bool ParmAveragingSignalGet(sSignalAveraging<uParm> & rSigAvg);
		bool ParmAveragingSignalGet(sSignalAveraging<sBoundedParm> & rSigAvg);
		bool ParmAveragingMeasurementSet(const sMeasurementAveraging<uParm> & rMeasAvg);
		bool ParmAveragingMeasurementGet(sMeasurementAveraging<uParm> & rMeasAvg);
		bool ParmAveragingMeasurementGet(sMeasurementAveraging<sBoundedParm> & rMeasAvg);
		
		///@}
		/// @name Input trigger parameters
		///
		/// Configure the input trigger subsystem of the instrument. Use these methods to configure which event will 
		/// latch a measurement. Available events are external trigger, internal timed trigger or continuous trigger. 
		/// 
		///@{
			
			// Input Trigger
		bool ParmTriggerInputSet(const sInputTrigger<uParm> & rInputTrig);
		bool ParmTriggerInputGet(sInputTrigger<uParm> & rInputTrig);
		bool ParmTriggerInputGet(sInputTrigger<sBoundedParm> & rInputTrig);
		
		///@}
		///
		/// @name Output trigger parameters
		///
		/// Configures the output trigger subsystem of the instrument. Use these methods to specify the measured 
		/// values that, when reached, must generate an external TTL level or edge. 
		///
		///@{
			
			// Output Trigger
		bool ParmTriggerOutputSet(const sOutputTrigger<uParm> & rOutputTrig);
		bool ParmTriggerOutputGet(sOutputTrigger<uParm> & rOutputTrig);
		bool ParmTriggerOutputGet(sOutputTrigger<sBoundedParm> & rOutputTrig);
		
		///@}
		/// @name RF Pulse parameters
		///
		/// Determines the rate and the width at which the sample is stimulated. The frequency of the pulse is not 
		/// under the scope of these method that are just meant to determine the modulation rate. A simple modulation 
		/// scheme is used, as for today, there is no way to generate other sequences (i.e spin echo, saturation inversion 
		/// recovery, etc.)
		///
		///@{
		
			// RF Pulse
		bool ParmRFPulseSet(const sPulse<uParm> & rPulse);
		bool ParmRFPulseGet(sPulse<uParm> & rPulse);
		bool ParmRFPulseGet(sPulse<sBoundedParm> & rPulse);
		
		///@}
		/// @name Digitization parameters
		///
		/// Configures the digitization subsystem. One may configure the dead-time (the duration during which the 
		/// FID signal isn't sampled right after the RF burst that stimulate the NMR sample), the acquisition duration 
		/// and the sampling rate.
		/// 
		///@{
		
			// Digitization
		bool ParmDigitizationSet(const sDigitization<uParm> & rDigitization);
		bool ParmDigitizationGet(sDigitization<uParm> & rDigitization);
		bool ParmDigitizationGet(sDigitization<sBoundedParm> & rDigitization);
		
		///@}
		/// @name Matching / Tuning parameters
		///
		/// The NMR measuring head consists in a matched / tuned RLC circuit that needs to be set according to the 
		/// operation frequency. These accessors are very specific and shouldn't be used. 
		/// 
		///@{
		
			// Matching / Tuning
		bool ParmMatchingTuningSet(const sMatchingTuning<uParm> & rMatchingTuning);
		bool ParmMatchingTuningGet(sMatchingTuning<uParm> & rMatchingTuning);
		bool ParmMatchingTuningGet(sMatchingTuning<sBoundedParm> & rMatchingTuning);
	
		///@}
		/// @name Units parameters
		///
		/// Determine the unit in which the measurements will be returned by the instrument. One may select Tesla, 
		/// milli-Tesla, Gauss, Kilo-Gauss, MHz relative to the sample, MHz relative to the proton or ppm. For ppm units, 
		/// one must state the reference value. Obviously, the reference must be set prior to set the units to ppm. 
		/// Once the ppm unit is selected, getting the reference will return 0.0, being the exact difference to the reference.
		///
		///@{
			
			// Units
		bool ParmUnitsSet(eUnits Units);
		bool ParmUnitsSetPPMRef(tPPMReference_UNITS Ref);
		bool ParmUnitsGet(eUnits & rUnits, tPPMReference_UNITS & rRef);
		
		///@}
		/// @name Reference Clock parameters
		///
		/// Sets the reference clock port direction. 
		///
		///@{
		
			// Ref Clock
		bool ParmRefClockSet(eReferenceClock RefClock);
		bool ParmRefClockGet(eReferenceClock & rRefClock);
		
		///@}
		
		/// @name Communication : USB
		///
		/// Enables or disables the USB communication port.
		///
		///@{
		
			// Communication : USB
		bool ParmComUsbSet(bool Enabled);
		bool ParmComUsbGet(bool & rEnabled);
		
		///@}
		/// @name Communication : Ethernet
		///
		/// Enables or disables the Ethernet communication port.
		///
		///@{
		
			// Communication : Ethernet
		bool ParmComEthernetSet(const sEthernet & rSettings);
		bool ParmComEthernetGet(sEthernet & rSettings);
		
		///@}
		/// @name Communication : VXI11
		///
		/// Enables or disables the VXI11 compatibility layer.
		///
		///@{
		
			// Communication : VXI11
		bool ParmComVXI11Set(bool Enabled);
		bool ParmComVXI11Get(bool & rEnabled);

		///@}
	///@}
		
		// Settings / Files
		/// @name Settings / Files
		///
		///@{
		
		//----------------------------------------------------------------------//
		// Settings / Files
		bool ReadFileDirectory(U32 & rUsedBytes, U32 & rAvailableBytes, tFileList & rFileList);
		bool ReadFile(std::string Path, std::string & rContent);
		bool WriteFile(std::string Path, const std::string & rContent);
		bool DeleteFile(std::string Path);
		bool GetSettingFileList(std::vector<std::string> & rSettingFileList);
		bool SaveSettings(eSettingType Type, std::string Name);
		bool RecallSettings(std::string Name);

		///@}
		
		/// @name Initiate / Abort / Trigger
		/// Operation control of the measurement system.
		///
		///@{
		
		//----------------------------------------------------------------------//
		// Initiate / Abort / Trigger
		bool Initiate(bool Continuous = false); ///<Starts the search operation and then immediately after NMR resonance detection proceed to NMR measurements. If no NMR resonance is found, the search operation is performed indefinitely.
		bool Abort(bool CancelContinuous = true); ///<Abort the current operation (Search or measure). Once the operation is aborted, the instrument will be waiting to be initiated again. When aborted, only Hall measurement can be read. 
		bool SendBusTrigger(); ///<Send a trigger to the instrument. If configured to react to this bus trigger, the acquisition sub system will collect a measurement and store it for future read out.
		bool ForceOutputTrigger(); ///<Set the output trigger according to current output trigger parametrization. This method is very specific and mainly present for debug purpose.
		
		///@}
		
		/// @name Measurements (data) gathering
		///
		///@{

		//----------------------------------------------------------------------//
		// Measurements / Data
		bool MeasurementsGet(U32 NoMeasurements, std::vector<tFlux> & rFlux, eUnits & rUnits); ///<Returns the NMR magnetic field density measurements, including the unit used. A measurement vector is consistent to the unit. When a measurement was lost or failed for unknown reason, the value is substituted by NaN (Not A Number). 
		bool MeasurementsGet(U32 NoMeasurements, sArbitraryMeasurements ArbSelect, sAllMeasurements & rMeas, eUnits & rUnits); ///<Returns all the data associated to a measurement, including the temporal free induction decay, its FFT, the polynom that was computed to determine the location of the peak, the intermediate frequency, the estimated T2*. This method is very specific and mainly present for debug purpose.
		bool GetSearchProgress(U8 & rSPR, F32 & rHallMeasurement_UNITS);
		bool SetFormat(eCommunicationFormat Format); ///<One can request the data in ASCII or binary format. The binary format, which needs extra operation to be decoded, are preferred as this format reduced the amount of data being sent on the communication link. Setting the format to either ASCII or binary won't affect the results gathered using the MeasurementsGet accessors. 
		
		///@}

		//----------------------------------------------------------------------//
		// High Level Measurement
		//bool Measure(tFlux & rFlux, MTL::SCPI::tChannelList ChanList, F64 ExpectedField = 0, unsigned int NoDigits = 0);
		//bool Measure(std::vector<tFlux> & rFluxArray, unsigned int NoMeasurements, MTL::SCPI::tChannelList ChanList, F64 ExpectedField = 0, unsigned int NoDigits = 0);

		/// @name Instrument Status Handling
		/// See the user manual for a detailed explanation of the instrument status subsystem.
		///
		///@{
		
		//----------------------------------------------------------------------//
		// Status Handling
		bool StatusPreset();
		bool StatusGet(sStatusRegister Reg, U16 & rStatus);
		bool StatusGet(std::vector<sStatusRegister> Regs, std::vector<U16> & rStatus);
		bool StatusSetEnableRegister(eStatusRegisterSet Set, U16 DisableMask, U16 EnableMask);
		bool StatusSetTransitionFilter(eStatusRegisterSet Set, eTransition Transition, U16 ClearMask, U16 SetMask);
		bool StandardStatusSetEnableRegister(eStandardStatusRegister Set, U16 DisableMask, U16 EnableMask);

		///@}
		
		/// @name Utilities
		///@{
		
		//----------------------------------------------------------------------//
		// Utilities
		bool GetIdentification(std::string & rIdentification); ///<Returns the instrument identification string including its serial number and make. This string contains also the various electronic boards versions along the internal firmware and files versions.
		bool PowerOff(bool Reboot = false); ///<Sending this command will shut the instrument down. Setting the reboot parameter to true will initiate a full shut-down and reboot of the instrument. Be careful, the link to the instrument is lost and must be re-acquired. It is a safe practice to release the instrument lock before performing to this operation.
		bool Reset(); ///<Calling this method will wipe out all user parameters that are stored in the file system and revert them back to factory defaults.  
		bool ReadCaseTemperature(F32 & rTemperatureDegC); ///<Returns the internal instrument temperature. An internal survey will shut the instrument down if this value reaches a temperature that is factory set to prevent the system to fail. The probe temperature isn't measured.
		bool SwitchToDFUMode(); ///<Restart the instrument in USB DFU mode, allowing the firmware to be upgraded.
		bool ReadDebugLog(std::string & rDebugLog); ///<Returns a long string containing all the operations that were processed before the last proper instrument shut down.
		bool ReadInformationDates(std::string & rSManufacturingDate, std::time_t & rManufacturingDate, std::string & rSCalibrationDate, std::time_t & rCalibrationDate); ///<Returns the manufacturing and calibration date in plain text or a number representing the amount of seconds since 00:00, Jan 1 1970 UTC.
		
		///@}
		
		// Low Level Utilities
		/// @name Low Level Utilities
		///@{
		
		//----------------------------------------------------------------------//
		// Low Level Utilities
		bool DeviceROMRead(const MTL::SCPI::tChannel & rChannel, std::vector<U8> & rROMContent);
		bool DeviceROMWrite(const MTL::SCPI::tChannel & rChannel, const std::vector<U8> & rROMContent);
		bool NormalizationGet(const MTL::SCPI::tChannel & rChannel, tNormalizationTable & rNormalizationTable);
		bool NormalizationSet(const MTL::SCPI::tChannel & rChannel, const tNormalizationTable & rNormalizationTable);
		bool AuthenticationStatus(bool & rIsAuthenticated, std::string & rCipherMsgGet);
		bool AuthenticationRequest(const std::string & rCipherMsgSet, bool & rIsAuthenticated);
		bool LowLevelTransfer(const std::vector<U8> & rDataToSend, std::vector<U8> & rRcvdData, U8 & rReturnedStatus);
		bool WriteCustomCommand(const std::string & rWriteStr);

		///@}
	};

	} // namespace Instrument
} // namespace MTL
