#pragma once

// Standard includes
#include <string>
#include <vector>

// Personnal includes
#include "SCPIParsing.h"
#include "OSDefines.h"

namespace MTL {
	namespace Instrument {
		namespace PT2026Types {

			//----------------------------------------------------------------------//
			//								Status									//
			//----------------------------------------------------------------------//
			enum eStandardStatusRegister {
				kStandardEventStatusRegister,
				kServiceRequestRegister
			};
			enum eStatusRegisterSet {
				kStatusQestionable,
				kStatusOperation,
				kStatusOperationBit11,
				kStatusOperationBit12,
				kStatusQuestionableBit12
			};
			enum eStatusRegisterType {
				kStatusEvent,
				kStatusCondition,
				kStatusEnable
			};
			struct sStatusRegister {
				eStatusRegisterSet	Set;
				eStatusRegisterType	Type;
				sStatusRegister(eStatusRegisterSet S, eStatusRegisterType T)
					: Set(S), Type(T) {}
			};
			enum eTransition {
				kPositive,
				kNegative
			};
			union uStatusByte {
				U16	RawSTB = 0;
				struct sStatusByte {
					// Contains a 1-byte status summary. The PT2026 uses the following bits:
					U8		NA0_1	: 2;	// Not Available
					U8		EAV		: 1;	// Error Available in Error / Event Queue
					U8		QSB		: 1;	// QUEStionable status summary bit
					U8		MAV		: 1;	// Message Available
					U8		ESB		: 1;	// Standard Event Status Register summary bit
					U8		RQS		: 1;	// ReQuest for Service
					U8		OSB		: 1;	// OPERation status summary bit
				} StatusByte;
			};

			union uStandardEvent {
				U8		RawSE = 0;
				struct sStandardEvent {
					// Latches certain standardized events. The PT2026 uses the following bits:
					U8		OperationComplete		: 1;	// *OPC has flagged operation complete
					U8		NA1						: 1;	// Not Available
					U8		QueryError				: 1;	// Error in preceding query
					U8		DeviceDependentError	: 1;	// Errors specific to the PT2026, including internal errors
					U8		ExecutionError			: 1;	// Error detected during command execution
					U8		CommandError			: 1;	// Error in preceding command
					U8		UserRequest				: 1;	// A parameter has been changed
					U8		PowerOn					: 1;	// PT2026 has been powered up
				} StandardEvent;
			};

			union uQUEStionable{
				U16	RawQUES = 0;
				struct sQUEStionable {
					// Indicates conditions that may reduce the quality of the measurement. The PT2026 sets the following bits:
					U16		NA0_8				: 9;	// Not Available
					U16		UnableToMeasure		: 1;	// Unable to acquire and measure an NMR signal
					U16		IgnoredSetting		: 1;	// A setting was set, but is not taken into account because of other settings
					U16		Measurement			: 1;	// The quality of the measurement is questionable (e.g. Due to unknown IF in manual search mode)
					U16		BIT12				: 1;	// Summary Bit: DSP Status
					U16		NA13_15				: 3;	// Not Available
				} QUEStionable;
			};

			union uQUEStionableBIT12{
				U16 RawQUES12 = 0;
				struct sQUEStionableBIT12 {
					// This is a fan-out for bit 12 of the QUEStionable register, a device-specific summary bit.
					U16		ExternalRefDetected	: 1;	// External reference clock is detected
					U16		ExternalRefLocked	: 1;	// Locked onto external reference clock
					U16		BeatingDetected		: 1;	// Measurement was rejected as spurious signal (cf. :CONFigure:MEASure:REJect command)
					U16		NA3_15				: 13;	// Not Available
				} QUEStionableBIT12;
			};

			union uOPERation {
				U16 RawOPER = 0;
				struct sOPERation {
					// Captures conditions which are part of the instrument s normal operation. The PT2026 uses the following bits:
					U16		NA0_1				: 2;	// Not Available
					U16		RANGing				: 1;	// Scanning for available probes
					U16		SWEeping			: 1;	// Searching for NMR signal
					U16		MEASuring			: 1;	// Measuring magnetic field strength
					U16		WaitingForTRIGger	: 1;	// Waiting for trigger
					U16		NA6_7				: 2;	// Not Available
					U16		NewAcqAvail			: 1;	// New Acquisition Data is Available
					U16		NewMeasAvail		: 1;	// New Measurement Available
					U16		INITiated			: 1;	// Currently in the INITiated state
					U16		BIT11				: 1;	// BIT11 Summary Bit: A parameter has changed
					U16		BIT12				: 1;	// BIT12 Summary Bit: Acquisition status
					U16		NA13_15				: 3;	// Not Available
				} OPERation;
			};

			union uOPERationBIT11 {
				U16 RawOPER11 = 0;
				struct sOPERationBIT11 {
					// This is a fan-out for bit 11 of the OPERation register, a device-specific summary bit. It indicates subsystems whose configuration has changed.
					U16		SYSTem			: 1;	// System: system ID, communications, lock
					U16		STATus			: 1;	// Status: enable bits
					U16		MEMory			: 1;	// Memory: parameter file loaded
					U16		MMEMory			: 1;	// Mass memory: parameters saved, created, deleted
					U16		CONFigure		: 1;	// Measurement: matching/tuning, search, tracking
					U16		ROUTe			: 1;	// Route: channel selection
					U16		INPut			: 1;	// Input: reference clock
					U16		OUTput			: 1;	// Output: trigger output signal
					U16		SENse			: 1;	// Sense: digitization
					U16		SOURce			: 1;	// Source: pulse
					U16		TRIGger			: 1;	// Trigger: measurement trigger, output trigger
					U16		CALCulate		: 1;	// Calculate: signal & measurement averaging
					U16		FORMat			: 1;	// Format: output format
					U16		UNIT			: 1;	// Units: output units
					U16		NAS14_15		: 2;	// Not Available
				} OPERationBIT11;
			};

			union uOPERationBIT12 {
				U16 RawOPER12 = 0;
				struct sOPERationBIT12 {
					// This is a fan-out for bit 12 of the OPERation register, a device-specific summary bit.
					U16		DSP_CPLDFatalError	: 1;	// Unable to use DSP and/or CPLD
					U16		NA1_2				: 2;	// Not Available
					U16		AcquisitionOverrun	: 1;	// DSP s acquisition buffer is full
					U16		TriggerOverrun		: 1;	// Trigger arrived before previous acquisition complete
					U16		NMRPathBroken		: 1;	// Active probe was unplugged
					U16		PeripheralPlugged	: 1;	// A peripheral was plugged into the PT2026
					U16		RemoteBoxStart		: 1;	// Start button of the remote box has been pressed
					U16		NA8_15				: 8;	// Not Available
				} OPERationBIT12;
			};

			//----------------------------------------------------------------------//
			//								Units									//
			//----------------------------------------------------------------------//
			enum eUnits {
				kT, kmT, kGauss, kkGauss, kppm, kMHzp, kMHz
			};

			//----------------------------------------------------------------------//
			//								Errors									//
			//----------------------------------------------------------------------//
			struct sError {
				I32				Code;
				std::string		Description;
				std::string		Context;

				explicit operator bool() const
				{	return Code != 0;	}
			};
			
			//----------------------------------------------------------------------//
			//								Parameters								//
			//----------------------------------------------------------------------//
			//------------------------------------------//
			// Generic
			template <class DataType>
			struct sBoundedParm
			{
				DataType	Val;	// Current value
				DataType	Min;	// Low bound
				DataType	Max;	// High bound
				DataType	Def;	// Default value
			};
			template <class DataType>
			using uParm = DataType;

			template <template<class> class ParmType>
			struct sExample
			{
				bool				ExampleState;
				ParmType<F64>		ExampleF64;
				ParmType<F32>		ExampleF32;
			};
			// Declare as: sExample<uParm> or sExample<sBoundedParm>

			//------------------------------------------//
			// Averaging
			enum eSignalAveragingType {
				kSigNone,
				kSigExponential,
				kSigRepeat
			};
			template <template<class> class ParmType>
			struct sSignalAveraging {
				eSignalAveragingType	Type;
				ParmType<U32>			NoPoints;
			};

			enum eMeasurementAveragingType {
				kMeasNone,
				kMeasExponential,
				kMeasMoving,
				kMeasRepeat
			};
			template <template<class> class ParmType>
			struct sMeasurementAveraging {
				eMeasurementAveragingType	Type;
				ParmType<U32>				NoPoints;
			};

			//------------------------------------------//
			// Digitization
			enum eDigitizationMode {
				kDigitizationAuto,
				kDigitizationManual
			};
			template <template<class> class ParmType>
			struct sDigitization {
				eDigitizationMode		Mode;
				ParmType<F64>			DeadTime_s;
				ParmType<F64>			SweepTime_s;
				ParmType<F64>			Rate_Hz;
			};

			//------------------------------------------//
			// Input Trigger
			enum eInputTriggerSource {
				kInputTrigSrceImmediate,
				kInputTrigSrceTimer,
				kInputTrigSrceBus,
				kInputTrigSrceExternal
			};
			enum eInputTriggerEdge {
				kInputTrigEdgeRising,
				kInputTrigEdgeFalling,
				kInputTrigEdgeBoth
			};
			template <template<class> class ParmType>
			struct sInputTrigger {
				eInputTriggerSource		Source;
				ParmType<F64>			Period_s;
				eInputTriggerEdge		Edge;
				ParmType<U8>			Count;
			};

			//------------------------------------------//
			// Output Trigger
			enum eOutputTriggerMode {
				kOutputTrigShapeOFF,
				kOutputTrigShapeDCBGreaterThan,
				kOutputTrigShapeDCBLowerThan,
				kOutputTrigShapePulseBRising,
				kOutputTrigShapePulseBFalling,
				kOutputTrigShapePulseBCrossing
			};
			template <template<class> class ParmType>
			struct sOutputTrigger {
				eOutputTriggerMode		Mode;
				ParmType<F64>			Level_UNITS;
				bool					InvertedPolarity;
				ParmType<F64>			PulseWidth_s;
				ParmType<F64>			OutputDelay_s;
			};

			//------------------------------------------//
			// Measure
			enum eMeasureMode {
				kMeasureAuto,
				kMeasureManual
			};
			template <template<class> class ParmType>
			struct sTracking {
				ParmType<F64>		MinIF_Hz;
				ParmType<F64>		MaxIF_Hz;
				ParmType<U32>		AllowedMissMeas;
			};
			template <template<class> class ParmType>
			struct sMeasure {
				eMeasureMode			Mode;
				bool					Reject;
				ParmType<F64>			DetectionLevel_V;
				ParmType<F64>			Bandwidth_Hz;
				ParmType<U32>			NoFitPoints;
				ParmType<U32>			AllowedMissMeas;
				sTracking<ParmType>		Tracking;
			};

			//------------------------------------------//
			// Matching-Tuning
			enum eMatchingTuningMode {
				kMatchTunAuto,
				kMatchTunManual
			};
			template <template<class> class ParmType>
			struct sMatchingTuning {
				eMatchingTuningMode		Mode;
				ParmType<F64>			Matching;
				ParmType<F64>			Tuning;
			};

			//------------------------------------------//
			// Pulse
			enum ePulseMode {
				kPulseAuto,
				kPulseManual
			};
			template <template<class> class ParmType>
			struct sPulse {
				ePulseMode			Mode;
				ParmType<F64>		Period_s;
				ParmType<F64>		Width_s;
			};

			//------------------------------------------//
			// Search
			enum eSearchMode {
				kSearchAuto,
				kSearchManual
			};
			template <template<class> class ParmType>
			struct sSearch {
				bool				Enable;
				ParmType<F64>		ManualValue_UNITS;
				eSearchMode			Mode;
				ParmType<F64>		FrequencyStep_Hz;
				ParmType<F64>		LowLimit_UNITS;
				ParmType<F64>		HighLimit_UNITS;
			};

			//------------------------------------------//
			// Reference
			typedef F64		tPPMReference_UNITS;

			//------------------------------------------//
			// Clock
			enum eReferenceClock {
				kRefClockInternal,
				kRefClockExternal
			};

			//------------------------------------------//
			// Format
			enum eCommunicationFormat {
				kComFormatAscii,
				kComFormatInteger
			};

			//------------------------------------------//
			// Ethernet
			union uIPAddress {
				U8		Address[4];
				struct {
					U8	A;
					U8	B;
					U8	C;
					U8	D;
				} Member;
			};
			struct sEthernet {
				bool			Enabled;
				uIPAddress		IP;
				uIPAddress		Mask;
				std::string		Domain;
				uIPAddress		Broadcast;
				uIPAddress		Gateway;
				uIPAddress		DNSServer;
				bool			NBNSEnabled;
				std::string		Identifier;
			};
			//------------------------------------------//
			// Modbus
			struct sModbus {
				bool						Enabled;
				std::vector<std::string>	Hosts;
				U32							AllowedNoConnections;
				bool						Use32bitsFloats;
			};
			//------------------------------------------//
			// Channels
			typedef U32		tModel;
			typedef U32		tSerialNumber;
			enum ePeriphObjectType		// Type as presented by PT2026 through SCPI
			{
				kPwProbe			= 0,	// Pulsed probe
				kPwCamera			= 1,	// Pulsed camera
				kPwCameraHead		= 2,	// Pulsed camera head
				kPwMultiplexer		= 3,	// Pulsed probe multiplexer
				kPwFCA				= 4,	// Pulsed camera adapter
				NbTypes
			};
			static const char* PeriphObjectTypeStr[] =
			{
				"kPwProbe",
				"kPwCamera",
				"kPwCameraHead",
				"kPwMultiplexer",
				"kPwFCA"
			};
			struct sExtendedChannel {
				SCPI::tChannel		Channel;
				ePeriphObjectType	Type;
				tModel				Model;
				tSerialNumber		SerialNumber;
				U8					NoPorts;
				bool				CanMeasure;

				sExtendedChannel()
					: Type(kPwProbe), Model(0), SerialNumber(0), NoPorts(0), CanMeasure(false)
				{}

				inline bool operator==(const sExtendedChannel & other)
				{
					return (	Channel == other.Channel &&
								Type == other.Type &&
								Model == other.Model &&
								SerialNumber == other.SerialNumber &&
								NoPorts == other.NoPorts &&
								CanMeasure == other.CanMeasure);
				}
				inline bool operator!=(const sExtendedChannel & other)
				{
					return !(*this == other);
				}
				void clear(void)
				{
					Channel.clear();
					Type = kPwProbe;
					Model = 0;
					SerialNumber = 0;
					NoPorts = 0;
					CanMeasure = false;
				}
			};
			class CExtendedChannelList : public std::vector<sExtendedChannel>
			{
			public:
				inline bool operator==(CExtendedChannelList & other)
				{
					if (size() != other.size())
						return false;
					for (CExtendedChannelList::iterator it = begin(), otherit = other.begin(); it != end(); it++, otherit++)
					{
						if (*it != *otherit)
							return false;
					}
					return true;
				}
				inline bool operator!=(CExtendedChannelList & other)
				{
					return !(*this == other);
				}
			};
			//------------------------------------------//
			// Channel Information
			typedef F64		tProbeLimit_UNITS;
			typedef U16		tFirmwareVersion;
			struct sChannelIformation {
				tModel				Model;
				tSerialNumber		SN;
				std::string			Designation;
				tFirmwareVersion	FirmwareVersion;
				sChannelIformation(tModel Mod, tSerialNumber Ser, std::string Des, tFirmwareVersion FWV)
					: Model(Mod), SN(Ser), Designation(Des), FirmwareVersion(FWV) {}
			};
			//------------------------------------------//
			// Hall
			struct sProbeHall {
				F64		B;
				F64		Bx;
				F64		By;
				F64		Bz;
				void clear()
				{
					B	= 0;
					Bx	= 0;
					By	= 0;
					Bz	= 0;
				}
			};
			struct sProbeRawHall {
				U16		Bx;
				U16		By;
				U16		Bz;
				U16		T;
				void clear()
				{
					Bx	= 0;
					By	= 0;
					Bz	= 0;
					T	= 0;
				}
			};
			//------------------------------------------//
			// Remote Busy LED modes
			enum eRemoteBusyLEDmode {
				kOff						= 0,	// Busy LED is always Off
				kOnWhenMeasuring			= 1,	// Busy LED is On when measuring, Off otherwise
				kClearsAfterMeasurement		= 2,	// Busy LED in immediately On, and clears after measurement
				kOn							= 3,	// Busy LED is always On
				kBlinkSlow					= 4,	// Busy LED blinks at a slow rate
				kBlinkFast					= 5		// Busy LED blinks at a fast rate
			};
			const U8 MAX_BUSY_LED_MODE		= kBlinkFast;

			//------------------------------------------//
			// Files
			struct sFile {
				std::string		Path;
				std::string		Type;
				size_t			Size;
			};
			typedef std::vector<sFile>		tFileList;
			enum eSettingType {
				kAll,
				kCommunication,
				kMeasure,
				kTriggers,
				kReferenceClock,
				kUnits
			};

			//----------------------------------------------------------------------//
			//								Measurements							//
			//----------------------------------------------------------------------//
			//------------------------------------------//
			// NMR
			typedef F64		tFlux;
			union uMeasStatus {
				U8		StatusByte;
				struct sMeasStatus {
					U8		Questionable : 1;	// Measurement is questionable
					U8		NA1_7 : 7;			// Not Available
				}		StatusBits;
			};
			struct sBasicMeasurement {
				tFlux					Flux_UNITS;
				F32						Deviation_ppm;
				F32						Uniformity;
				MTL::SCPI::tChannel		Channel;
				U64						Timestamp;
				uMeasStatus				Status;
				U32						NbValidMeas;
				void clear()
				{
					Flux_UNITS		= DPqNaN;
					Deviation_ppm	= SPqNaN;
					Uniformity		= SPqNaN;
					Channel.clear();
					Timestamp		= 0;
					Status			= { 0 };
					NbValidMeas		= 0;
				}
			};
			struct sAdvancedMeasurements {
				std::vector<sBasicMeasurement>	Basic;
				F64								IntermediateFrequency_Hz;
				F32								RelaxationTime_s;
				F64								RadioFrequency_Hz;
				void clear()
				{
					Basic.clear();
					IntermediateFrequency_Hz = DPqNaN;
					RelaxationTime_s = SPqNaN;
					RadioFrequency_Hz = DPqNaN;
				}
			};
			struct sNMRSignal {
				F32					SamplePeriod_s;
				std::vector<F32>	Samples_V;
				void clear()
				{
					SamplePeriod_s = SPqNaN;
					Samples_V.clear();
				}
			};
			struct sFFTBuffer {
				F32					FrequencyResolution_Hz;
				std::vector<F32>	Samples_V;
				void clear()
				{
					FrequencyResolution_Hz = SPqNaN;
					Samples_V.clear();
				}
			};
			struct sSpectrum {
				F32					CenterFrequency_Hz;
				F32					FrequencyResolution_Hz;
				std::vector<F32>	Samples_V;
				void clear()
				{
					CenterFrequency_Hz = SPqNaN;
					FrequencyResolution_Hz = SPqNaN;
					Samples_V.clear();
				}
			};
			struct sFitPolynom {
				F32					Residue;
				std::vector<F32>	Coeffs;
				void clear()
				{
					Residue = SPqNaN;
					Coeffs.clear();
				}
			};
			struct sArbitraryMeasurements {
				bool	Flux				= false;
				bool	Deviation			= false;
				bool	Uniformity			= false;
				bool	Channel				= false;
				bool	Timestamp			= false;
				bool	Status				= false;
				bool	NbValidMeas			= false;
				bool	IF					= false;
				bool	RelaxationTime		= false;
				bool	RadioFrequency		= false;
				U32		NoNMRSamples		= 0;
				U32		NoFFTSamples		= 0;
				U32		NoSpectrumSamples	= 0;
				U32		PolynomDegree		= 0;
			};
			struct sAllMeasurements {
				sAdvancedMeasurements	AdvMeas;
				F64						IntermediateFrequency_Hz;
				F32						RelaxationTime_s;
				F64						RadioFrequency_Hz;
				sNMRSignal				NMR;
				sFFTBuffer				FFT;
				sSpectrum				Spectrum;
				sFitPolynom				FitPolynom;
				void clear()
				{
					AdvMeas.clear();
					IntermediateFrequency_Hz = 0;
					RelaxationTime_s = 0;
					RadioFrequency_Hz = 0;
					NMR.clear();
					FFT.clear();
					Spectrum.clear();
					FitPolynom.clear();
				}
			};

			//------------------------------------------//
			// Hall
			struct sHallMeasurement {
				F64		B;
				F64		Bx;
				F64		By;
				F64		Bz;
			};

			//------------------------------------------//
			// Normalization
			typedef F32		tNormalizationFactor_Hz;								// [Hz]
			typedef std::vector<tNormalizationFactor_Hz>	tNormalizationTable;


		} // namespace PT2026Types
	} // namespace Instrument
} // namespace MTL
