#pragma once

#include "Catalog.h"
#include <functional>

namespace MTL {
	namespace InstrumentCatalogInterface {

		//----------------------------------------------------------------------//
		//							Catalog Interface							//
		//----------------------------------------------------------------------//
		//----------------------------------------------------------------------
		// Control
		template <class DataType>
		class CControl
		{
		private:
			CVariable<DataType>	& m_rVariable;
		public:
			// Constrution / Destruction
			CControl(CVariable<DataType> & rVariable)
				: m_rVariable(rVariable)
			{}
			virtual ~CControl()
			{}

			// Control
			void RequestControl(const DataType NewValue)		// Request a control with the given value.
			{
				m_rVariable.Write(NewValue);
			}
			void rRequestControl(const DataType & rNewValue)	// Request a control with the given value, by reference.
			{
				m_rVariable.rWrite(rNewValue);
			}
			void operator() (const DataType & rNewValue)		// Request a control with the given value, by reference.
			{
				m_rVariable.rWrite(rNewValue);
			}
			void operator= (const DataType & rNewValue)	// Request a control with the given value, by reference.
			{
				m_rVariable.rWrite(rNewValue);
			}

			// Reading
			DataType ReadCurrent(void)						// Read the current value of the control.
			{
				return m_rVariable.Read();
			}
			DataType operator() ()							// Read the current value of the control.
			{
				return  m_rVariable.Read();
			}
			void rReadCurrent(DataType & rReturnedValue)	// Read the current value of the control, by reference.
			{
				m_rVariable.rRead(rReturnedValue);
			}
		};

		//----------------------------------------------------------------------
		// Untyped observer
		class CObserver_untyped	// Provide a generic class which is untyped
		{
		public:
			// Constrution / Destruction
			virtual ~CObserver_untyped()
			{}

			// Identification
			bool Is(const CObserver_untyped & rObs)
			{ return (this == &rObs); }
			bool Is(const CObserver_untyped * pObs)
			{ return (this == pObs); }

			// Copy manipulation
			virtual void * Copy(void) = 0;
			virtual void Delete(void * const pAllocatedData) = 0;
		};

		//----------------------------------------------------------------------
		// Observer
		template <class DataType>
		class CObserver : public CObserver_untyped
		{
		protected:
			CVariable<DataType>	& m_rVariable;
		public:
			// Constrution / Destruction
			CObserver(CVariable<DataType> & rVariable)
				: m_rVariable(rVariable) {}
			virtual ~CObserver()
			{}

			// Subscription
			void Subscribe(CObserverNotifier_untyped & rNotifier)		// Subscribe to observer change. Notify that changes through the given Notifier.
			{
				m_rVariable.Subscribe(rNotifier, *this);
			}
			void Unsubscribe(CObserverNotifier_untyped & rNotifier)		// Unsubscribe from observer change.
			{
				m_rVariable.Unsubscribe(rNotifier);
			}

			// Reading
			DataType ReadCurrent(void)						// Read the current value of the observer.
			{
				return m_rVariable.Read();
			}
			DataType operator() ()							// Read the current value of the control.
			{
				return  m_rVariable.Read();
			}
			void rReadCurrent(DataType & rReturnedValue)	// Read the current value of the observer, by reference.
			{
				m_rVariable.rRead(rReturnedValue);
			}
			
			// Copy manipulation
			void * Copy(void)								// Allocate a copy of the content of the observer and return its pointer.
			{
				DataType * l_pDataCopy = new DataType;
				m_rVariable.rRead(*l_pDataCopy);
				return l_pDataCopy;
			}
			void Delete(void * const pAllocatedData)		// Delete the content from the allocated buffer.
			{
				if (NULL != pAllocatedData)
					return;
				delete reinterpret_cast<DataType *>(pAllocatedData);
			}
		};

		//----------------------------------------------------------------------//
		//						Synchronous Notifications						//
		//----------------------------------------------------------------------//
		// There is no synchronous notification class since a synchronous
		// notifier consists in a direct call of a callback function.

		//----------------------------------------------------------------------//
		//						Asynchronous Notifications						//
		//----------------------------------------------------------------------//
		//----------------------------------------------------------------------
		// Generic asynchronous notification (not to be used directly)
		class CAsyncNotification
		{
		private:
			CObserverNotifier_untyped &		m_rNotifier;
		protected:
			CObserver_untyped &				m_rObserver;
		protected:
			// Construction / Destruction
			CAsyncNotification(CObserverNotifier_untyped & rNot, CObserver_untyped & rObs)
				: m_rNotifier(rNot), m_rObserver(rObs)
			{}
		public:
			virtual ~CAsyncNotification()
			{}

			// Member access
			CObserverNotifier_untyped & Notifier() const
			{ return m_rNotifier; }
			CObserver_untyped & Observer() const
			{ return m_rObserver; }
		};

		//----------------------------------------------------------------------
		// Asynchronous dynamic notification (notification whose data can change between emission and reception of this notification)
		class CAsyncDynamicNotification : public CAsyncNotification
		{
		public:
			CAsyncDynamicNotification(CObserverNotifier_untyped & rNot, CObserver_untyped & rObs)
				: CAsyncNotification(rNot, rObs)
			{}
			virtual ~CAsyncDynamicNotification()
			{}
		};

		//----------------------------------------------------------------------
		// Asynchronous snapshop notification (notification which embedds a copy of the associated observer data)
		class CAsyncSnapshotNotification : public CAsyncNotification
		{
		private:
			void *			m_pDataCopy;
		public:
			CAsyncSnapshotNotification(CObserverNotifier_untyped & rNot, CObserver_untyped & rObs)
				: CAsyncNotification(rNot, rObs), m_pDataCopy(NULL)
			{
				m_pDataCopy = m_rObserver.Copy();
			}
			virtual ~CAsyncSnapshotNotification()
			{ 
				m_rObserver.Delete(m_pDataCopy);
			}

			const void * DataCopy () const
			{ return m_pDataCopy;	}
		};

		//----------------------------------------------------------------------//
		//								Notifiers								//
		//----------------------------------------------------------------------//
		//----------------------------------------------------------------------
		// Generic untyped notifier class
		class CObserverNotifier_untyped	// Provide a generic class which is untyped
		{
		public:
			typedef enum NotifierType {
				kSync,			// Synchronous notifier.
				kAsyncDynamic,	// Asynchronous notifier with dynamic data.
				kAsyncSnapshot	// Asynchronous notifier including a snapshot of the data.
			} tNotifierType;
		private:
			const tNotifierType		m_NotifierType;
		public:
			CObserverNotifier_untyped(tNotifierType NotifType)
				: m_NotifierType(NotifType)
			{}
			virtual ~CObserverNotifier_untyped()
			{}

			bool Is(CObserverNotifier_untyped & rNot)
			{ return (this == &rNot); }
			bool Is(const CObserverNotifier_untyped * pNot)
			{ return (this == pNot); }
			const tNotifierType & Type()
			{ return m_NotifierType; }

			virtual void NotifyVarUpdate(CObserver_untyped & rObs) = 0;
		};

		//----------------------------------------------------------------------
		// Synchronous notifier
		template <class DataType>
		class CSyncNotifier : public CObserverNotifier_untyped
		{
		public:
			typedef std::function<void(CObserver<DataType> & rObserver)>	tCallback;
		private:
			tCallback					m_Callback;
		public:
			// Constructor as: CSyncNotifier<DataType> Notifier(Function)
			CSyncNotifier(const tCallback & rCallB)
				: CObserverNotifier_untyped(kSync), m_Callback(rCallB)
			{}
			// Constructor as: CSyncNotifier<DataType> Notifier(&Class::method, &ClassObj)
			template<class Fn, class... Args>
			CSyncNotifier(Fn&& MemberCallback, Args&&... ClassObj)
				: CObserverNotifier_untyped(kSync), m_Callback(std::bind(MemberCallback, ClassObj..., std::placeholders::_1))
			{}
			virtual ~CSyncNotifier()
			{}

			void NotifyVarUpdate(CObserver_untyped & rObs)
			{
				// Call the callback
				m_Callback(reinterpret_cast<CObserver<DataType> &>(rObs));
			}
		};

		//----------------------------------------------------------------------
		// Asynchronous notifier
		template <class NotificationType>
		class CAsyncNotifier : public CObserverNotifier_untyped
		{
		public:
			typedef std::function<void(CAsyncNotification * pAsyncNotification)>	tPushEventCallback;
		private:
			tPushEventCallback			m_PushNotifCallback;
		protected:
			CAsyncNotifier(tNotifierType NotifType, const tPushEventCallback & rCallB)
				: CObserverNotifier_untyped(NotifType), m_PushNotifCallback(rCallB)
			{}
			virtual ~CAsyncNotifier()
			{}

			void NotifyVarUpdate(CObserver_untyped & rObs)
			{
				CAsyncNotification * l_pNotif = new NotificationType(*this, rObs);
				m_PushNotifCallback(l_pNotif);
			}
		};

		//----------------------------------------------------------------------
		// Asynchronous Dynamic notifier
		class CAsyncDynamicNotifier : public CAsyncNotifier<CAsyncDynamicNotification>
		{
		public:
			// Constructor as: CAsyncDynamicNotifier Notifier(Function)
			CAsyncDynamicNotifier(const tPushEventCallback & rCallB)
				: CAsyncNotifier<CAsyncDynamicNotification>(CObserverNotifier_untyped::kAsyncDynamic, rCallB) {}
			// Constructor as: CAsyncDynamicNotifier Notifier(&Class::method, &ClassObj)
			template<class Fn, class... Args>
			CAsyncDynamicNotifier(Fn&& MemberCallback, Args&&... ClassObj)
				: CAsyncNotifier<CAsyncDynamicNotification>(CObserverNotifier_untyped::kAsyncDynamic, std::bind(MemberCallback, ClassObj..., std::placeholders::_1)) {}
		};
		
		//----------------------------------------------------------------------
		// Asynchronous Snapshot notifier (embedded data copy)
		class CAsyncSnaphsotNotifier : public CAsyncNotifier<CAsyncSnapshotNotification>
		{
		public:
			// Constructor as: CAsyncSnaphsotNotifier Notifier(Function)
			CAsyncSnaphsotNotifier(const tPushEventCallback & rCallB)
				: CAsyncNotifier<CAsyncSnapshotNotification>(CObserverNotifier_untyped::kAsyncSnapshot, rCallB) {}
			// Constructor as: CAsyncSnaphsotNotifier Notifier(&Class::method, &ClassObj)
			template<class Fn, class... Args>
			CAsyncSnaphsotNotifier(Fn&& MemberCallback, Args&&... ClassObj)
				: CAsyncNotifier<CAsyncSnapshotNotification>(CObserverNotifier_untyped::kAsyncSnapshot, std::bind(MemberCallback, ClassObj..., std::placeholders::_1)) {}
		};

	} // namespace InstrumentCatalogInterface;
} // namespace MTL
