//////////////////////////////////////////////////////////////////////////
/// \file
/// \brief C++ wrapper for NI-VISA: type definitions.

#pragma once

// External tools
#include "visa.h"

// Standard libraries
#include <string>
#include <vector>
#include <iostream>

namespace MTL {
	namespace Instrument
	{
		//----------------------------------------------------------------------//
		//								Types									//
		//----------------------------------------------------------------------//
		typedef std::string		tResourceName;

		class CResourceList : public std::vector<tResourceName>
		{
		public:
			std::string Dump() const
			{
				std::string l_RsrcList;
				for (CResourceList::const_iterator it = this->begin(); it != this->end(); it++)
					l_RsrcList += *it + '\n';
				return l_RsrcList;
			}
			bool operator==(CResourceList const & other)
			{
				if (other.size() != size())
					return false;
				for (CResourceList::const_iterator thisit = begin(), otherit = other.begin(); thisit != end(); thisit++, otherit++)
				{
					if (*thisit != *otherit)
						return false;
				}
				return true;
			}
			bool operator!=(CResourceList const & other)
			{
				if (other.size() != size())
					return true;
				for (CResourceList::const_iterator thisit = begin(), otherit = other.begin(); thisit != end(); thisit++, otherit++)
				{
					if (*thisit != *otherit)
						return true;
				}
				return false;
			}
		};

		enum class eInterfaceType {
			GPIB		= VI_INTF_GPIB,
			VXI			= VI_INTF_VXI,
			GPIB_VXI	= VI_INTF_GPIB_VXI,
			ASRL		= VI_INTF_ASRL,
			PXI			= VI_INTF_PXI,
			TCPIP		= VI_INTF_TCPIP,
			USB			= VI_INTF_USB
		};

		typedef ViUInt16		tInterfaceNumber;

		enum class eResourceClass {
			INSTR = 0,
			MEMACC,
			INTFC,
			BACKPLANE,
			SERVANT,
			SOCKET,
			RAW
		};

		enum class eOpenAccessMode {
			NoLock			= VI_NO_LOCK,				// Open session without using an exclusive lock or loading configuration information
			ExclusiveLock	= VI_EXCLUSIVE_LOCK,		// Acquire an exclusive lock immediately upon opening a session. If a lock cannot be acquired, the session is closed and an error is returned
			LoadConfig		= VI_LOAD_CONFIG			// Configure attributes to values specified by an external configuration utility, such as the following utilities:
														// - (Windows) Measurement & Automation Explorer
														// - (Mac OS X) NI - VISA Configuration
														// - (Linux) visaconf
		};

		enum class eEventMechanism {
			Queue				= VI_QUEUE,				// Enable the session to queue events
			Handle				= VI_HNDLR,				// Enable the session to invoke a callback function
			Suspendedhandler	= VI_SUSPEND_HNDLR,		// Enable the session to execute the handler
			All					= VI_ALL_MECH			// Enable the session to all events
		};

		enum class eEventType : unsigned long {
			ServiceRequest		= VI_EVENT_SERVICE_REQ,			// Notifies the application that a service request was received from the device or interface associated with the given session.
			Trig				= VI_EVENT_TRIG,				// Notifies the application that a trigger interrupt was received from the device. This may be either a hardware or software trigger, depending on the interface and the current session settings.
			Clear				= VI_EVENT_CLEAR,				// Notifies the application that a device clear message was sent to the local controller.
			IOCompletion		= VI_EVENT_IO_COMPLETION,		// Notifies the application that an asynchronous operation has completed.
			Exception			= VI_EVENT_EXCEPTION,			// Notifies the application that an error condition has occured during an operation invocation.
			GPIB_CIC			= VI_EVENT_GPIB_CIC,			// Notifies the application that the GPIB controller has gained or lost CIC (controller-in-charge) status.
			GPIB_TALK			= VI_EVENT_GPIB_TALK,			// Notifies the application that the GPIB controller was addressed to talk.
			GPIB_LISTEN			= VI_EVENT_GPIB_LISTEN,			// Notifies the application that the GPIB controller has been addressed to listen.
			VXI_VME_SYSFAIL		= VI_EVENT_VXI_VME_SYSFAIL,		// Notifies the application that the VXI/VME SYSFAIL* line was asserted.
			VXI_VME_SYSRESET	= VI_EVENT_VXI_VME_SYSRESET,	// Notifies the application that the VXI/VME SYSRESET* line was asserted.
			VXI_SIGP			= VI_EVENT_VXI_SIGP,			// Notifies the application that a VXIbus signal or VXIbus interrupt was received from the device associated with the given session.
			VXI_VME_INTR		= VI_EVENT_VXI_VME_INTR,		// Notifies the application that a VXIbus interrupt was received from the device associated with the given session.
			PXI_INTR			= VI_EVENT_PXI_INTR,			// Notifies the application that a PXI interrupt occurred.
			TCPIP_CONNECT		= VI_EVENT_TCPIP_CONNECT,		// Notifies the application that ...? No documentation...
			USB_INTR			= VI_EVENT_USB_INTR,			// Notifies the application that a USB interrupt has occurred.
			ALL					= VI_ALL_ENABLED_EVENTS
		};

		enum class eTriggerProtocol {
			Default			= VI_TRIG_PROT_DEFAULT,
			VXI_Default		= VI_TRIG_PROT_DEFAULT,
			VXI_On			= VI_TRIG_PROT_ON,
			VXI_Off			= VI_TRIG_PROT_OFF,
			VXI_Sync		= VI_TRIG_PROT_SYNC,
			PXI_Reserve		= VI_TRIG_PROT_RESERVE,
			PXI_Unreserve	= VI_TRIG_PROT_UNRESERVE
		};

		enum class eSerialBaudrate {		// Serial baudrate
			k110 = 110,
			k300 = 300,
			k600 = 600,
			k1200 = 1200,
			k2400 = 2400,
			k4800 = 4800,
			k9600 = 9600,
			k14400 = 14400,
			k19200 = 19200,
			k28800 = 28800,
			k38400 = 38400,
			k57600 = 57600,
			k115200 = 115200,
			k230400 = 230400,
			k460800 = 460800,
			k921600 = 921600
		};
		enum class eSerialDataBits {		// Number of data bits for the serial communication
			k7 = 7,
			k8 = 8
		};
		enum class eSerialParity {		// Parity configuration for the serial communication
			kNone	= VI_ASRL_PAR_NONE,		//set Parity to None
			kOdd	= VI_ASRL_PAR_ODD,		//set Parity to Odd
			kEven	= VI_ASRL_PAR_EVEN,		//set Parity to Even
			kMark	= VI_ASRL_PAR_MARK,		//set Parity to Mark
			kSpace	= VI_ASRL_PAR_SPACE		//set Parity to Space

		};
		enum class eSerialStopBits {	// Stop bit configuration for the serial communication
			k1		= VI_ASRL_STOP_ONE,		//set Stop bit to 1
			k1_5	= VI_ASRL_STOP_ONE5,	//set Stop bit to 1.5
			k2		= VI_ASRL_STOP_TWO		//set Stop bit to 2
		};
		enum class eSerialHandshake {		// Handshake configuration for the serial communication
			kNone			= VI_ASRL_FLOW_NONE,								//set None
			kXonXoff		= VI_ASRL_FLOW_XON_XOFF,							//set Software Handshake Xon/Xoff
			kHardware		= VI_ASRL_FLOW_RTS_CTS,								//set Hardware Handshake RTS/CTS
			kHardAndSoft	= (VI_ASRL_FLOW_XON_XOFF | VI_ASRL_FLOW_RTS_CTS)	//set Hardware & Software Handshake (Xon/Xoff & RTS/CTS)
		};

		enum class eSerialTermMode {		//Serial termination mode
			kEndNone = VI_ASRL_END_NONE,			//set termination character to None
			kEndLast = VI_ASRL_END_LAST_BIT,		//set termination character to Last bit
			kEndTermChar = VI_ASRL_END_TERMCHAR,	//set termination character to the one specified by VI_ATTR_TERMCHAR()
			kEndBreak = VI_ASRL_END_BREAK			//set termination character to Break (Only available in Write termination)
		};

		struct sSerialPortSettings {		// Sets serial port settings
			eSerialBaudrate		Baudrate;			// [bd]	Speed of the serial communication
			eSerialDataBits		DataBits;			// [-]	Number of transmitted bits per packet
			eSerialParity		Parity;				// [-]	Parity configuration of the serial communication
			eSerialStopBits		StopBits;			// [-]	Stop bit configuration of the serial communication
			eSerialHandshake	Handshake;			// [-]	Handshake configuration of the serial communication
			eSerialTermMode		ReadTermMode;		// [-]	Termination Read mode of the serial communication
			char				ReadTermChar;		// [-]	Specific termination character when readTermMode = kEndTermChar
			sSerialPortSettings()
				: Baudrate(eSerialBaudrate::k9600),
				DataBits(eSerialDataBits::k8),
				Parity(eSerialParity::kNone),
				StopBits(eSerialStopBits::k1),
				Handshake(eSerialHandshake::kNone), 
				ReadTermMode(eSerialTermMode::kEndTermChar),
				ReadTermChar('\n')
			{}
		};

		// For future use
		class VISAResourceInfo
		{
		public:
			typedef std::string		tClass;				// INSTR / RAW / ...
			typedef std::string		tExpandedName;
			typedef std::string		tAlias;				// If any

		private:
			eInterfaceType		m_Type;
			tInterfaceNumber	m_Number;
			tClass				m_Class;
			tExpandedName		m_ExpName;
			tAlias				m_Alias;
		public:
			VISAResourceInfo()
			{}
			VISAResourceInfo(eInterfaceType Type, tInterfaceNumber Number, tClass Class, tExpandedName Expname, tAlias Alias)
				: m_Type(Type), m_Number(Number), m_Class(Class), m_ExpName(Expname), m_Alias(Alias)
			{}
			const eInterfaceType & Type() const
			{
				return m_Type;
			}
			const tInterfaceNumber & Number() const
			{
				return m_Number;
			}
			const tClass & Class() const
			{
				return m_Class;
			}
			const tExpandedName & ExpName() const
			{
				return m_ExpName;
			}
			const tAlias & Alias() const
			{
				return m_Alias;
			}
		};

		struct sParsedResource
		{
			tResourceName		Name;
			VISAResourceInfo	Info;

			sParsedResource(tResourceName & rName, VISAResourceInfo & rInfo)
				: Name(rName), Info(rInfo)
			{}

			bool operator==(sParsedResource const & other) const
			{
				return (Name == other.Name);
			}
			bool operator!=(sParsedResource const & other) const

			{
				return !(*this == other);
			}
		};
		class CParsedResourceList : public std::vector<sParsedResource>
		{
		public:
			bool operator==(CParsedResourceList const & other)
			{
				if (other.size() != size())
					return false;
				for (CParsedResourceList::const_iterator thisit = begin(), otherit = other.begin(); thisit != end(); thisit++, otherit++)
				{
					if (*thisit != *otherit)
						return false;
				}
				return true;
			}
			bool operator!=(CParsedResourceList const & other)
			{
				if (other.size() != size())
					return true;
				for (CParsedResourceList::const_iterator thisit = begin(), otherit = other.begin(); thisit != end(); thisit++, otherit++)
				{
					if (*thisit != *otherit)
						return true;
				}
				return false;
			}
		};

}}	// namespace MTL::Instrument
