//--------------------------------------------------------------------------------------------------------
///	\file			MFC3045.cpp
///	\date			2022
///	\copyright (c)	Metrolab Technology S.A., Geneva, Switzerland (www.metrolab.com)
///	See the included file License Terms.pdf for the licensing conditions.
///	\brief			MFC3045 Metrolab Probe-Array Instrument Driver
///	Note:
///	This API is working with MFC3045 Firmware version 2.22 (february 2001)
///	The MFC - 3045 is controlled with a set of commands sent by a computer via a RS - 232C link. An
///	MFC - 3045 command consists of 3 ASCII characters (which are the command mnemonic) followed
///	either by a terminator or by a comma, a decimal number (the parameter) and a terminator.The
///	terminator consists of either the 2 characters "Carriage - Return and Line - Feed" <CR - LF> or of a
///	semicolon <;>. The semicolon terminator is used to send several commands on one line.
///	See MFC3045 manual (ver3.0) for more details on RS 232 commands.
///	The following commands are not supported by this driver : SMA,x ; UPD,x ; ERA
///	(In comment text the abreviations PA is used for the probe array, MU for the main unit)
//--------------------------------------------------------------------------------------------------------

// Standard includes
#include <locale>
#include <regex>
#include <iomanip>
#include <sstream>

//Personal includes
#include "MFC3045.h"
#include "OSDefines.h"
#include "Helpers.h"
#include "Exception.h"

// External tools includes
#include "date.h"

//------------------------------------------------------------------//
//							Definitions
//------------------------------------------------------------------//
#define DEBUG_MTL_INSTRUMENT_MFC3045					1
#define DEBUG_MTL_INSTRUMENT_MFC3045_ERRORS_ONLY		1
#if (defined(_DEBUG) && defined(DEBUG_MTL_INSTRUMENT_MFC3045) && DEBUG_MTL_INSTRUMENT_MFC3045)
#if (defined(DEBUG_MTL_INSTRUMENT_MFC3045_ERRORS_ONLY) && DEBUG_MTL_INSTRUMENT_MFC3045_ERRORS_ONLY)
#define MTL_INSTRUMENT_MFC3045_DEBUG_COUT(__X__)
#else
#define MTL_INSTRUMENT_MFC3045_DEBUG_COUT(__X__)		COUT(__X__)
#endif
#define MTL_INSTRUMENT_MFC3045_DEBUG_CERR(__X__)		CERR(__X__)
#else
#define MTL_INSTRUMENT_MFC3045_DEBUG_COUT(__X__)
#define MTL_INSTRUMENT_MFC3045_DEBUG_CERR(__X__)
#endif

#define SMA_PA		"PA\r\n"
#define SMA_UP		"UP\r\n"
#define SMA_DN		"DN\r\n"
#define SMA_EE		"EE\r\n"
#define SMA_RS		"RS\r\n"
#define SMA_ME		"ME\r\n"
#define SMA_CE		"CE\r\n"
#define SMA_DR		"DR\r\n"



using namespace MTL::Instrument;
using namespace MTL::Instrument::MFC3045Types;

//----------------------------------------------------------------------//
//								Exceptions								//
//----------------------------------------------------------------------//
typedef MTL::CException<CMFC3045Instrument>		CMFC3045InsException;


//----------------------------------------------------------------------//
//							Private Utilities							//
//----------------------------------------------------------------------//
// Configure the Host Serial Port (VISA) to communicate with MFC3045 Main Unit.
// Note: To be able to communicate, the host serial setting must be the same as the MFC3045 main unit serial settings.
//
bool CMFC3045Instrument::l_SerialPortHostConfig(const sSerialPortSettings & rSerialSettings)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!CVISAInstrument::ConfigSerialPort(rSerialSettings))
			throw CMFC3045InsException(std::string("Failed to configure serial port ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Clear (flush) RX and TX (VISA) Host serial Buffer and send a break
bool CMFC3045Instrument::l_ClearHostSerialBufferRxTx()
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try
	{
		if (!CVISAInstrument::Clear()) // clear the visa input and output buffer
			throw CMFC3045InsException("Could not clear RX TX Visa buffer", MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Get the number of unreaded byte in the Rx (VISA) buffer
bool CMFC3045Instrument::l_GetNewDataRxBufferLength(U32 & rRxBufferLength)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try
	{
		U32 l_rxBuffLength(0);

		if (!CVISAInstrument::GetAttribute(VI_ATTR_ASRL_AVAIL_NUM, &l_rxBuffLength)) // get visa input rx buffer length
			throw CMFC3045InsException("Could not get new data length in RX Visa buffer", MTL__LOCATION__);
		rRxBufferLength = l_rxBuffLength;
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		rRxBufferLength =0;
		return false;
	}
	return true;
}

// Wait on new data available in host RX (VISA) buffer
// The wait timeout is in millisecond. Each check of RX new data length is done every 10 ms
bool CMFC3045Instrument::l_WaitAvailableDataRX(const U32 & rTimeoutRx, U32 & rRXBufferLength, const bool & rErrorOnTimoute)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		rRXBufferLength = 0;
		auto l_EndTime = std::chrono::steady_clock::now() + std::chrono::milliseconds(rTimeoutRx);
		auto l_CurrentTime = std::chrono::steady_clock::now();

		for (U32 it = 0; (0 == rRXBufferLength && l_CurrentTime < l_EndTime); it++)
		{
			if (m_AbortSerialTransfert)
				throw CMFC3045InsException("Abort serial command received.", MTL__LOCATION__);
			MTL_SleepMs(10); //sleep for 10ms
			if (!l_GetNewDataRxBufferLength(rRXBufferLength)) // read number of available data in received buffer
				throw CMFC3045InsException("Failed to get data length", MTL__LOCATION__);
			l_CurrentTime = std::chrono::steady_clock::now();
		}
		if (0 == rRXBufferLength)
		{
			if(rErrorOnTimoute)
				throw CMFC3045InsException("Timeout reached on wait data RX buffer.", MTL__LOCATION__);
			else
				return false; // return false but without exception.
		}
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Extract any SMA (send message automatically) message string from an input string and set the according flag in SMAValue if required.
bool CMFC3045Instrument::l_ExtractSmaMsgFromDataRX(const std::string & rNewRxBufferStrIn, uSMAByte & rSMAValue, std::string & rNewRxBufferStrOut)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);
	rSMAValue.RawSMAB = 0;
	rNewRxBufferStrOut.clear();
	std::string l_StrTemp;
	try {
		if (0 == rNewRxBufferStrIn.size())
			throw CMFC3045InsException("Empty RX buffer.", MTL__LOCATION__);
		rNewRxBufferStrOut = rNewRxBufferStrIn;

		// find "DR\r\n" data ready message
		std::size_t l_Pos = rNewRxBufferStrOut.find(SMA_DR);
		if (std::string::npos != l_Pos)
		{
			rSMAValue.SMAByte.DRST1 = 1;
			rNewRxBufferStrOut.erase(l_Pos, 4);
		}

		// find "PA\r\n" Probe Array not conected message
		l_Pos = rNewRxBufferStrOut.find(SMA_PA);
		if (std::string::npos != l_Pos)
		{
			rSMAValue.SMAByte.PANST2 = 1;
			rNewRxBufferStrOut.erase(l_Pos, 4);
		}

		// find "CE\r\n" Command Error message
		l_Pos = rNewRxBufferStrOut.find(SMA_CE);
		if (std::string::npos != l_Pos)
		{
			rSMAValue.SMAByte.CEST1 = 1;
			rNewRxBufferStrOut.erase(l_Pos, 4);
		}

		// find "RS\r\n" Serial communication error message
		l_Pos = rNewRxBufferStrOut.find(SMA_RS);
		if (std::string::npos != l_Pos)
		{
			rSMAValue.SMAByte.RSST1 = 1;
			rNewRxBufferStrOut.erase(l_Pos, 4);
		}

		// find "ME\r\n" Modulation error message
		l_Pos = rNewRxBufferStrOut.find(SMA_ME);
		if (std::string::npos != l_Pos)
		{
			rSMAValue.SMAByte.MEST1 = 1;
			rNewRxBufferStrOut.erase(l_Pos, 4);
		}

		// find "EE\r\n" Miscelaneous (EEPROM) error message
		l_Pos = rNewRxBufferStrOut.find(SMA_EE);
		if (std::string::npos != l_Pos)
		{
			rSMAValue.SMAByte.EEST1 = 1;
			rNewRxBufferStrOut.erase(l_Pos, 4);
		}

		// find "DN\r\n" Remote Box button pressed message
		l_Pos = rNewRxBufferStrOut.find(SMA_DN);
		if (std::string::npos != l_Pos)
		{
			rSMAValue.SMAByte.DNRB = 1;
			rNewRxBufferStrOut.erase(l_Pos, 4);
		}

		// find "UP\r\n" Remote Box button released message
		l_Pos = rNewRxBufferStrOut.find(SMA_UP);
		if (std::string::npos != l_Pos)
		{
			rSMAValue.SMAByte.UPRB = 1;
			rNewRxBufferStrOut.erase(l_Pos, 4);
		}

	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Look for "PA\r\n" and "DR\r\n" SMA (send message automatically) message string from an input string and set the according flag in rSMAValue.
bool CMFC3045Instrument::l_ExtractPaDrSmaMsgFromDataRX(const std::string & rNewRxBufferStrIn, uSMAByte & rSMAValue, std::string & rNewRxBufferStrOut)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);
	rSMAValue.RawSMAB = 0;
	rNewRxBufferStrOut.clear();
	std::string l_StrTemp;
	try {
		if (0 == rNewRxBufferStrIn.size())
			throw CMFC3045InsException("Empty RX buffer.", MTL__LOCATION__);
		rNewRxBufferStrOut = rNewRxBufferStrIn;

		// find "DR\r\n" data ready message
		std::size_t l_Pos = rNewRxBufferStrOut.find(SMA_DR);
		if (std::string::npos != l_Pos)
		{
			rSMAValue.SMAByte.DRST1 = 1;
			rNewRxBufferStrOut.erase(l_Pos, 4);
		}


		// find "PA\r\n" Probe Array not conected message
		l_Pos = rNewRxBufferStrOut.find(SMA_PA);
		if (std::string::npos != l_Pos)
		{
			rSMAValue.SMAByte.PANST2 = 1;
			rNewRxBufferStrOut.erase(l_Pos, 4);
		}

	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write a string adding the termination char "\r\n"
bool CMFC3045Instrument::l_Write(const std::string & rWriteStr, const bool & rForceWrite)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);
	
	try
	{
		if(m_AbortSerialTransfert && !rForceWrite)
			throw CMFC3045InsException("Abort serial command received.", MTL__LOCATION__);
		// Write
		std::string l_WriteStr;
		l_WriteStr = rWriteStr + "\r\n"; // add termination char
		MTL_SleepMs(10); // Wait to let the MFC3045 to process any older Write command before to send any other one.
		if (!CVISAInstrument::Write(reinterpret_cast<ViBuf>(const_cast<char *>(l_WriteStr.data())), static_cast<ViUInt32>(l_WriteStr.size())))
			throw CMFC3045InsException("Could not write command " + l_WriteStr, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read N strings (with Terrm Char "\r\n") in the Serial Input Buffer and concatenate them.
// Termination character is specified by SerialPortInit() (default should be "\n")
// Termination char is remove if it is "\r\n" or End of transmission Block (0x17)
bool CMFC3045Instrument::l_ReadString(std::string & rRxBufferStr, const U8 & rNStringAssembly, const bool & rCheckSmaMsg)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try
	{
		U32 l_RxNewDataLength(0);
		rRxBufferStr.clear();
		std::string l_TempRdStrOut;
		uSMAByte l_SMAMsg;
		l_SMAMsg.RawSMAB =0;
		U8 l_NbSmaMsg = 0;

		// Read received N strings with term char and concatenate them
		for (size_t i = 0; i < rNStringAssembly + l_NbSmaMsg; i++)
		{
			if(!l_WaitAvailableDataRX(m_WaitTimeRdData_ms, l_RxNewDataLength)) // check for new availabled data in Rx Buffer before to read it. We have a time_out of m_WaitTimeRdData_ms before to abort waiting
				throw CMFC3045InsException("Failed to read data length in RX buffer or buffer is empty. In ", MTL__LOCATION__);
			m_ReadBuffer.clear();
			if (!CVISAInstrument::Read(m_ReadBuffer,false))
			{
				ViStatus l_VisaStatus = Status();
				if ((VI_ERROR_TMO == l_VisaStatus) && 
					(1 <= m_ReadBuffer.size()) && (static_cast<char>(DATA_END_ETB) == m_ReadBuffer[m_ReadBuffer.size() - 1]))
				{
					m_ReadBuffer.resize(m_ReadBuffer.size() - 1); // receveid End of transmission block character
					l_TempRdStrOut = std::string(m_ReadBuffer.begin(), m_ReadBuffer.end());
					break; // quit loop, but we do not consider it as an error!
				}
				else
					throw CMFC3045InsException("Could not read", MTL__LOCATION__);
			}
			else
			{
				if(m_SMAEnabled || rCheckSmaMsg)
				{
					l_ExtractSmaMsgFromDataRX(std::string(m_ReadBuffer.begin(), m_ReadBuffer.end()), l_SMAMsg, l_TempRdStrOut);
					m_SMAValue.RawSMAB |= l_SMAMsg.RawSMAB;
					if (l_SMAMsg.RawSMAB >0) // message received was an automatic sent message (SMA) from main unit
						l_NbSmaMsg ++; // allow to do one more "for loop" to read the expected message (other than SMA message)
				}
				else
				{
					l_TempRdStrOut = std::string(m_ReadBuffer.begin(), m_ReadBuffer.end());
				}
				if(0 != l_TempRdStrOut.size())
				{
					// Remove the term char "\r\n" or the end of transmission bloc char from the read string (cleaned from the SMA message if exists)
					if (l_TempRdStrOut.size() >= 2 && l_TempRdStrOut[l_TempRdStrOut.size() - 2] == '\r' && l_TempRdStrOut[l_TempRdStrOut.size() - 1] == '\n')
						l_TempRdStrOut.resize(l_TempRdStrOut.size() - 2);
					else if (1 <= l_TempRdStrOut.size() && static_cast<char>(DATA_END_ETB) == l_TempRdStrOut[l_TempRdStrOut.size() - 1])
					{
						l_TempRdStrOut.resize(l_TempRdStrOut.size() - 1);
					}
					else
						throw CMFC3045InsException("Read doesn't contain termination Chars", MTL__LOCATION__);
				}
			}
			rRxBufferStr.append(l_TempRdStrOut); // concatenate received string messages
		}
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read serial input buffer of size N-char without any End of Char termination mode.
// This mode is used to read EEPROM content received after a specific EEPROM command (SMU, SPA, CPS, ...)
bool CMFC3045Instrument::l_ReadNCharNoTerm(CVISABuffer & rReadBuffer, const U16 & rReadNChar)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try
	{		
		// Change the Read Termination Mode to None for EEPROM reading
		sSerialPortSettings l_SerialParmsNoTerm = m_HostSerialConfig;
		l_SerialParmsNoTerm.ReadTermMode = eSerialTermMode::kEndNone;
		l_SerialPortHostConfig(l_SerialParmsNoTerm); //Disable serial port termination.

		rReadBuffer.clear();
		ViUInt32 l_RetLength(0);
		ViUInt32 l_NbCharToRead = static_cast<ViUInt32>(rReadNChar) + 1; // add Last char "End of Transmission Block" to Char count
		U32 l_RxNewDataLength(0);

		if (!l_WaitAvailableDataRX(m_WaitTimeRdData_ms, l_RxNewDataLength)) // wait and check durring m_WaitTimeRdData_ms for new data available in Rx Buffer before to read it
			throw CMFC3045InsException("Failed to read data length in RX buffer, or buffer is empty. In ", MTL__LOCATION__);
		if (!CVISAInstrument::Read(reinterpret_cast<ViPBuf>(rReadBuffer.data()), l_NbCharToRead, l_RetLength))
		{
			throw CMFC3045InsException("Failed to Read EEPROM in ", MTL__LOCATION__);
		}
		rReadBuffer.resize(l_RetLength);
		if ((l_RetLength == l_NbCharToRead) && static_cast<char>(DATA_END_ETB) == rReadBuffer[l_RetLength - 1])
		{
			rReadBuffer.resize(l_RetLength - 1);//Remove last character "End of Transmission Block"
		}
		else
			throw CMFC3045InsException("Failed to Read EEPROM. Wrong expected data size read, in ", MTL__LOCATION__);

		// Set back the read termination mode as it was (with end character) & clear serial Rx Tx buffers
		l_SerialPortHostConfig(m_HostSerialConfig);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		l_SerialPortHostConfig(m_HostSerialConfig); // set back the read termination mode as it was (with end character)
		return false;
	}
	return true;
}

// Converts string date format dd/mm/yy (at UTC) to number representing the amount of seconds since 00:00:00, Jan 1 1970 UTC
bool CMFC3045Instrument::l_DateStrToDateTimeUtc(const std::string & rUtcDateStr, std::time_t & rUtcDateTime)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);

	try {
		std::match_results<std::string::const_iterator> l_DateMatch;
		// Match 3 groups of 2 digits seperated by "/" (String date format read is : dd/mm/yy)
		if (!std::regex_match(rUtcDateStr, l_DateMatch, std::regex("^([0-9]{2})/([0-9]{2})/([0-9]{2})")))
			throw CMFC3045InsException("Invalid date answer format", MTL__LOCATION__);

		std::stringstream l_DateStream;
		std::chrono::system_clock::time_point l_TimePointUTC;
		int l_TempDD, l_TempMM, l_TempYY;
		l_TempDD = std::stoi(l_DateMatch[1]); //dd
		l_TempMM = std::stoi(l_DateMatch[2]);	//mm
		l_TempYY = std::stoi(l_DateMatch[3]);	//yy

		// if date not set yet
		if (0 == l_TempDD || 0 == l_TempMM || 0 == l_TempYY)
			rUtcDateTime = 0; // set 0 correpsonding to 1970/01/01
		else
		{ //test if date format is valid
			if(l_TempDD < 1 || l_TempDD > 31)
				throw CMFC3045InsException("Invalid Day format", MTL__LOCATION__);

			if (l_TempMM < 0 || l_TempMM > 12)
				throw CMFC3045InsException("Invalid Month format", MTL__LOCATION__);

			if (l_TempYY >= 80) // if above than 80 we consider the date to be between 1980 to 1999
			{
				l_DateStream << "19" + l_DateMatch[3].str() + "-" + l_DateMatch[2].str() + "-" + l_DateMatch[1].str() + " 00:00:00";
			}
			else // we consider to be in year 2000 to 2079
			{
				l_DateStream << "20" + l_DateMatch[3].str() + "-" + l_DateMatch[2].str() + "-" + l_DateMatch[1].str() + " 00:00:00";
			}

			l_DateStream >> date::parse("%F %T", l_TimePointUTC); // parse string (UTC) date to time_point UTC
			rUtcDateTime = std::chrono::system_clock::to_time_t(l_TimePointUTC); // convert to time_t (UTC)
		}
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		rUtcDateTime = -1;
		return false;
	}
	return true;
}

// Converts UTC date (number representing the amount of seconds since 00:00, Jan 1 1970 UTC) to a string date format ddmmyy (UTC) 
// (as it must be when writing in CDP and CDU registers). Note: No "/" in string between day month year values when writing.
bool CMFC3045Instrument::l_DateTimeToDateStrUtc(const std::time_t & rUtcDateTime, std::string & rUtcDateStr)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);

	try {
		auto l_TimePointUTC = std::chrono::system_clock::from_time_t(rUtcDateTime);
		rUtcDateStr = date::format("%d%m%g", date::floor<std::chrono::seconds>(l_TimePointUTC)); // convert date UTC in ddmmy string format
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_Unused(rE)
		rUtcDateStr.clear();
		return false;
	}
	return true;
}

//------------------------------------------------------------------//
//					Constructors / Destructors
//------------------------------------------------------------------//
CMFC3045Instrument::CMFC3045Instrument(CVISAResourceManager & rRM, tResourceName Rsrc)
	: CVISAInstrument(rRM, Rsrc), 
	m_WaitTimeRdData_ms(RS_WAIT_DATARD_MS),
	m_AbortSerialTransfert(false),
	m_ReadBuffer(RS_RD_BUFF_DEFAULT_SIZE), 
	m_AdvancedMode(eAdvancedMode::kNone),		// default value in MFC3045 after reset
	m_DataBlockMode(eDataBlockMode::kOneByOne),	// default value in MFC3045 after reset
	m_NbProbesInPA(0),
	m_SMAEnabled(false)
{

}

CMFC3045Instrument::~CMFC3045Instrument()
{
	Abort();
	DisconnectInstrument();
}

//------------------------------------------------------------------//
//		Connect / Disconnect to Instrument with Serial Port (VISA)
//------------------------------------------------------------------//
// Check serial connection to instrument (Main Unit)
// Send a command to check MFC3045 main unit firmware version. Search "MFC" string in answer.
bool CMFC3045Instrument::CheckInstrumentConnection(const U8 & rNbRetry)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try
	{
		std::string l_Answer;
		bool l_TestOk(false);

		for (int i = 0; i < rNbRetry; i++)
		{
			if (!l_Write("VER"))
				throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
			if (l_ReadString(l_Answer,1,true))	// try to read the version information. Answer must start with "MFC"
			{
				if (std::string::npos != l_Answer.find(MFC_ASCII))
				{
					l_TestOk = true;
					break;
				}
			}
		}
		if (!l_TestOk)
			throw CMFC3045InsException(std::string("No MFC3045 Connected on Serial Port"), MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Connect to MFC3045 main unit instrument
// Open host serial port, check serial connection to MFC3045 main unit and 
// configure some main unit registers (AdvancedMode to level 0 and DataBlockMode to Decimal)
bool CMFC3045Instrument::ConnectInstrument(const sSerialPortSettings & rSerialSettings, const U32 & rInitialTimeout_ms)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try
	{
		// Open Host Serial Port and set Visa Timeout
		if (!CVISAInstrument::Open(eOpenAccessMode::ExclusiveLock, rInitialTimeout_ms))
			throw CMFC3045InsException("Failed to open VISA serial port", MTL__LOCATION__);
		if (!CVISAInstrument::SetTimeout(rInitialTimeout_ms))
			throw CMFC3045InsException("Failed to set VISA timeout", MTL__LOCATION__);
		// Acquire VISA lock
		if (!CVISAInstrument::LockExclusive(rInitialTimeout_ms))
			throw CMFC3045InsException("Failed acquiring VISA lock", MTL__LOCATION__);
		// Clear the VISA input and output serial buffer
		if (!CVISAInstrument::Clear())
			throw CMFC3045InsException("Failed to clear serial RX TX buffers", MTL__LOCATION__);
		
		// Configure (VISA) host serial port
		m_HostSerialConfig = rSerialSettings; // store current serial configuration
		m_HostSerialConfig.ReadTermMode = eSerialTermMode::kEndTermChar; // by default set end char mode to activ
		m_HostSerialConfig.ReadTermChar = '\n'; // by default set end char to '\n'
		if(!l_SerialPortHostConfig(m_HostSerialConfig))
			throw CMFC3045InsException("Failed configure serial port, in ", MTL__LOCATION__);

		//Check instrument connection
		m_WaitTimeRdData_ms = 2000; // for checking connection set x sec of waiting RX buffer to be field by any response.
		if (!CheckInstrumentConnection(3)) // try 3 times to communicate and verify presence of MFC3045 main unit
			throw CMFC3045InsException(std::string("No MFC3045 connected on serial port, in "), MTL__LOCATION__);

		m_WaitTimeRdData_ms = rInitialTimeout_ms; // for further use set wait time on RX buffer to receive data at the same value as VISA timeout

		U8 l_RegStat(0);
		if(!StatusGet(eStatusRegister::kStatus3, l_RegStat)) // Read status register 3 doesn't clear bits
			throw CMFC3045InsException(std::string("Failed to read Status 3, in"), MTL__LOCATION__);
		if (l_RegStat > 1) //measurement may be in progress.
		{
			if (!BreakMeasSet())
				throw CMFC3045InsException(std::string("Failed to send Break command, in "), MTL__LOCATION__);
		}
		// Configure instrument
		if (!DisableSMA()) // Disable SMA (already default state in MFC3045) . We do not want to use it in normal case to avoid conflict during a read command
			throw CMFC3045InsException(std::string("Failed to disable MFC3045 SMA, in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!AdvancedCommandModeSet(eAdvancedMode::kNone))	// Set Advanced Command Mode register to 0 (kNone)
			throw CMFC3045InsException(std::string("Failed to set MFC3045 Advanced Mode to 0, in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!DataBlockModeDecimalSet()) // Set Data transfert block mode to "Decimal"
			throw CMFC3045InsException(std::string("Failed to set MFC3045 Decimal Data Block Mode, in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		// Try to release the VISA lock
		CVISAInstrument::Unlock();
		// Try to close the VISA serial connection
		CVISAInstrument::Close();
		return false;
	}
	return true;
}

// Disconnect instrument => unlock and close (VISA) Serial Port from main Unit
bool CMFC3045Instrument::DisconnectInstrument()
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	bool l_Success(true);
	// Try to clear the VISA input and output serial buffer
	CVISAInstrument::Clear();
	// Try to release the VISA lock
	if (!CVISAInstrument::Unlock())
		l_Success = false;
	// Try to close the VISA serial connection
	CVISAInstrument::Close();
	return l_Success;
}

//------------------------------------------------------------------------//
//						Special public utilities
//------------------------------------------------------------------------//
// Abort any serial command and send a break acquisition to instrument.
bool CMFC3045Instrument::Abort()
{
	bool AbortSucceeded = true;
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);
	m_AbortSerialTransfert = true;
	if(!BreakMeasSet()) // the only write cmd allowed.
		AbortSucceeded = false;
	MTL_SleepMs(200);
	m_AbortSerialTransfert = false;
	return AbortSucceeded;
}

// Read a register with a custom command String (see MFC3045 Manual for correct command syntax).
// The read register must be only one single string. (termination \r\n is automatically removed)
bool CMFC3045Instrument::ReadWithCustomCmd(const std::string & rCommandStr, std::string & rRegValStr)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_Write(rCommandStr))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(rRegValStr))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Get maximum time in ms to wait before expecting new data in RX buffer to read
void CMFC3045Instrument::GetCurrentTimeoutRdDataMs(U32 & rRdDataTimeout_ms)
{
	rRdDataTimeout_ms = m_WaitTimeRdData_ms;
}

// Set maximum time in ms to wait before expecting new data in RX buffer to read
void CMFC3045Instrument::SetTimeoutRdNewDataMs(const U32 & rNewdataTimeout_ms)
{
	m_WaitTimeRdData_ms = rNewdataTimeout_ms;
}

//------------------------------------------------------------------//
//				RS232 Settings
//------------------------------------------------------------------//
// Parse Serial RS232 Parmameters from a sSerialPortSettings to RSP register format
bool CMFC3045Instrument::ParseParmForRSP(const sSerialPortSettings & rSerialSettings, U16 & rParmRSPSettings)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		uRSPByte l_Setting;
		l_Setting.RawRSPB = 0;

		switch (rSerialSettings.DataBits) // Data bit (bit 0 of RSP)
		{
		case eSerialDataBits::k7:
			l_Setting.RSPByte.PBL8 = 0;
			break;
		case eSerialDataBits::k8:
			l_Setting.RSPByte.PBL8 = 1;
			break;
		default:
			throw CMFC3045InsException(std::string("Failed to parse Data bit value in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		}

		switch (rSerialSettings.StopBits)	//Stop bit (bit 1 of RSP). Value 1 or 2 only; 1.5 not supported
		{
		case eSerialStopBits::k1:
			l_Setting.RSPByte.PSB2 = 0;
			break;
		case eSerialStopBits::k2:
			l_Setting.RSPByte.PSB2 = 1;
			break;
		default:
			throw CMFC3045InsException(std::string("Failed to parse Stop_bit value in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		}

		switch (rSerialSettings.Parity) // Parity (bit 2 and 3 of RSP). Only None, Odd, Even Parity supported.
		{
		case eSerialParity::kNone:
			l_Setting.RSPByte.PPAS = 0;
			break;
		case eSerialParity::kEven:
			l_Setting.RSPByte.PPAS = 1;
			l_Setting.RSPByte.PPAE = 1;
			break;
		case eSerialParity::kOdd:
			l_Setting.RSPByte.PPAS = 1;
			l_Setting.RSPByte.PPAE = 0;
			break;
		default:
			throw CMFC3045InsException(std::string("Failed to parse Parity value in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		}

		switch (rSerialSettings.Handshake) // Handshake (bit 4 and 5 of RSP)
		{
		case eSerialHandshake::kNone:
			l_Setting.RSPByte.PSHE = 0;
			l_Setting.RSPByte.PHHE = 0;
			break;
		case eSerialHandshake::kXonXoff: // Software
			l_Setting.RSPByte.PSHE = 1;
			l_Setting.RSPByte.PHHE = 0;
			break;
		case eSerialHandshake::kHardware: //RTS/CTS
			l_Setting.RSPByte.PSHE = 0;
			l_Setting.RSPByte.PHHE = 1;
			break;
		case eSerialHandshake::kHardAndSoft: // RTS/CTS & Xon/Xoff
			l_Setting.RSPByte.PSHE = 1;
			l_Setting.RSPByte.PHHE = 1;
			break;
		default:
			throw CMFC3045InsException(std::string("Failed to parse Handshake value in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		}

		switch (rSerialSettings.Baudrate) // Baud Rate (bit 8-10 of RSP)
		{
		case eSerialBaudrate::k2400:
			l_Setting.RSPByte.PRBR = static_cast<U8>(eMFCSerialBDR::k2400);
			break;
		case eSerialBaudrate::k4800:
			l_Setting.RSPByte.PRBR = static_cast<U8>(eMFCSerialBDR::k4800);
			break;
		case eSerialBaudrate::k9600:
			l_Setting.RSPByte.PRBR = static_cast<U8>(eMFCSerialBDR::k9600);
			break;
		case eSerialBaudrate::k19200:
			l_Setting.RSPByte.PRBR = static_cast<U8>(eMFCSerialBDR::k19200);
			break;
		case eSerialBaudrate::k28800:
			l_Setting.RSPByte.PRBR = static_cast<U8>(eMFCSerialBDR::k28800);
			break;
		case eSerialBaudrate::k38400:
			l_Setting.RSPByte.PRBR = static_cast<U8>(eMFCSerialBDR::k38400);
			break;
		case eSerialBaudrate::k57600:
			l_Setting.RSPByte.PRBR = static_cast<U8>(eMFCSerialBDR::k57600);
			break;
		case eSerialBaudrate::k115200:
			l_Setting.RSPByte.PRBR = static_cast<U8>(eMFCSerialBDR::k115200);
			break;
		default:
			throw CMFC3045InsException(std::string("Failed to parse Baud Rate value in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		}
		rParmRSPSettings = l_Setting.RawRSPB;
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Parse Serial RS232 Parmameters from RSP register to a sSerialPortSettings
bool CMFC3045Instrument::ParseParmFromRSP(const U16 & rParmRSPSettings, sSerialPortSettings & rSerialSettings)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		uRSPByte l_ParmSetting;
		sSerialPortSettings l_ParsedSetting;

		l_ParmSetting.RawRSPB = rParmRSPSettings;
		l_ParsedSetting = rSerialSettings; // preserve COM Port number, End Term Char if already stored

		if (l_ParmSetting.RSPByte.PBL8) // Data bit (bit 0 of RSP)
			l_ParsedSetting.DataBits = eSerialDataBits::k8;
		else
			l_ParsedSetting.DataBits = eSerialDataBits::k7;

		if (l_ParmSetting.RSPByte.PSB2) // Stop bit (bit 1 of RSP). Value 1 or 2 only; 1.5 not supported
			l_ParsedSetting.StopBits = eSerialStopBits::k2;		// 1 => 2 stop bit
		else
			l_ParsedSetting.StopBits = eSerialStopBits::k1;		// 0 => 1 stop bit

		if (l_ParmSetting.RSPByte.PPAS) // Parity (bit 2 and 3 of RSP). Only None, Odd, Even Parity supported.
		{
			if (l_ParmSetting.RSPByte.PPAE)
				l_ParsedSetting.Parity = eSerialParity::kEven;
			else
				l_ParsedSetting.Parity = eSerialParity::kOdd;
		}
		else
			l_ParsedSetting.Parity = eSerialParity::kNone;

		if (l_ParmSetting.RSPByte.PSHE) // Handshake (bit 4 of RSP)
		{
			if (l_ParmSetting.RSPByte.PHHE)
				l_ParsedSetting.Handshake = eSerialHandshake::kHardAndSoft;
			else
				l_ParsedSetting.Handshake = eSerialHandshake::kXonXoff;
		}
		else
		{
			if (l_ParmSetting.RSPByte.PHHE)
				l_ParsedSetting.Handshake = eSerialHandshake::kHardware;
			else
				l_ParsedSetting.Handshake = eSerialHandshake::kNone;
		}

		switch (static_cast<eMFCSerialBDR>(l_ParmSetting.RSPByte.PRBR)) // Baud Rate (bit 8-10 of RSP)
		{
		case eMFCSerialBDR::k2400:
			l_ParsedSetting.Baudrate = eSerialBaudrate::k2400;
			break;
		case eMFCSerialBDR::k4800:
			l_ParsedSetting.Baudrate = eSerialBaudrate::k4800;
			break;
		case eMFCSerialBDR::k9600:
			l_ParsedSetting.Baudrate = eSerialBaudrate::k9600;
			break;
		case eMFCSerialBDR::k19200:
			l_ParsedSetting.Baudrate = eSerialBaudrate::k19200;
			break;
		case eMFCSerialBDR::k28800:
			l_ParsedSetting.Baudrate = eSerialBaudrate::k28800;
			break;
		case eMFCSerialBDR::k38400:
			l_ParsedSetting.Baudrate = eSerialBaudrate::k38400;
			break;
		case eMFCSerialBDR::k57600:
			l_ParsedSetting.Baudrate = eSerialBaudrate::k57600;
			break;
		case eMFCSerialBDR::k115200:
			l_ParsedSetting.Baudrate = eSerialBaudrate::k115200;
			break;
		default:
			throw CMFC3045InsException(std::string("Failed to parse Baud Rate value in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		}

		rSerialSettings = l_ParsedSetting;
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register RSP (RS 232 Parameters). Command {RSP,x}, with x = deciaml value (see manual)
// Set main unit RS232 communication settings. New RS232 settings are active only if saved in main unit EEPROM (use MainUnitEepromWrite() method), 
// and after a main unit reset (use SystemResetSet() method).
bool CMFC3045Instrument::SerialRSPSet(const U16 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command("RSP,");

		if (eAdvancedMode::k2 != m_AdvancedMode)
			throw CMFC3045InsException(std::string("RSP Set not allowed. Advanced mode < 2 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if(rRegVal < RSP_MIN || rRegVal > RSP_MAX)
			throw CMFC3045InsException(std::string("RSP Set not allowed. Value out of range. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		l_Command += std::to_string(rRegVal);
		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register RSP (RS 232 Parameters). Command [RSP]
// Get main unit RS232 communication settings
bool CMFC3045Instrument::SerialRSPGet(U16 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_RdStr;
		if (eAdvancedMode::kNone == m_AdvancedMode)
			throw CMFC3045InsException(std::string("RSP Get not allowed. Advanced mode < 1 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_Write("RSP"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_RdStr, 2)) // read and concatenate 2 strings of 8 bit with therminal char \r\n
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		rRegVal = static_cast<U16>(std::stoul(l_RdStr,0,2));
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

//------------------------------------------------------------------//
//				Miscellaneous Commands
//------------------------------------------------------------------//
// Write register ADV (ADVanced commands). Command ADV,x with x = 0 or 1 or 2
// Set advanced command mode level and update m_AdvancedMode attribute.
bool CMFC3045Instrument::AdvancedCommandModeSet(const eAdvancedMode & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command("ADV,");

		l_Command += std::to_string(static_cast<U8>(rRegVal));

		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		m_AdvancedMode = rRegVal;
	}
	catch (CMFC3045InsException & rE)
	{
		m_AdvancedMode = eAdvancedMode::kNone;
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register ADV (ADVanced commands). Command ADV
// Get advanced command mode level and update m_AdvancedMode attribute.
bool CMFC3045Instrument::AdvancedCommandModeGet(eAdvancedMode & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		U32 l_AdvancedMode(255);
		std::string l_RdStr;
		if (!l_Write("ADV"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_RdStr))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		l_AdvancedMode = static_cast<U32>(std::stoul(l_RdStr));
		switch (l_AdvancedMode)
		{
		case static_cast<U32>(eAdvancedMode::kNone):
			rRegVal = eAdvancedMode::kNone;
			break;
		case static_cast<U32>(eAdvancedMode::k1):
			rRegVal = eAdvancedMode::k1;
			break;
		case static_cast<U32>(eAdvancedMode::k2):
			rRegVal = eAdvancedMode::k2;
			break;
		default:
			MTL_Assert("UnHandled value Type");
			throw CMFC3045InsException(std::string("Failed: Get wrong value for ADV in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
			break;
		} 
		m_AdvancedMode = rRegVal; // store Advanced Mode result
	}
	catch (CMFC3045InsException & rE)
	{
		m_AdvancedMode = eAdvancedMode::kNone; // store as if it was the most restrictive one
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register RST (ReSeT the system). Command RST
// Reset the main unit
bool CMFC3045Instrument::SystemResetSet()
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_Write("RST"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		m_AdvancedMode = eAdvancedMode::kNone;
		m_DataBlockMode = eDataBlockMode::kOneByOne;
		l_ClearHostSerialBufferRxTx(); // try to clear RX TX buffers
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register VER (VERsion). Command VER
// Get the main unit MFC3045 firmware version information
bool CMFC3045Instrument::VersionGet(std::string & rVersionInfo, F32 & rFwVersion)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_Write("VER"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(rVersionInfo))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		if (rVersionInfo.size() > 30)
			rFwVersion = static_cast<F32>(std::stof(std::string(rVersionInfo.end() - 4, rVersionInfo.end()).c_str()));
		else
			rFwVersion = 0.0;

	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register VER (VERsion). Command VER,x with x= 0 or 1 or 2
// Get MFC3045 firmware version (x=0), or Main unit EEPROM version (x=1), or Probe-Array version (x=2)
bool CMFC3045Instrument::VersionGet(std::string & rVersionInfo, F32 & rFwVersion, const eWhoVersion & rSelect)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command("VER,");

		l_Command += std::to_string(static_cast<U8>(rSelect));

		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(rVersionInfo))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		// Parse version number
		rFwVersion = 0.0;
		switch (rSelect)
		{
		case eWhoVersion::kMUFirmVer:
			if (rVersionInfo.size() > 30)
				rFwVersion = static_cast<F32>(std::stof(std::string(rVersionInfo.end() - 4, rVersionInfo.end()).c_str()));
			break;
		case eWhoVersion::kMUEepromVer:
		case eWhoVersion::kPAEepromVer:
			if (rVersionInfo.size() > 0)
				rFwVersion = static_cast<F32>(std::stof(rVersionInfo));
			break;
		default:
			MTL_Assert(false);
			break;
		}
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register S/N (Serial Number). Command S/N
// Get the main unit serial number
bool CMFC3045Instrument::SerialNumberGet(std::string & rSerialNum)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_Write("S/N"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(rSerialNum))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register S/N (Serial Number). Command S/N,x with x = 0 or 1 or 2
// Get serial number of main unit (x=0,1) or probe-array (x=2)
bool CMFC3045Instrument::SerialNumberGet(std::string & rSerialNum, const eWhoSerialNum & rSelect)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command("S/N,");

		l_Command += std::to_string(static_cast<U8>(rSelect));

		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(rSerialNum))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register LED (Remote box busy led). Command LED,x with x = 0 to 5 (see manual)
// Set the state of the remote box busy led.
bool CMFC3045Instrument::RemoteBusyLedSet(const eRemoteBusyLed & rRemoteLed)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command("LED,");

		l_Command += std::to_string(static_cast<U8>(rRemoteLed));

		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register LED (Remote box busy led). Command LED
// Get the state of the remote box busy led
bool CMFC3045Instrument::RemoteBusyLedGet(eRemoteBusyLed & rRemoteLed)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		U8 l_RemoteLed(255);
		std::string l_RdStr;
		if (!l_Write("LED"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_RdStr))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		l_RemoteLed = static_cast<U8>(std::stoul(l_RdStr));
		switch (l_RemoteLed)
		{
		case static_cast<U8>(eRemoteBusyLed::kRemoteLedOff):
			rRemoteLed = eRemoteBusyLed::kRemoteLedOff;
			break;
		case static_cast<U8>(eRemoteBusyLed::kRemoteLedMeasuring):
			rRemoteLed = eRemoteBusyLed::kRemoteLedMeasuring;
			break;
		case static_cast<U8>(eRemoteBusyLed::kRemoteLedOffEndMeas):
			rRemoteLed = eRemoteBusyLed::kRemoteLedOffEndMeas;
			break;
		case static_cast<U8>(eRemoteBusyLed::kRemoteLedOn):
			rRemoteLed = eRemoteBusyLed::kRemoteLedOn;
			break;
		case static_cast<U8>(eRemoteBusyLed::kRemoteLedBlinkSlow):
			rRemoteLed = eRemoteBusyLed::kRemoteLedBlinkSlow;
			break;
		case static_cast<U8>(eRemoteBusyLed::kRemoteLedBlinkFast):
			rRemoteLed = eRemoteBusyLed::kRemoteLedBlinkFast;
			break;
		default:
			MTL_Assert("UnHandled value Type");
			throw CMFC3045InsException(std::string("Failed: Get wrong value for LED in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
			break;
		}
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register DFF (Dds Fixed Frequency). Command [DFF,x] with x = decimal value in dHz
// Set DDS fixed frequency value (dHz) for the un-modulated RF signal output.
// A command DFF,0 or BRK stops the DDS frequecncy RF output
bool CMFC3045Instrument::DdsFrequencySet(const U32 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command("DFF,");

		if (eAdvancedMode::kNone == m_AdvancedMode)
			throw CMFC3045InsException(std::string("DFF Set not allowed. Advanced mode < 1 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (rRegVal > FSW_DFF_MAX)
			throw CMFC3045InsException(std::string("Failed, DDS Fix Freq out of Range in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		else
			l_Command += std::to_string(rRegVal);
		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register DFF (Dds Fixed Frequency). Command [DFF]
// Get the DDS fixed frequency unmodulated RF signal output in dHz
bool CMFC3045Instrument::DdsFrequencyGet(U32 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_RdStr;

		if (eAdvancedMode::kNone == m_AdvancedMode)
			throw CMFC3045InsException(std::string("DFF Set not allowed. Advanced mode < 1 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		if (!l_Write("DFF"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_RdStr))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		rRegVal = static_cast<U32>(std::stoul(l_RdStr));
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register FSW (Frequency Sweep). Command [FSW,x] with x = -1,0,1 or decimal value in dHz >1
// Set the modulated central frequency sweep shift value (see manual)
// The attribute rFreqStepVal is the frequency step value in dHz to set if mode eSweepFreqMode::kSetValue is choosed, otherwise it's not used.
bool CMFC3045Instrument::FrequencySweepSet(const eSweepFreqMode & rSweepMode, const U32 & rFreqStepVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command("FSW,");

		if (eAdvancedMode::kNone == m_AdvancedMode)
			throw CMFC3045InsException(std::string("FSW Set not allowed. Advanced mode < 1 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		switch (rSweepMode)
		{
		case eSweepFreqMode::kdisable:
			l_Command += "0";
			break;
		case eSweepFreqMode::k20Step:
			l_Command += "1";
			break;
		case eSweepFreqMode::kPaRange:
			l_Command += "-1";
			break;
		case eSweepFreqMode::kSetValue:
			if (rFreqStepVal < 2  || rFreqStepVal > FSW_DFF_MAX)
				throw CMFC3045InsException(std::string("Failed, FreqSweep out of Range in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
			else
				l_Command += std::to_string(rFreqStepVal);
			break;
		default:
			MTL_Assert(false);
			break;
		}
		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Shift the modulated central frequency up or down after a command FSW,x
bool CMFC3045Instrument::FrequencySweepShiftSet(const eShiftFreq & rShiftDirection)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command;
		l_Command += std::to_string(static_cast<U8>(rShiftDirection));
		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

//----------------------------------------------------------------------//
//						Status Handling
//----------------------------------------------------------------------//
// Read register ST1 to ST6 (Status). Command ST1 ST2 to ST6
// Get one Status Register
bool CMFC3045Instrument::StatusGet(const eStatusRegister & rStatRegSel, U8 & rStatus)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	std::string l_Command;

	// Prepare status register command
	switch (rStatRegSel)
	{
	case eStatusRegister::kStatus1:
		l_Command += "ST1";		// Note: Status Bits are cleared after reading
		break;
	case eStatusRegister::kStatus2:
		l_Command += "ST2";		// Note: Status Bits are cleared after reading
		break;
	case eStatusRegister::kStatus3:
		l_Command += "ST3";
		break;
	case eStatusRegister::kStatus4:
		l_Command += "ST4";		// Note: Status Bits 2,4,7 are cleared after reading
		break;
	case eStatusRegister::kStatus5:
		l_Command += "ST5";		// Note: Status Bits 3 to 7 are cleared after reading
		break;
	case eStatusRegister::kStatus6:
		l_Command += "ST6";
		break;
	default:
		MTL_Assert(false);
		break;
	}
	try {	
		//Write cmd
		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		//Reads Status register
		std::string l_RdStr;
		if (!l_ReadString(l_RdStr))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		rStatus = static_cast<U8>(std::stoul(l_RdStr, nullptr, 2));	//String Binary results convert to decimal
	} // try
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		rStatus = 0;
		return false;
	}
	return true;
}

// Read register ST1 to ST6 (Status). Command ST1;ST2; to ST6
// Get several Status Registers
bool CMFC3045Instrument::StatusGet(const std::vector<eStatusRegister> & rStatRegSel, std::vector<U8> & rStatus)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	rStatus.clear();
	std::string l_Command;

	for (std::vector<eStatusRegister>::const_iterator it = rStatRegSel.begin(); it != rStatRegSel.end(); it++)
	{
		if (it != rStatRegSel.begin())
			l_Command += ";";
		switch (*it)
		{
		case eStatusRegister::kStatus1:
			l_Command += "ST1";
			break;
		case eStatusRegister::kStatus2:
			l_Command += "ST2";
			break;
		case eStatusRegister::kStatus3:
			l_Command += "ST3";
			break;
		case eStatusRegister::kStatus4:
			l_Command += "ST4";
			break;
		case eStatusRegister::kStatus5:
			l_Command += "ST5";
			break;
		case eStatusRegister::kStatus6:
			l_Command += "ST6";
			break;
		default:
			MTL_Assert(false);
			break;
		}
	}
	try {
		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		//Reads N Status register
		std::string l_RdStr;
		for (size_t i = rStatRegSel.size(); i > 0; i--)
		{
			if(!l_ReadString(l_RdStr))
				throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
			rStatus.push_back(static_cast<U8>(std::stoul(l_RdStr, nullptr, 2)));	//String Binary results convert to decimal
		}
	} // try
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read several status register from ST1 to ST6. Register command reading is send one by one to instrument
// Get several Status Registers using bit selection and returning status Register in srtuct sStatusAllReg
bool CMFC3045Instrument::StatusGet(const uStatusRegBitSel & rStatRegBitSel, sStatusAllReg & rSomeStatusReg)
{
	try
	{
		if(rStatRegBitSel.StRegSelBit.ST1 || rStatRegBitSel.StRegSelBit.ALLST > 0)
			if(!StatusGet(eStatusRegister::kStatus1, rSomeStatusReg.StReg1.RawSTB1))
				throw CMFC3045InsException(std::string("Unable to read Status register 1") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		if(rStatRegBitSel.StRegSelBit.ST2 || rStatRegBitSel.StRegSelBit.ALLST > 0)
		{
			if(!StatusGet(eStatusRegister::kStatus2, rSomeStatusReg.StReg2.RawSTB2))
				throw CMFC3045InsException(std::string("Unable to read Status register 3") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		}
		if(rStatRegBitSel.StRegSelBit.ST3 || rStatRegBitSel.StRegSelBit.ALLST > 0)
		{
			if(!StatusGet(eStatusRegister::kStatus3, rSomeStatusReg.StReg3.RawSTB3))
				throw CMFC3045InsException(std::string("Unable to read Status register 3") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		}
		if(rStatRegBitSel.StRegSelBit.ST4 || rStatRegBitSel.StRegSelBit.ALLST > 0)
		{
			if(!StatusGet(eStatusRegister::kStatus4, rSomeStatusReg.StReg4.RawSTB4))
				throw CMFC3045InsException(std::string("Unable to read Status register 4") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		}
		if(rStatRegBitSel.StRegSelBit.ST5 || rStatRegBitSel.StRegSelBit.ALLST > 0)
		{
			if(!StatusGet(eStatusRegister::kStatus5, rSomeStatusReg.StReg5.RawSTB5))
				throw CMFC3045InsException(std::string("Unable to read Status register 5") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		}
		if(rStatRegBitSel.StRegSelBit.ST6 || rStatRegBitSel.StRegSelBit.ALLST > 0)
		{
			if(!StatusGet(eStatusRegister::kStatus6, rSomeStatusReg.StReg6.RawSTB6))
				throw CMFC3045InsException(std::string("Unable to read Status register 6") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		}
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	catch (...)
	{
		CMFC3045InsException  l_Error(std::string("Unhandled exception type") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(l_Error.what());
		return false;
	}
	return true;
}

// Read register ERR (ERRor). Command ERR
// Get 3 first caracters acccording to the Last Command Error (see also bit 1 of Status1)
bool CMFC3045Instrument::LastCmdErrorGet(std::string & rCmdError)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_Write("ERR"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(rCmdError))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register SMA (Send Message Automatically). Command SMA
// Get SMA register state
bool CMFC3045Instrument::SMAGet(U8 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_RdStr;
		if (!l_Write("SMA"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_RdStr))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		rRegVal = static_cast<U8>(std::stoul(l_RdStr, nullptr, 2)); //String Binary results convert to decimal
		if(rRegVal)
			m_SMAEnabled = 1;
		else
			m_SMAEnabled = 0;
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Disable SMA (Send Message Automatically) mode and clear m_SMAValue message
bool CMFC3045Instrument::DisableSMA()
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		U32 l_RxNewDataLength(0);
		U32 l_NbSmaMsg(0);
		uSMAByte l_SMARegValRead;
		l_SMARegValRead.RawSMAB = 0;
		bool l_DoWork(true);

		if(0 == m_SMAEnabled)
		{
			if (!SMAGet(l_SMARegValRead.RawSMAB)) //check that SMA is really disabled in MFC3045 Main unit
				throw CMFC3045InsException(std::string("Failed to read back SMA register in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
			if(0 == l_SMARegValRead.RawSMAB)
			{	//SMA is already disabled
				l_DoWork = false;
				m_SMAValue.RawSMAB = 0; // clear any previous SMA message received
			}
			l_SMARegValRead.RawSMAB = 0;
		}

		if (l_DoWork)
		{
			if (!l_SMASet(0))
				throw CMFC3045InsException(std::string("Failed to disable SMA in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
			if(l_WaitAvailableDataRX(100, l_RxNewDataLength, false)) // wait and check during 100ms and for any new data available in Rx Buffer before to read it
			{
				if (l_RxNewDataLength >= 4)
					l_NbSmaMsg = l_RxNewDataLength / 4; // size per SMA message (= 2 ascii char + 2 term char)
				for(size_t l_i = 0; l_i < l_NbSmaMsg ; l_i++)// purge any SMA message received
				{
					m_ReadBuffer.clear();
					CVISAInstrument::Read(m_ReadBuffer,false); // read any other SMA message received before SMA has been really disabled.
				}
			}
			if (!SMAGet(l_SMARegValRead.RawSMAB)) //check that SMA is correctly disabled in MFC3045 instrument
				throw CMFC3045InsException(std::string("Failed to read back SMA register in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
			if(0 != l_SMARegValRead.RawSMAB)
				throw CMFC3045InsException(std::string("Fail, SMA is not disabled") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
			m_SMAEnabled = 0; //SMA is disabled
			m_SMAValue.RawSMAB = 0; // clear any previous SMA message received
		}
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Set SMA (Send Message Automatically) mode
bool CMFC3045Instrument::l_SMASet(const U8 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command("SMA,");
		l_Command += std::to_string(rRegVal);
		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if(rRegVal)
			m_SMAEnabled = 1;
		else
			m_SMAEnabled = 0;
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

//----------------------------------------------------------------------//
//						Modulation Parameters                           //
//----------------------------------------------------------------------//
// Write register MDA (Modulation Amplitude). Command MDA,x with x = value in ppm
// Set the total amplitude peak peak value of freaquency modulation in ppm (referenced to PCF register : PA central freq)
bool CMFC3045Instrument::ModulationAmplitudeSet(const U32 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command("MDA,");
		if ((rRegVal < MDA_MIN) || (rRegVal > MDA_MAX))
			throw CMFC3045InsException(std::string("Failed, MDA out of Range in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		else
			l_Command += std::to_string(rRegVal);
		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register MDA (Modulation Amplitude). Command MDA
// Get the total amplitude peak peak value of freaquency modulation in ppm (referenced to PCF register : PA central freq and not MCF!)
bool CMFC3045Instrument::ModulationAmplitudeGet(U32 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_RdStr;
		if (!l_Write("MDA"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_RdStr))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		rRegVal = static_cast<U32>(std::stoul(l_RdStr));
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register MCF (Modulation Central Frequency). Command MCF,x with x = value in dHz
// Set the central frequency of modulation signal in dHz
bool CMFC3045Instrument::ModulationCentralFreqSet(const U32 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command("MCF,");
		if ((rRegVal < MOD_FREQ_MIN_DHZ) || (rRegVal > MOD_FREQ_MAX_DHZ))
			throw CMFC3045InsException(std::string("Failed, MCF out of Range in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		else
			l_Command += std::to_string(rRegVal);
		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register MCF (Modulation Central Frequency). Command MCF
// Get the central frequency of modulation signal in dHz
bool CMFC3045Instrument::ModulationCentralFreqGet(U32 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_RdStr;
		if (!l_Write("MCF"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_RdStr))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		rRegVal = static_cast<U32>(std::stoul(l_RdStr));
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register MLF (Modulation Lowest Frequency). Command MLF,x with x = value in dHz
// Set the lowest (bottom plateau) frequency of modulation signal in dHz
bool CMFC3045Instrument::ModulationLowestFreqSet(const U32 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command("MLF,");
		if ((rRegVal < MOD_FREQ_MIN_DHZ) || (rRegVal > MOD_FREQ_MAX_DHZ))
			throw CMFC3045InsException(std::string("Failed, MLF out of Range in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		else
			l_Command += std::to_string(rRegVal);
		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register MLF (Modulation Lowest Frequency). Command MLF
// Get the lowest (bottom plateau) frequency of modulation signal in dHz
bool CMFC3045Instrument::ModulationLowestFreqGet(U32 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_RdStr;
		if (!l_Write("MLF"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_RdStr))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		rRegVal = static_cast<U32>(std::stoul(l_RdStr));
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}


// Write register MHF (Modulation Highest Frequency). Command MHF,x with x = value in dHz
// Set the highest (top plateau) frequency of modulation signal in dHz
bool CMFC3045Instrument::ModulationHighestFreqSet(const U32 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command("MHF,");
		if ((rRegVal < MOD_FREQ_MIN_DHZ) || (rRegVal > MOD_FREQ_MAX_DHZ))
			throw CMFC3045InsException(std::string("Failed, MHF out of Range in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		else
			l_Command += std::to_string(rRegVal);
		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register MHF (Modulation Highest Frequency). Command MHF
// Get the highest (top plateau) frequency of modulation signal in dHz
bool CMFC3045Instrument::ModulationHighestFreqGet(U32 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_RdStr;
		if (!l_Write("MHF"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_RdStr))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		rRegVal = static_cast<U32>(std::stoul(l_RdStr));
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register MRE (Modulation Referenced frequency). Command MRE,x with x = 0 to 3 (see manual)
// Set which modulation frequency (MDA, MCF, MLF, MHF) is used as a reference
bool CMFC3045Instrument::ModulationReferenceSet(const eModulationRef & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command("MRE,");

		l_Command += std::to_string(static_cast<U8>(rRegVal));

		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register MRE (Modulation Highest Frequency). Command MRE
// Get the modulation reference
bool CMFC3045Instrument::ModulationReferenceGet(eModulationRef & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		U32 l_Modulation(255);
		std::string l_RdStr;
		if (!l_Write("MRE"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_RdStr))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		l_Modulation = static_cast<U32>(std::stoul(l_RdStr));
		switch (l_Modulation)
		{
		case static_cast<U32>(eModulationRef::kMDA):
			rRegVal = eModulationRef::kMDA;
			break;
		case static_cast<U32>(eModulationRef::kMCF):
			rRegVal = eModulationRef::kMCF;
			break;
		case static_cast<U32>(eModulationRef::kMLF):
			rRegVal = eModulationRef::kMLF;
			break;
		case static_cast<U32>(eModulationRef::kMHF):
			rRegVal = eModulationRef::kMHF;
			break;
		default:
			MTL_Assert(false);			
			break;
		}
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register MDP (Modulation Period). Command [MDP,x] with x = value in ms
// Set the period of frequency modulation cycle in ms
bool CMFC3045Instrument::ModulationPeriodSet(const U16 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command("MDP,");

		if (eAdvancedMode::kNone == m_AdvancedMode)
			throw CMFC3045InsException(std::string("MDP Set not allowed. Advanced mode < 1 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (rRegVal < 1)
			throw CMFC3045InsException(std::string("Failed, MDP out of Range in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		else
			l_Command += std::to_string(rRegVal);
		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register MDP (Modulation Period). Command MDP
// Get the period of frequency modulation cycle in ms
bool CMFC3045Instrument::ModulationPeriodGet(U16 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_RdStr;
		if (!l_Write("MDP"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_RdStr))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		rRegVal = static_cast<U16>(std::stoul(l_RdStr));
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register NSR (Number of Steps Ramp). Command [NSR]
// Get the number of steps for ramp modulation
bool CMFC3045Instrument::ModulationNumberStepRampGet(U32 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_RdStr;
		if (eAdvancedMode::kNone == m_AdvancedMode)
			throw CMFC3045InsException(std::string("NSR get not allowed. Advanced mode < 1 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_Write("NSR"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_RdStr))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		rRegVal = static_cast<U32>(std::stoul(l_RdStr));
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register NSP (Number of Steps Plateau). Command [NSP]
// Get the number of steps for plateau modulation
bool CMFC3045Instrument::ModulationNumberStepPlateauGet(U32 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_RdStr;
		if (eAdvancedMode::kNone == m_AdvancedMode)
			throw CMFC3045InsException(std::string("NSP get not allowed. Advance mode < 1 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_Write("NSP"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_RdStr))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		rRegVal = static_cast<U32>(std::stoul(l_RdStr));
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register DBR (DDS Bit Resolution of Steps Plateau). Command [DBR]
// Get the register value to calculate the resolution of the RF steps during ramping up and down. (see manual)
bool CMFC3045Instrument::DdsBitResolutionGet(U32 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_RdStr;
		if (eAdvancedMode::kNone == m_AdvancedMode)
			throw CMFC3045InsException(std::string("DBR get not allowed. Advance mode < 1 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_Write("DBR"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_RdStr))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		rRegVal = static_cast<U32>(std::stoul(l_RdStr));
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register DBR (DDS Bit Resolution of Steps Plateau). Command [DBR]
// Get the DDS resolution in dHZ with the given formula: 0.256*(DBR+1)*RFH
// DBR is DDS Bit Resolution register and RFH is the RF Harmonic Register
bool CMFC3045Instrument::DdsBitResolutionFdhzGet(F32 & rDdsResolutionFdhz)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		U32 l_DdsResolution;
		eRFH l_RfHarmonic;

		if (!DdsBitResolutionGet(l_DdsResolution))
			throw CMFC3045InsException(std::string("Failed to read DBR register in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!RfHarmonicGet(l_RfHarmonic))
			throw CMFC3045InsException(std::string("Failed to read DBR register in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		rDdsResolutionFdhz = 0.256f*(static_cast<F32>(l_DdsResolution) + 1.0f)*(static_cast<F32>(l_RfHarmonic));
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

//----------------------------------------------------------------------//
//						Measurement Parameters                          //
//----------------------------------------------------------------------//
// Write register NCY (Number of Measurement Cycles). Command NCY,x with x = decimal value (see manual)
// Set the number of measurement (or modulation) cycles
bool CMFC3045Instrument::NumberMeasCycleSet(const U16 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command("NCY,");

		if ((rRegVal < NCY_MIN) || (rRegVal > NCY_MAX))
			throw CMFC3045InsException(std::string("Failed, NCY out of Range in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		else
		l_Command += std::to_string(rRegVal);
		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register NCY (Number of Measurement Cycles). Command NCY
// Get the number of measurement (or modulation) cycles
bool CMFC3045Instrument::NumberMeasCycleGet(U16 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_RdStr;
		if (!l_Write("NCY"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_RdStr))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		rRegVal = static_cast<U16>(std::stoul(l_RdStr));
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register NPC (Number of Preliminary Cycles). Command {NPC,x} with x = decimal value (see manual)
// Set the number of dummy preliminary cycle before any measurement
bool CMFC3045Instrument::NumberPreCycleSet(const U16 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command("NPC,");

		if (eAdvancedMode::k2 != m_AdvancedMode)
			throw CMFC3045InsException(std::string("NPC Set not allowed. Advance mode < 2 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (rRegVal > NPC_MAX)
			throw CMFC3045InsException(std::string("Failed, NPC out of Range in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		else
			l_Command += std::to_string(rRegVal);
		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register NPC (Number of Preliminary Cycles). Command [NPC]
// Get the number of dummy preliminary cycle before any measurement
bool CMFC3045Instrument::NumberPreCycleGet(U16 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_RdStr;
		if (eAdvancedMode::kNone == m_AdvancedMode)
			throw CMFC3045InsException(std::string("NPC Get not allowed. Advance mode < 1 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_Write("NPC"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_RdStr))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		rRegVal = static_cast<U16>(std::stoul(l_RdStr));
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register NPT (Time duration of Preliminary cycles). Command {NPT,x} with x = value in ms (see manual)
// Set the maximum time duration in ms of the preminilary cycles.
bool CMFC3045Instrument::TimePreCycleSet(const U16 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command("NPT,");

		if (eAdvancedMode::k2 != m_AdvancedMode)
			throw CMFC3045InsException(std::string("NPT Set not allowed. Advance mode < 2 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (rRegVal > NPT_MAX)
			throw CMFC3045InsException(std::string("Failed, NPT out of Range in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		else
			l_Command += std::to_string(rRegVal);
		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register NPT (Time duration of Preliminary Cycles). Command [NPT]
// Get the maximum time duration in ms of the preminilary cycles.
bool CMFC3045Instrument::TimePreCycleGet(U16 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_RdStr;
		if (eAdvancedMode::kNone == m_AdvancedMode)
			throw CMFC3045InsException(std::string("NPT Get not allowed. Advance mode < 1 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_Write("NPT"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_RdStr))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		rRegVal = static_cast<U16>(std::stoul(l_RdStr));
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register RSO (Rejection Signal Offset). Command [RSO,x] with x = value in ppm (see manual)
// Set the offset between NMR signals of the ramps up and down in ppm (referenced to PCF register : PA central freq) 
bool CMFC3045Instrument::RejectionOffsetSet(const U16 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command("RSO,");

		if (eAdvancedMode::kNone == m_AdvancedMode)
			throw CMFC3045InsException(std::string("RSO Set not allowed. Advance mode < 1 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if ((rRegVal < RSO_MIN) || (rRegVal > RSO_MAX))
			throw CMFC3045InsException(std::string("Failed, RSO out of Range in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		else
			l_Command += std::to_string(rRegVal);
		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register RSO (Rejection Signal Offset). Command [RSO]
// Get the offset between NMR signals of the ramps up and down in ppm (referenced to PCF register : PA central freq) 
bool CMFC3045Instrument::RejectionOffsetGet(U16 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_RdStr;
		if (eAdvancedMode::kNone == m_AdvancedMode)
			throw CMFC3045InsException(std::string("RSO Get not allowed. Advance mode < 1 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_Write("RSO"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_RdStr))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		rRegVal = static_cast<U16>(std::stoul(l_RdStr));
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register RSG (Rejection Signal Gap). Command [RSG,x] with x = value in ppm (see manual)
// Set the gap limit in ppm (referenced to PCF register : PA central freq) between NMR signal ramps to reject a measurement
bool CMFC3045Instrument::RejectionGapSet(const U16 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command("RSG,");

		if (eAdvancedMode::kNone == m_AdvancedMode)
			throw CMFC3045InsException(std::string("RSG Set not allowed. Advance mode < 1 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if ((rRegVal < RSG_MIN) || (rRegVal > RSG_MAX))
			throw CMFC3045InsException(std::string("Failed, RSG out of Range in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		else
			l_Command += std::to_string(rRegVal);
		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register RSG (Rejection Signal Gap). Command [RSG]
// Get the gap limit in ppm (referenced to PCF register : PA central freq) between NMR signal ramps to reject a measurement
bool CMFC3045Instrument::RejectionGapGet(U16 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_RdStr;
		if (eAdvancedMode::kNone == m_AdvancedMode)
			throw CMFC3045InsException(std::string("RSG Get not allowed. Advance mode < 1 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_Write("RSG"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_RdStr))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		rRegVal = static_cast<U16>(std::stoul(l_RdStr));
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register TVP (Time Versus Precision). Command TVP,x with x = 0 or 1 (see manual)
// Set time versus precision measurement optimisation
bool CMFC3045Instrument::TimeVsPrecisionSet(const ePrecision & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command("TVP,");

		l_Command += std::to_string(static_cast<U8>(rRegVal));
		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register TVP (Time Versus Precision). Command TVP
// Get time versus precision measurement optimisation register value
bool CMFC3045Instrument::TimeVsPrecisionGet(ePrecision & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		U8 l_Precision(255);
		std::string l_RdStr;
		if (!l_Write("TVP"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_RdStr))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		l_Precision = static_cast<U8>(std::stoul(l_RdStr));
		switch (l_Precision)
		{
		case static_cast<U8>(ePrecision::kHighPrecision):
			rRegVal = ePrecision::kHighPrecision;
			break;
		case static_cast<U8>(ePrecision::kLowPrecision):
			rRegVal = ePrecision::kLowPrecision;
			break;
		default:
			MTL_Assert(false);
			break;
		}
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

//----------------------------------------------------------------------//
//						Data Acquisition  (Measurement)                 //
//----------------------------------------------------------------------//
// Write register RUN (Run measurement). Command RUN
// Starts a NMR measurement (of NCY Cycles) on all probes of probe-array.
// (Note: then read measurement results with BFV command)
bool CMFC3045Instrument::MeasCycleStart()
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_Write("RUN"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register RUN (Run measurement). Command RUN,x wih x= probe number 0 to nbProbe in PA
// Starts a NMR measurement of (NCY Cycles) on an individual probe (if x>1 to nbProbes PA) or on all PA (if x=0)
// (Note: then read measurement result with BFV,x command)
bool CMFC3045Instrument::MeasCycleStart(const U8 & rProbeSel)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (rProbeSel > m_NbProbesInPA)
			throw CMFC3045InsException(std::string("Probe selection out of range in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		std::string l_Command("RUN,");
		l_Command += std::to_string(rProbeSel);
		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Activate SMA (send message automatically from main unit) and Start a NMR measurement on the whole probe array.
// Only PA not connected and Data Ready SMA bit in register are activated.
// Warning: after this method, avoid using any get (Write + Read) commands. Use only the WaitSmaMeasureDataReady()
// to wait on the measurement data ready message.
// You must disable the SMA before to send any get (Write + Read) commands.
bool CMFC3045Instrument::MeasStartWithSMA()
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		m_SMAValue.RawSMAB =0; // clear previous SMA message flag received

		// Set SMA for PA & DR
		uSMAByte l_SMARegValRead;
		uSMAByte l_SMARegVal;
		l_SMARegValRead.RawSMAB =0;
		l_SMARegVal.RawSMAB =0;
		l_SMARegVal.SMAByte.DRST1 =1;
		l_SMARegVal.SMAByte.PANST2 =1;
		if (!l_SMASet(l_SMARegVal.RawSMAB))
			throw CMFC3045InsException(std::string("Failed to set SMA register in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!SMAGet(l_SMARegValRead.RawSMAB))
			throw CMFC3045InsException(std::string("Failed to read back SMA register in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if(l_SMARegVal.RawSMAB != l_SMARegValRead.RawSMAB)
			throw CMFC3045InsException(std::string("Error. SMA is not enable") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		// Start measurement
		if (!MeasCycleStart())
			throw CMFC3045InsException(std::string("Failed to start measurement in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Activate SMA (send message automatically from main unit) and Start a NMR measurement on a specific single probe.
// Only PA not connected and Data Ready SMA bit in register are activated.
// Warning: after this method, avoid using any get (Write + Read) commands. Use only the WaitSmaMeasureDataReady()
// to wait on the measurement data ready message.
bool CMFC3045Instrument::MeasStartWithSMA(const U8 & rProbeSel)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		m_SMAValue.RawSMAB =0; // clear previous SMA message flag

		// Set SMA for PA & DR
		uSMAByte l_SMARegValRead;
		uSMAByte l_SMARegVal;
		l_SMARegValRead.RawSMAB =0;
		l_SMARegVal.RawSMAB =0;
		l_SMARegVal.SMAByte.DRST1 =1;
		l_SMARegVal.SMAByte.PANST2 =1;
		if (!l_SMASet(l_SMARegVal.RawSMAB))
			throw CMFC3045InsException(std::string("Failed to set SMA register in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!SMAGet(l_SMARegValRead.RawSMAB))
			throw CMFC3045InsException(std::string("Failed to read back SMA register in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if(l_SMARegVal.RawSMAB != l_SMARegValRead.RawSMAB)
			throw CMFC3045InsException(std::string("Error. SMA is not enable") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		// Start measurement
		if (!MeasCycleStart(rProbeSel))
			throw CMFC3045InsException(std::string("Failed to start measurement in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// When measurement is processing and SMA is activated use this method to wait during
// x milliseconds on SMA message "DR" (data ready) or "PA" (probe array not connected).
// SMA is disable after "DR" or "PA" message.
// Note: be sure to never use this method while waiting for another standard Read command!
bool CMFC3045Instrument::WaitSmaMeasureDataReady(const U32 & rMeasWaitTimeout_ms, uSMAByte & rSmaValue)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);
	rSmaValue.RawSMAB = 0;

	try {
		// Wait for SMA Flag DR or PA with a milliseconds timeout
		if(m_SMAEnabled)
		{
			U32 l_RxNewDataLength(0);
			std::string l_TempRdStrOut;
			uSMAByte l_SMAMsg;
			l_SMAMsg.RawSMAB =0;
			U32 l_NbSmaMsg(0);

			// Wait and check during rMeasWaitTimeout_ms for new data available in Rx Buffer before to read it. If no new SMA message received before timeout we go on further.
			if(l_WaitAvailableDataRX(rMeasWaitTimeout_ms, l_RxNewDataLength, false))
			{
				if (l_RxNewDataLength >= 4)
					l_NbSmaMsg = l_RxNewDataLength / 4; // size per SMA message (= 2 ascii char + 2 term char)
				for(size_t l_i =0; l_i < l_NbSmaMsg ; l_i++)
				{
					m_ReadBuffer.clear();
					if (!CVISAInstrument::Read(m_ReadBuffer,false))
					{
						ViStatus l_VisaStatus = Status();
						if (!((VI_ERROR_TMO == l_VisaStatus) &&
								(1 <= m_ReadBuffer.size()) && (static_cast<char>(DATA_END_ETB) == m_ReadBuffer[m_ReadBuffer.size() - 1])))
						{
							throw CMFC3045InsException("Could not read", MTL__LOCATION__);
						}
					}
					else
					{
						l_ExtractPaDrSmaMsgFromDataRX(std::string(m_ReadBuffer.begin(), m_ReadBuffer.end()), l_SMAMsg, l_TempRdStrOut); //Extract only PA and DR MSA message from input string, any other message are left in l_TempRdStrOut.
						m_SMAValue.RawSMAB |= l_SMAMsg.RawSMAB;
					}
				}
			}
			rSmaValue = m_SMAValue;
			if (m_SMAValue.SMAByte.DRST1 || m_SMAValue.SMAByte.PANST2)
			{
				// Stop SMA
				if (!DisableSMA())
					throw CMFC3045InsException(std::string("Failed to disable SMA in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
			}
			m_SMAValue.RawSMAB =0; // global SMA value has been read. Clear it!
		}
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register SRC (Search NMR signal). Command SRC
// Starts a NMR search over the full probe-array range frequency using all probes
// (Note: then read search results with BFV,x command)
bool CMFC3045Instrument::SearchStart()
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_Write("SRC"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register SRC (Search NMR signal). Command SRC,x wih x= probe number 0 to nbProbe in PA 
// Starts a NMR search over the full probe-array range frequency using an individual probe (if x>1 to nbProbes PA) or on all PA (if x=0)
// (Note: then read measurement result with BFV,x command)
bool CMFC3045Instrument::SearchStart(const U8 & rProbeSel)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (rProbeSel > m_NbProbesInPA)
			throw CMFC3045InsException(std::string("Probe selection out of range in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		
		std::string l_Command("SRC,");
		l_Command += std::to_string(rProbeSel);
		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}


// Activate SMA (send message automatically from main unit) and Start a NMR Search on the whole probe array.
// Only PA not connected and Data Ready SMA bit in register are activated.
// Warning: after this method, avoid using any get (Write + Read) commands. Use only the WaitSmaMeasureDataReady()
// to wait on the measurement data ready message.
// You must disable the SMA before to send any get (Write + Read) commands.
bool CMFC3045Instrument::SearchStartWithSMA()
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		m_SMAValue.RawSMAB =0; // clear previous SMA message flag received

		// Set SMA for PA & DR
		uSMAByte l_SMARegValRead;
		uSMAByte l_SMARegVal;
		l_SMARegValRead.RawSMAB =0;
		l_SMARegVal.RawSMAB =0;
		l_SMARegVal.SMAByte.DRST1 =1;
		l_SMARegVal.SMAByte.PANST2 =1;
		if (!l_SMASet(l_SMARegVal.RawSMAB))
			throw CMFC3045InsException(std::string("Failed to set SMA register in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!SMAGet(l_SMARegValRead.RawSMAB))
			throw CMFC3045InsException(std::string("Failed to read back SMA register in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if(l_SMARegVal.RawSMAB != l_SMARegValRead.RawSMAB)
			throw CMFC3045InsException(std::string("Error. SMA is not enable") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		// Start search
		if (!SearchStart())
			throw CMFC3045InsException(std::string("Failed to start Search in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Activate SMA (send message automatically from main unit) and Start a NMR search on a specific single probe.
// Only PA not connected and Data Ready SMA bit in register are activated.
// Warning: after this method, avoid using any get (Write + Read) commands. Use only the WaitSmaMeasureDataReady()
// to wait on the measurement data ready message.
// You must disable the SMA before to send any get (Write + Read) commands.
bool CMFC3045Instrument::SearchStartWithSMA(const U8 & rProbeSel)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		m_SMAValue.RawSMAB =0; // clear previous SMA message flag received

		// Set SMA for PA & DR
		uSMAByte l_SMARegValRead;
		uSMAByte l_SMARegVal;
		l_SMARegValRead.RawSMAB =0;
		l_SMARegVal.RawSMAB =0;
		l_SMARegVal.SMAByte.DRST1 =1;
		l_SMARegVal.SMAByte.PANST2 =1;
		if (!l_SMASet(l_SMARegVal.RawSMAB))
			throw CMFC3045InsException(std::string("Failed to set SMA register in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!SMAGet(l_SMARegValRead.RawSMAB))
			throw CMFC3045InsException(std::string("Failed to read back SMA register in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if(l_SMARegVal.RawSMAB != l_SMARegValRead.RawSMAB)
			throw CMFC3045InsException(std::string("Error. SMA is not enable") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		// Start search
		if (!SearchStart(rProbeSel))
			throw CMFC3045InsException(std::string("Failed to start Search in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register CTN (Continuous measurement). Command CTN
// Starts a continuous measurement (infinite modulation cycles) until a BRK command
bool CMFC3045Instrument::ContinuousMeasStart()
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_Write("CTN"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register BRK (Break). Command BRK
// Interrupts any data acquisition (stop measurement or data block transfert)
bool CMFC3045Instrument::BreakMeasSet()
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_Write("BRK", true))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

//----------------------------------------------------------------------//
//						Data Reading                                    //
//----------------------------------------------------------------------//
// Write register BLK (Data transfert Block Mode). Command BLK,x with x= 0 to 2 mode
// Set the data transfert block mode. (see manual)
// This mode is used for reading data as measurements & calibration corrections table
bool CMFC3045Instrument::l_DataBlockModeSet(const eDataBlockMode & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command("BLK,");
		l_Command += std::to_string(static_cast<U8>(rRegVal));
		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		m_DataBlockMode = rRegVal;
	}
	catch (CMFC3045InsException & rE)
	{
		m_DataBlockMode = eDataBlockMode::kOneByOne; // default main unit value after reset
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write & Read register BLK (Data transfert Block Mode) to decimal. Commands BLK,1 and BLK
// Set data transfert block mode to decimal and read back value
bool CMFC3045Instrument::DataBlockModeDecimalSet()
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		eDataBlockMode l_BlockMode;
		if(!l_DataBlockModeSet(eDataBlockMode::kAllDecimal))
			throw CMFC3045InsException(std::string("Failed to set decimal data block mode in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!DataBlockModeGet(l_BlockMode))
			throw CMFC3045InsException(std::string("Failed to get data block mode in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if(eDataBlockMode::kAllDecimal != l_BlockMode)
			throw CMFC3045InsException(std::string("Failed: read back data block mode is not decimal in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register BLK (Data transfert Block Mode). Command BLK
// Get the data transfert block mode. (see manual)
// This mode is used for reading data as measurements & calibration corrections table
bool CMFC3045Instrument::DataBlockModeGet(eDataBlockMode & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		U32 l_BlockMode(255);
		std::string l_RdStr;
		if (!l_Write("BLK"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_RdStr))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		l_BlockMode = static_cast<U32>(std::stoul(l_RdStr));
		switch (l_BlockMode)
		{
		case static_cast<U32>(eDataBlockMode::kOneByOne):
			rRegVal = eDataBlockMode::kOneByOne;
			break;
		case static_cast<U32>(eDataBlockMode::kAllDecimal):
			rRegVal = eDataBlockMode::kAllDecimal;
			break;
		case static_cast<U32>(eDataBlockMode::kAllHexa):
			rRegVal = eDataBlockMode::kAllHexa;
			break;
		default:
			MTL_Assert("UnHandled value Type");
			throw CMFC3045InsException(std::string("Failed: Get wrong value for BLK in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
			break;
		}
		m_DataBlockMode = rRegVal;
	}
	catch (CMFC3045InsException & rE)
	{
		m_DataBlockMode = eDataBlockMode::kOneByOne;
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read the expected End of transmission Block character
// Note: Change the termination character to end of transmission block 0x17 and read it.
// then set back the previous Termination Char.
// (This is needed after the last measurement data read of a block transfer)
bool CMFC3045Instrument::l_ReadEndTransmissionBlockChar()
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_RdStr;
		// Change the Termination Char to End of transmission Block 0x17 to read it after the last measurement data
		sSerialPortSettings l_SerialParms = m_HostSerialConfig;
		l_SerialParms.ReadTermChar = static_cast<char>(DATA_END_ETB);
		if(!l_SerialPortHostConfig(l_SerialParms))
			throw CMFC3045InsException(std::string("Failed: to modify VISA serial term char in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if(!l_ReadString(l_RdStr))	// read end of transmission Block ETB character
			throw CMFC3045InsException(std::string("Failed: Didn't get end of transmission block char in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_SerialPortHostConfig(m_HostSerialConfig)) // set back previous ReadTermChar in Serial Port config & clear Serial Rx Tx buffers
			throw CMFC3045InsException(std::string("Failed: to modify VISA serial term char in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		l_SerialPortHostConfig(m_HostSerialConfig);	// set back previous ReadTermChar in Serial Port config & clear Serial Rx Tx buffers
		return false;
	}
	return true;
}

// Get, according to a write string cmd, data measurements in a block of decimal data
// Note: Data is read according to the active block mode (BLK). We need BLK = 1 = kAllDecimal
// The rData vector size must have been initialized to the exepcted number of data to be received.
// NumberProbesGet() or NumberProbesSet() must have been called once before it to set attribute m_NbProbesInPA
bool CMFC3045Instrument::l_ReadMeasurementTableBlock(const std::string rStringCMD, std::vector<U32> & rData)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if ((0 == rData.size()) || (rData.size() < m_NbProbesInPA) || (NPR_MAX < rData.size()))
			throw CMFC3045InsException(std::string("Failed. Vector size out of range to store received data.") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (eDataBlockMode::kAllDecimal != m_DataBlockMode)
			throw CMFC3045InsException(std::string("Failed. Wrong Block mode. Block mode must be 1 in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		if (!l_Write(rStringCMD))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		for (size_t i = 0; i < rData.size(); i++)
		{
			std::string l_StrData;
			if (!l_ReadString(l_StrData))
				throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
			if (l_StrData.empty())
				rData[i] = 0;
			else
				rData[i] = static_cast<U32>(std::stoul(l_StrData));
		}
		// Read the char End of transmission Block 0x17 after the last measurement data
		if (!l_ReadEndTransmissionBlockChar())
			throw CMFC3045InsException(std::string("Read failed ETB in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register BFV (B Magnetic Field Value). Command BFV
// Get the magnetic field value in dHz from last measurements or search operation for all probes
// Note: Data is read according to the active block mode (BLK). We need BLK = 1 = kAllDecimal
// m_NbProbesInPA must have been initialized once befor by a call of NumberProbesGet()
bool CMFC3045Instrument::DataMagneticFieldGet(std::vector<U32> & rData)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);
	rData.clear();
	try {
		if (eDataBlockMode::kAllDecimal != m_DataBlockMode)
			throw CMFC3045InsException(std::string("Failed. Wrong Block mode. Block mode must be 1 in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if ((0 == m_NbProbesInPA) || (NPR_MAX < m_NbProbesInPA))
			throw CMFC3045InsException(std::string("Failed. Unvalid probes number in PA. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		rData.resize(m_NbProbesInPA);
		if (!l_ReadMeasurementTableBlock("BFV", rData))
			throw CMFC3045InsException(std::string("Read failed measurement block in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register BFV (B Magnetic Field Value). Command BFV,x with x= probe number 1 to nbProbe in PA
// Get the magnetic field value in dHz from last measurement or search operation of a specified probe
// Note: Data is read according to the active block mode (BLK). We need BLK = 1 = kAllDecimal
bool CMFC3045Instrument::DataMagneticFieldGet(U32 & rData, const U8 & rProbeSel)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (rProbeSel < 1 || rProbeSel > m_NbProbesInPA)
			throw CMFC3045InsException(std::string("Probe selection out of range in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (eDataBlockMode::kAllDecimal != m_DataBlockMode)
			throw CMFC3045InsException(std::string("Failed. Wrong Block mode. Block mode must be 1 in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		std::string l_Command("BFV,");
		l_Command += std::to_string(rProbeSel);
		std::string l_StrData;
		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_StrData))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (l_StrData.empty())
			rData = 0;
		else
			rData = static_cast<U32>(std::stoul(l_StrData));
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register BSD (B Magnetic standard Deviation). Command BSD
// Get the standard deviation in dHz of each probe measurement
// Note: Block mode must be 1 (kAllDecimal). BSD is not available in CTN measurement mode
// m_NbProbesInPA must have been initialized once befor by a call of NumberProbesGet()
bool CMFC3045Instrument::DataStdDeviationGet(std::vector<U32> & rData)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);
	rData.clear();
	try {
		if (eDataBlockMode::kAllDecimal != m_DataBlockMode)
			throw CMFC3045InsException(std::string("Failed. Wrong Block mode. Block mode must be 1 in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if ((0 == m_NbProbesInPA) || (NPR_MAX < m_NbProbesInPA))
			throw CMFC3045InsException(std::string("Failed. Unvalid probes number in PA. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		rData.resize(m_NbProbesInPA);
		if (!l_ReadMeasurementTableBlock("BSD", rData))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register BSD (B Magnetic standard Deviation). Command BSD,x with x= probe number 1 to nbProbe in PA
// Get the standard deviation in dHz of a selected probe measurement
// Block mode must be 1 (kAllDecimal). BSD is not available in CTN measurement mode
bool CMFC3045Instrument::DataStdDeviationGet(U32 & rData, const U8 & rProbeSel)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (rProbeSel < 1 || rProbeSel > m_NbProbesInPA)
			throw CMFC3045InsException(std::string("Probe selection out of range in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (eDataBlockMode::kAllDecimal != m_DataBlockMode)
			throw CMFC3045InsException(std::string("Failed. Wrong Block mode. Block mode must be 1 in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		
		std::string l_Command("BSD,");
		l_Command += std::to_string(rProbeSel);
		std::string l_StrData;
		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_StrData))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (l_StrData.empty())
			rData = 0;
		else
			rData = static_cast<U32>(std::stoul(l_StrData));
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register BNC (Number of valid Cycle). Command BNC
// Get the number of valid NMR Cycles acquisition during last measurements of each probe
// Note: Block mode must be 1 (kAllDecimal). BNC is not available in CTN measurement mode
// m_NbProbesInPA must have been initialized once befor by a call of NumberProbesGet()
bool CMFC3045Instrument::DataNumberValidCycleGet(std::vector<U32> & rData)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);
	rData.clear();
	try {
		if (eDataBlockMode::kAllDecimal != m_DataBlockMode)
			throw CMFC3045InsException(std::string("Failed. Wrong Block mode. Block mode must be 1 in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if ((0 == m_NbProbesInPA) || (NPR_MAX < m_NbProbesInPA))
			throw CMFC3045InsException(std::string("Failed. Unvalid probes number in PA. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		rData.resize(m_NbProbesInPA);
		if (!l_ReadMeasurementTableBlock("BNC", rData))
			throw CMFC3045InsException(std::string("Failed to read measurement block in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register BNC (Number of valid Cycle). Command BNC,x with x= probe number 1 to nbProbe in PA
// Get the number of valid NMR Cycles acquisition during last measurement of selected probe
// Block mode must be 1 (kAllDecimal). BNC is not available in CTN measurement mode
bool CMFC3045Instrument::DataNumberValidCycleGet(U32 & rData, const U8 & rProbeSel)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (rProbeSel < 1 || rProbeSel > m_NbProbesInPA)
			throw CMFC3045InsException(std::string("Probe selection out of range in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (eDataBlockMode::kAllDecimal != m_DataBlockMode)
			throw CMFC3045InsException(std::string("Failed. Wrong Block mode. Block mode must be 1 in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	
		std::string l_Command("BNC,");
		l_Command += std::to_string(rProbeSel);
		std::string l_StrData;
		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_StrData))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (l_StrData.empty())
			rData = 0;
		else
			rData = static_cast<U32>(std::stoul(l_StrData));
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register BIN (Individual NMR frequency). Command BIN
// Get two NMR Frequency measurements in dHz, one for ramp up and one for ramp down according to the last previous individual RUN,x or SRC,x cmd
// The first measurement value returned is ramp up, the second value is ramp down.
// Block mode must be 1 (kAllDecimal). BIN is not available in CTN measurement mode
bool CMFC3045Instrument::DataIndividualFrequencyGet(U32  & rDataUp, U32 & rDataDown)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (eDataBlockMode::kAllDecimal != m_DataBlockMode)
			throw CMFC3045InsException(std::string("Failed. Wrong Block mode. Block mode must be 1 in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		if (!l_Write("BIN"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		std::string l_StrDataUp;
		if (!l_ReadString(l_StrDataUp))
			throw CMFC3045InsException(std::string("Read failed ramp up in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (l_StrDataUp.empty())
			rDataUp = 0;
		else
			rDataUp = static_cast<U32>(std::stoul(l_StrDataUp));

		std::string l_StrDataDwn;
		if (!l_ReadString(l_StrDataDwn))
			throw CMFC3045InsException(std::string("Read failed ramp down in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (l_StrDataDwn.empty())
			rDataDown = 0;
		else
			rDataDown = static_cast<U32>(std::stoul(l_StrDataDwn));

		// Read the char End of transmission Block 0x17 after the last measurement data
		if (!l_ReadEndTransmissionBlockChar())
			throw CMFC3045InsException(std::string("Read failed ETB in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register BFC (Central NMR frequency). Command BFC
// Get the central (median, not average) NMR Frequency in dHz of all valid probes measurement.
bool CMFC3045Instrument::DataCentralFrequencyGet(U32 & rData)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_StrData;
		if (!l_Write("BFC"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_StrData))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (l_StrData.empty())
			rData = 0;
		else
			rData = static_cast<U32>(std::stoul(l_StrData));
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register BFL (Lowest NMR frequency). Command BFL
// Get the lowest NMR Frequency in dHz of all valid probes measurement.
bool CMFC3045Instrument::DataLowestFrequencyGet(U32 & rData)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_StrData;
		if (!l_Write("BFL"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_StrData))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (l_StrData.empty())
			rData = 0;
		else
			rData = static_cast<U32>(std::stoul(l_StrData));
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register BFH (Highest NMR frequency). Command BFH
// Get the highest NMR Frequency in dHz of all valid probes measurement.
bool CMFC3045Instrument::DataHighestFrequencyGet(U32 & rData)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_StrData;
		if (!l_Write("BFH"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_StrData))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (l_StrData.empty())
			rData = 0;
		else
			rData = static_cast<U32>(std::stoul(l_StrData));
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register BFD (Difference highest-lowest NMR frequency). Command BFD
// Get the difference (highest - lowest) NMR Frequency in ppm (referenced to BFC register value) of all valid probes measurement.
bool CMFC3045Instrument::DataDifferenceFrequencyGet(U32 & rData)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_StrData;
		if (!l_Write("BFD"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_StrData))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (l_StrData.empty())
			rData = 0;
		else
			rData = static_cast<U32>(std::stoul(l_StrData));
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

//----------------------------------------------------------------------//
//                  Probe Array Informations                            //
//----------------------------------------------------------------------//
// Write register NPR (Number of PRobes). Command {NPR,x} with x = 1 to Nb_of_ProbesPA
// Set the total number of probes (measurement points) in the Probe-Array
bool CMFC3045Instrument::NumberProbesSet(const U8 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command("NPR,");

		if (eAdvancedMode::k2 != m_AdvancedMode)
			throw CMFC3045InsException(std::string("NPR Set not allowed. Advance mode < 2 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if ((rRegVal < 1) || (rRegVal > NPR_MAX))
			throw CMFC3045InsException(std::string("Failed, NPR out of Range in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		l_Command += std::to_string(rRegVal);
		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		m_NbProbesInPA = rRegVal; // save number of Probe-array probes
	}
	catch (CMFC3045InsException & rE)
	{
		m_NbProbesInPA = 0; // clear number of Probe-array probes
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register NPR (Number of PRobes). Command NPR
// Get the total number of probes (measurement points) in the Probe-Array
bool CMFC3045Instrument::NumberProbesGet(U8 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_RdStr;
		if (!l_Write("NPR"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_RdStr))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		rRegVal = static_cast<U8>(std::stoul(l_RdStr));
		m_NbProbesInPA = rRegVal; // save number of probes in Probe-array 
	}
	catch (CMFC3045InsException & rE)
	{
		m_NbProbesInPA = 0; // clear number of Probe-array probes
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register PCF (Probe array Central Frequency). Command {PCF,x} with x = value in dHz
// Set the probe array central frequency in dHz
bool CMFC3045Instrument::PaCentralFreqSet(const U32 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command("PCF,");
		if ((rRegVal < PA_FREQ_MIN_DHZ) || (rRegVal > PA_FREQ_MAX_DHZ))
			throw CMFC3045InsException(std::string("Failed, PCF out of Range in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (eAdvancedMode::k2 != m_AdvancedMode)
			throw CMFC3045InsException(std::string("PCF Set not allowed. Advance mode < 2 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		l_Command += std::to_string(rRegVal);
		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register PCF (Probe array Central Frequency). Command PCF
// Get the probe array central frequency in dHz
bool CMFC3045Instrument::PaCentralFreqGet(U32 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_RdStr;
		if (!l_Write("PCF"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_RdStr))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		rRegVal = static_cast<U32>(std::stoul(l_RdStr));
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register PLF (Probe array Lowest Frequency). Command {PLF,x} with x = value in dHz
// Set probe array lowest frequency in dHz
bool CMFC3045Instrument::PaLowestFreqSet(const U32 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command("PLF,");
		if ((rRegVal < PA_FREQ_MIN_DHZ) || (rRegVal > PA_FREQ_MAX_DHZ))
			throw CMFC3045InsException(std::string("Failed, PLF out of Range in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (eAdvancedMode::k2 != m_AdvancedMode)
			throw CMFC3045InsException(std::string("PLF Set not allowed. Advance mode < 2 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		l_Command += std::to_string(rRegVal);
		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register PLF (Probe array Lowest Frequency). Command PLF
// Get the probe array lowest entral frequency in dHz
bool CMFC3045Instrument::PaLowestFreqGet(U32 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_RdStr;
		if (!l_Write("PLF"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_RdStr))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		rRegVal = static_cast<U32>(std::stoul(l_RdStr));
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register PHF (Probe array Highest Frequency). Command {PHF,x} with x = value in dHz
// Set probe array highest frequency in dHz
bool CMFC3045Instrument::PaHighestFreqSet(const U32 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command("PHF,");
		if ((rRegVal < PA_FREQ_MIN_DHZ) || (rRegVal > PA_FREQ_MAX_DHZ))
			throw CMFC3045InsException(std::string("Failed, PHF out of Range in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (eAdvancedMode::k2 != m_AdvancedMode)
			throw CMFC3045InsException(std::string("PHF Set not allowed. Advance mode < 2 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		l_Command += std::to_string(rRegVal);
		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register PHF (Probe array Highest Frequency). Command PHF
// Get the probe array highest entral frequency in dHz
bool CMFC3045Instrument::PaHighestFreqGet(U32 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_RdStr;
		if (!l_Write("PHF"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_RdStr))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		rRegVal = static_cast<U32>(std::stoul(l_RdStr));
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register RFH (RF Harmonic). Command {RFH,x} with x = 1,3,5,7
// Set RF harmonic 
bool CMFC3045Instrument::RfHarmonicSet(const eRFH & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command("RFH,");

		if (eAdvancedMode::k2 != m_AdvancedMode)
			throw CMFC3045InsException(std::string("RFH Set not allowed. Advance mode < 2 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		l_Command += std::to_string(static_cast<U8>(rRegVal));
		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register RFH (RF Harmonic). Command RFH
// Get RF harmonic 
bool CMFC3045Instrument::RfHarmonicGet(eRFH & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		U32 l_RFHarmonic(0);
		std::string l_RdStr;
		if (!l_Write("RFH"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_RdStr))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		l_RFHarmonic = static_cast<U32>(std::stoul(l_RdStr));
		switch (l_RFHarmonic)
		{
		case static_cast<U32>(eRFH::k1):
			rRegVal = eRFH::k1;
			break;
		case static_cast<U32>(eRFH::k3):
			rRegVal = eRFH::k3;
			break;
		case static_cast<U32>(eRFH::k5):
			rRegVal = eRFH::k5;
			break;
		case static_cast<U32>(eRFH::k7):
			rRegVal = eRFH::k7;
			break;
		default:
			MTL_Assert("UnHandled value Type");
			throw CMFC3045InsException(std::string("Failed: Get wrong value for RFH in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
			break;
		}
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register NST (NMR Signal Threshold). Command {NST,x} with x = 0 to 255
// Set DC threshold level to detect NMR signal
bool CMFC3045Instrument::NmrThresholdSet(const U8 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command("NST,");

		if (eAdvancedMode::k2 != m_AdvancedMode)
			throw CMFC3045InsException(std::string("NST Set not allowed. Advance mode < 2 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		l_Command += std::to_string(rRegVal);
		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// read register NST (NMR Signal Threshold). Command [NST]
// Get DC threshold level to detect NMR signal
bool CMFC3045Instrument::NmrThresholdGet(U8 & rRegVal)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_RdStr;
		if (eAdvancedMode::kNone == m_AdvancedMode)
			throw CMFC3045InsException(std::string("NST Get not allowed. Advance mode < 1 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_Write("NST"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_RdStr))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		rRegVal = static_cast<U8>(std::stoul(l_RdStr));
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

//----------------------------------------------------------------------//
//           EEPROMs (Main unit & PA) and Probe Array Normalisation     //
//----------------------------------------------------------------------//
// Write register SMU (Settings of Main Unit). Command [SMU,>EEP]
// Copy main unit settings parameters from main unit RAM to main unit EEPROM.
bool CMFC3045Instrument::MainUnitEepromWrite()
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (eAdvancedMode::kNone == m_AdvancedMode)
			throw CMFC3045InsException(std::string("SMU Set not allowed. Advance mode < 1 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);	
		
		if (!l_Write("SMU,>EEP"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register SMU (Settings of Main Unit). Command [SMU]
// Read main unit settings parameters from main unit EEPROM.
bool CMFC3045Instrument::MainUnitEepromRead(std::vector<char> & rEEPROMContent)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (eAdvancedMode::kNone == m_AdvancedMode)
			throw CMFC3045InsException(std::string("SMU Set not allowed. Advance mode < 1 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		if (!l_Write("SMU"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		if (!l_ReadNCharNoTerm(m_ReadBuffer, MU_EEPROM_CHARSIZE + CHECKSUM_CHARSIZE)) // Main unit EEPROM has 80 bytes => 160 ASCII Hexa characters. We add 4 char for CheckSum-16bit
			throw CMFC3045InsException(std::string("Read failed Main Unit EEPROM in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		
		rEEPROMContent.assign(m_ReadBuffer.begin(), m_ReadBuffer.end());
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register SPA (Settings of Probe Array). Command [SPA,>EEP]
// Write probe array settings parameters and calibration probe table (from Main unit RAM) to probe array EEPROM. 
bool CMFC3045Instrument::ProbeArrayEepromWrite()
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (eAdvancedMode::kNone == m_AdvancedMode)
			throw CMFC3045InsException(std::string("SPA Set not allowed. Advance mode < 1 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		if (!l_Write("SPA,>EEP"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		MTL_SleepMs(200); //let Main unit time to do it
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register SPA (Settings of Probe Array). Command [SPA]
// Read probe array settings parameters and calibration probe table from the probe array EEPROM.
bool CMFC3045Instrument::ProbeArrayEepromRead(std::vector<char> & rEEPROMContent)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (eAdvancedMode::kNone == m_AdvancedMode)
			throw CMFC3045InsException(std::string("SPA Set not allowed. Advance mode < 1 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		if (!l_Write("SPA"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadNCharNoTerm(m_ReadBuffer, PA_EEPROM_CHARSIZE + CHECKSUM_CHARSIZE))
			throw CMFC3045InsException(std::string("Read failed Probe Array EEPROM in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		rEEPROMContent.assign(m_ReadBuffer.begin(), m_ReadBuffer.end());
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register CPS (Calibration Probe Selection). Command [CPS,x], with x = a probe number
// Fill main unit RAM calibration measurement table with a probe measurement.
// After a measurement, copy the measurement of a selected probe to the main unit RAM probe measurement table (without correction) used for calibration.
bool CMFC3045Instrument::CalibrationProbeMeasurementSet(const U8 & rProbeSel)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command("CPS,");
		U8 l_PaProbeNumber(0);

		if (eAdvancedMode::kNone == m_AdvancedMode)
			throw CMFC3045InsException(std::string("CPS Set not allowed. Advance mode < 1 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if(!NumberProbesGet(l_PaProbeNumber))
			throw CMFC3045InsException(std::string("Failed Get PA Probes Number in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if ((rProbeSel == 0) || (rProbeSel > m_NbProbesInPA))
			throw CMFC3045InsException(std::string("Failed, CPS probe target out of Range in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		l_Command += std::to_string(rProbeSel);
		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register CPS (Calibration Probe Selection). Command [CPS,0]
// Clear all main unit RAM probe measurement table (without correction) used for calibration.
bool CMFC3045Instrument::CalibrationProbeMeasurementTableClear()
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (eAdvancedMode::kNone == m_AdvancedMode)
			throw CMFC3045InsException(std::string("CPS Clear Table not allowed. Advance mode < 1 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_Write("CPS,0"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register CPS (Calibration Probe Selection). Command [CPS]
// Get from main unit RAM the whole calibration probe measurement table (without correction) in dHz.
// Block mode must be 1 (kAllDecimal). Return rData size will be NbProbes in PA
// Data[0] is the first probe measurement. Data[n] is the measurement of probe n.
bool CMFC3045Instrument::CalibrationProbeMeasurementTableGet(std::vector<U32> & rData)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);
	rData.clear();
	try {
		U8 l_PaProbeNumber(0);
		if (eAdvancedMode::kNone == m_AdvancedMode)
			throw CMFC3045InsException(std::string("CPS Set not allowed. Advance mode < 1 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!NumberProbesGet(l_PaProbeNumber))
			throw CMFC3045InsException(std::string("Failed Get PA Probes Number in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if ((0 == m_NbProbesInPA) || (NPR_MAX < m_NbProbesInPA))
			throw CMFC3045InsException(std::string("Failed to get valid probes number from PA. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (eDataBlockMode::kAllDecimal != m_DataBlockMode)
			throw CMFC3045InsException(std::string("Failed. Wrong Block mode. Block mode must be 1 in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		rData.resize(m_NbProbesInPA);
		if (!l_ReadMeasurementTableBlock("CPS", rData))
			throw CMFC3045InsException(std::string("Failed to read Calibration measurement block in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register CPS (Calibration Probe Selection). Command [CPS,x], with x > number of probes in PA
// Compute and get the average (in dHz) of the calibration probe measurement table (RAM measurement table without correction).
bool CMFC3045Instrument::CalibrationMeasurementAverageProbeGet(U32 & rData)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command("CPS,");
		std::string l_RdStr;
		if (eAdvancedMode::kNone == m_AdvancedMode)
			throw CMFC3045InsException(std::string("CPS Set not allowed. Advance mode < 1 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		l_Command += std::to_string(NPR_MAX + 1);
		if (!l_Write(l_Command))	 // send parameter value > than NPR
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_RdStr))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		rData = static_cast<U32>(std::stoul(l_RdStr));
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register CBT (Calibration Build Table). Command {CBT,-1}
// Copy the calibration correction table from the probe array EEPROM to the main unit RAM table.
bool CMFC3045Instrument::CalibrationBuildTableCopyEEPROMToRAM()
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (eAdvancedMode::k2 != m_AdvancedMode)
			throw CMFC3045InsException(std::string("CBT Table Copy not allowed. Advance mode < 2 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_Write("CBT,-1"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		MTL_SleepMs(200);// let main unit to do it
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register CBT (Calibration Build Table). Command {CBT,0}
// Clear in main unit RAM the calibration correction table.
bool CMFC3045Instrument::CalibrationBuildTableClearRAM()
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (eAdvancedMode::k2 != m_AdvancedMode)
			throw CMFC3045InsException(std::string("CBT Table Clear not allowed. Advance mode < 2 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_Write("CBT,0"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register CBT (Calibration Build Table). Command {CBT,x}, with x > number of probes in PA
// Set the target value (in dHz) and create the calibration correction table from the difference between the target and the measurement table (previously filled).
// rTarget must be bigger than number of probes in the probe-array
// The correction table will be store in main-unit RAM. To store it permanently in the probe-array EEPROM, call the ProbeArrayEepromWrite() method after it.
bool CMFC3045Instrument::CalibrationBuildTableWithTargetSet(const U32 & rTarget)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command("CBT,");
		U8 l_PaProbeNumber(0);

		if (eAdvancedMode::k2 != m_AdvancedMode)
			throw CMFC3045InsException(std::string("CBT Set not allowed. Advance mode < 2 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!NumberProbesGet(l_PaProbeNumber))
			throw CMFC3045InsException(std::string("Failed: Get PA Probes Number in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (rTarget <= m_NbProbesInPA)
			throw CMFC3045InsException(std::string("Failed: CBT target out of Range in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		l_Command += std::to_string(rTarget);
		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register CBT (Calibration Build Table). Command [CBT]
// Get all calibration correction table. The stored values, are the corrections in dHz to apply to each probe measurement to be normalized to the target value.
// Data[0] is the first probe correction. Data[n] is the n probe correction.
// Block mode must be 1 (kAllDecimal)
// rData.size will be = m_NbProbesInPA
bool CMFC3045Instrument::CalibrationBuildTableGet(std::vector<I32> & rData)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);
	rData.clear();
	try {
		U8 l_PaProbeNumber(0);

		if (!NumberProbesGet(l_PaProbeNumber)) // update and read m_NbProbesInPA
			throw CMFC3045InsException(std::string("Failed to get PA probes number in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if ((0 == m_NbProbesInPA) || (NPR_MAX < m_NbProbesInPA))
			throw CMFC3045InsException(std::string("Failed to get valid probes number from PA. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (eAdvancedMode::kNone == m_AdvancedMode)
			throw CMFC3045InsException(std::string("CBT Get not allowed. Advance mode < 1 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (eDataBlockMode::kAllDecimal != m_DataBlockMode)
			throw CMFC3045InsException(std::string("Failed. Wrong Block mode. Block mode must be 1 in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_Write("CBT"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		rData.resize(m_NbProbesInPA);
		std::string l_RdStr;
		for (size_t i = 0; i < rData.size(); i++)
		{
			if (!l_ReadString(l_RdStr))
				throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
			rData[i] = static_cast<I32>(std::stoul(l_RdStr));
		}
		// Read the End of transmission block char after the last data in block
		if(!l_ReadEndTransmissionBlockChar())
			throw CMFC3045InsException(std::string("Read failed ETB char in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register CBT (Calibration Build Table). Command [CBT,r], with r = a valid probe unmber 
// Get correction value (dHz) for a specific probe; (read a correction in main-unit RAM correction table)
// This method set as well a pointer to the correction table for a CET command
bool CMFC3045Instrument::CalibrationBuildProbeGet(I32 & rData, const U8 & rProbeSel)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command("CBT,");
		U8 l_PaProbeNumber(0);
		std::string l_RdStr;

		if (eAdvancedMode::kNone == m_AdvancedMode)
			throw CMFC3045InsException(std::string("CBT Get not allowed. Advance mode < 1 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!NumberProbesGet(l_PaProbeNumber))
			throw CMFC3045InsException(std::string("Failed: Get PA Probes Number in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if ((rProbeSel == 0) || (rProbeSel > m_NbProbesInPA))
			throw CMFC3045InsException(std::string("Failed, CBT target out of Range in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		l_Command += std::to_string(rProbeSel);
		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_RdStr))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		rData = static_cast<I32>(std::stoul(l_RdStr));
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register CET (Calibration Edit Table). Command {CET,x}, with x = correction value in dHz
// Set a correction value in dHz for a specific probe in the calibration correction table (in RAM). Correction value must be in range [-37768, 37767]
// The CalibrationBuildProbeGet() method must be called before to set a pointer to the desired probe in the correction table.
// The correction will be store in Main-Unit RAM. To Store it permanently in the Probe-Array EEPROM, call the ProbeArrayEepromWrite() method after it.
bool CMFC3045Instrument::CalibrationEditTableSet(const I32 & rData)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if ((rData < CORRECT_POBE_MIN_DHZ) || (rData > CORRECT_POBE_MAX_DHZ))
			throw CMFC3045InsException(std::string("Failed, Correction value is out of Range in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		std::string l_Command("CET,");

		if (eAdvancedMode::k2 != m_AdvancedMode)
			throw CMFC3045InsException(std::string("Failed, CET Set not allowed. Advance mode < 2 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		l_Command += std::to_string(rData);
		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register CET (Calibration Edit Table). Command {CET,x}, with x = correction value in dHz
// Set a correction value in dHz for a specific probe in the calibration correction table (in RAM). Correction value must be in range [-37768, 37767]
// The CalibrationBuildProbeGet() is called in this to set a pointer to the desired probe in the correction table.
// The correction will be store in Main-Unit RAM. To Store it permanently in the Probe-Array EEPROM, call the ProbeArrayEepromWrite() method after it.
bool CMFC3045Instrument::CalibrationEditTableSet(const I32 & rData, const U8 & rProbeSel)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		U8 l_probeIdxCET(0);
		I32 l_Correction(0);
		
		if(!CalibrationBuildProbeGet(l_Correction,rProbeSel))	// Read the correction value of a probe in the table to set the CET pointer at the correct position
			throw CMFC3045InsException(std::string("Failed to set pointer to the desired probe with CBT,r cmd in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		if(!CalibrationEditTableGetProbePointed(l_probeIdxCET))	// Read probe number pointed in table
			throw CMFC3045InsException(std::string("Failed with read CET cmd in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (l_probeIdxCET != rProbeSel)
			throw CMFC3045InsException(std::string("Failed, CET not at the deisred probe position in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if(!CalibrationEditTableSet(rData))			// write the new correction value for the desired probe
			throw CMFC3045InsException(std::string("Failed to edit Correction table for this probe in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;

}

// Read register CET (Calibration Edit Table). Command [CET]
// Get probe number pointed by CET (This allows to verify the probe position in table to be correct before edition).(Manual info is wrong!)
bool CMFC3045Instrument::CalibrationEditTableGetProbePointed(U8 & rProbeIdxCET)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_RdStr;
		if (eAdvancedMode::kNone == m_AdvancedMode)
			throw CMFC3045InsException(std::string("CET Get not allowed. Advance mode < 1 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_Write("CET"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(l_RdStr))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		rProbeIdxCET = static_cast<U8>(std::stoul(l_RdStr));
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register CDP (Calibration Date of Probe Array). Command {CDP,d}, with d = date in decimal format ddmmyy
// Set probe array calibration date (in string format ddmmyy )
// This method saves permanently the calibration date in PA-EEPROM
bool CMFC3045Instrument::CalibrationDatePASet(const U8 & rDayDD, const U8 & rMonthMM, const U8 & rYearYY, const bool & rSaveInEeprom)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::ostringstream l_StringFormated;
		std::string l_Command("CDP,");

		if (eAdvancedMode::k2 != m_AdvancedMode)
			throw CMFC3045InsException(std::string("CDP Set not allowed. Advance mode < 2 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		if(rDayDD < 1 || rDayDD >31)
			throw CMFC3045InsException(std::string("CDP Set not allowed. Wrong input Day format. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		l_StringFormated << std::setw(2) << std::setfill('0') << std::to_string(rDayDD);

		if (rMonthMM < 1 || rMonthMM > 12)
			throw CMFC3045InsException(std::string("CDP Set not allowed. Wrong input Month format. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		l_StringFormated << std::setw(2) << std::setfill('0') << std::to_string(rMonthMM);

		if (rYearYY > 99)
			throw CMFC3045InsException(std::string("CDP Set not allowed. Wrong input Year format. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		l_StringFormated << std::setw(2) << std::setfill('0') << std::to_string(rYearYY);
		l_Command += l_StringFormated.str();
		
		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if(rSaveInEeprom)
		{
			if(!ProbeArrayEepromWrite()) // save Calibration date in PA EEPROM
				throw CMFC3045InsException(std::string("Failed to save in PA EEPROM in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		}
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register CDP (Calibration Date of Probe Array). Command {CDP,d}, with d = date in decimal format ddmmyy
// Set probe array calibration date from an UTC date format (in second since 1970)
// This method saves permanently the calibration date in probe array EEPROM
bool CMFC3045Instrument::CalibrationDatePASet(const std::time_t & rUtcDate, const bool & rSaveInEeprom)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command("CDP,");
		std::string l_StrDate("");

		if (eAdvancedMode::k2 != m_AdvancedMode)
			throw CMFC3045InsException(std::string("CDP Set not allowed. Advance mode < 2 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		if (!l_DateTimeToDateStrUtc(rUtcDate, l_StrDate)) //convert UTC time_t format to string date format
			throw CMFC3045InsException("Failed to convert date from UTC time_t format", MTL__LOCATION__);
		l_Command += l_StrDate;

		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if(rSaveInEeprom)
		{
			if (!ProbeArrayEepromWrite()) // save Calibration date in PA EEPROM
				throw CMFC3045InsException(std::string("Failed to save in PA EEPROM in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		}
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register CDP (Calibration Date of Probe Array). Command [CDP]
// Get probe array calibration date. Get string as date (local time) format "dd/mm/yy", and get a time_t (at UTC) date format in second since 1970 UTC
bool CMFC3045Instrument::CalibrationDatePAGet(std::string & rStrDate, std::time_t & rUtcDate)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (eAdvancedMode::kNone == m_AdvancedMode)
			throw CMFC3045InsException(std::string("CDP Get not allowed. Advance mode < 1 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_Write("CDP"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(rStrDate))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		if (!l_DateStrToDateTimeUtc(rStrDate, rUtcDate)) //convert string date (UTC) format to time_t UTC
			throw CMFC3045InsException("Failed to convert UTC string date in time_t format", MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		rStrDate.clear();
		rUtcDate = -1;
		return false;
	}
	return true;
}

// Write register CDU (Calibration Date of main Unit). Command {CDU,d}, with d = date in decimal format ddmmyy
// Set main unit calibration date (in string format ddmmyy )
// This method saves permanently the calibration date in main unit EEPROM
bool CMFC3045Instrument::CalibrationDateMUSet(const U8 & rDayDD, const U8 & rMonthMM, const U8 & rYearYY, const bool & rSaveInEeprom)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::ostringstream l_StringFormated;
		std::string l_Command("CDU,");

		if (eAdvancedMode::k2 != m_AdvancedMode)
			throw CMFC3045InsException(std::string("CDU Set not allowed. Advance mode < 2 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		if (rDayDD < 1 || rDayDD >31)
			throw CMFC3045InsException(std::string("CDU Set not allowed. Wrong input Date format. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		l_StringFormated << std::setw(2) << std::setfill('0') << std::to_string(rDayDD); //left padding with 0
		

		if (rMonthMM < 1 || rMonthMM > 12)
			throw CMFC3045InsException(std::string("CDU Set not allowed. Wrong input Date format. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		l_StringFormated << std::setw(2) << std::setfill('0') << std::to_string(rMonthMM); //left padding with 0
		

		if (rYearYY > 99)
			throw CMFC3045InsException(std::string("CDU Set not allowed. Wrong input Date format. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		l_StringFormated << std::setw(2) << std::setfill('0') << std::to_string(rYearYY); //left padding with 0
		l_Command += l_StringFormated.str();

		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if(rSaveInEeprom)
		{
			if (!MainUnitEepromWrite()) // save Calibration date in the Main-Unit EEPROM
				throw CMFC3045InsException(std::string("Failed to save in MU EEPROM in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		}
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Write register CDU (Calibration Date of main Unit). Command {CDU,d}, with d = date in decimal format ddmmyy
// Set  main unit calibration date from an UTC date format (in second since 1970)
// This method saves permanently the calibration date in main unit-EEPROM
bool CMFC3045Instrument::CalibrationDateMUSet(const std::time_t & rUtcDate, const bool & rSaveInEeprom)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command("CDU,");
		std::string l_StrDate("");

		if (eAdvancedMode::k2 != m_AdvancedMode)
			throw CMFC3045InsException(std::string("CDU Set not allowed. Advance mode < 2 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		if (!l_DateTimeToDateStrUtc(rUtcDate, l_StrDate)) //convert UTC time_t format to string date format
			throw CMFC3045InsException("Failed to convert date from UTC time_t format", MTL__LOCATION__);
		l_Command += l_StrDate;

		if (!l_Write(l_Command))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if(rSaveInEeprom)
		{
			if (!MainUnitEepromWrite()) // save Calibration date in the Main-Unit EEPROM
				throw CMFC3045InsException(std::string("Failed to save in MU EEPROM in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		}
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		return false;
	}
	return true;
}

// Read register CDU (Calibration Date of main Unit). Command [CDU]
// Get main unit calibration date. Returned string date format is "dd/mm/yy", UTC date format is in second since 1970
bool CMFC3045Instrument::CalibrationDateMUGet(std::string & rStrDate, std::time_t & rUtcDate)
{
	MTL_INSTRUMENT_MFC3045_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (eAdvancedMode::kNone == m_AdvancedMode)
			throw CMFC3045InsException(std::string("CDU Get not allowed. Advance mode < 1 in MFC3045. ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_Write("CDU"))
			throw CMFC3045InsException(std::string("Write failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (!l_ReadString(rStrDate))
			throw CMFC3045InsException(std::string("Read failed in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		if (!l_DateStrToDateTimeUtc(rStrDate, rUtcDate)) //convert string date UTC format to time_t UTC
			throw CMFC3045InsException("Failed to convert date string in UTC format", MTL__LOCATION__);
	}
	catch (CMFC3045InsException & rE)
	{
		MTL_INSTRUMENT_MFC3045_DEBUG_CERR(rE.what());
		MTL_Unused(rE)
		rStrDate.clear();
		rUtcDate = -1;
		return false;
	}
	return true;
}


